/*
	Copyright (C) 2006 yopyop
	Copyright (C) 2007-2012 DeSmuME team
	This file is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	This file is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the this software.  If not, see <http://www.gnu.org/licenses/>.
*/

/* 00 0000 0000 */ TABDECL(OP_LSL_0),		// OP_LSL_0
/* 00 0000 0001 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 0010 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 0011 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 0100 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 0101 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 0110 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 0111 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 1000 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 1001 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 1010 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 1011 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 1100 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 1101 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 1110 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0000 1111 */ TABDECL(OP_LSL),		// OP_LSL

/* 00 0001 0000 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 0001 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 0010 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 0011 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 0100 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 0101 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 0110 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 0111 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 1000 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 1001 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 1010 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 1011 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 1100 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 1101 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 1110 */ TABDECL(OP_LSL),		// OP_LSL
/* 00 0001 1111 */ TABDECL(OP_LSL),		// OP_LSL

/* 00 0010 0000 */ TABDECL(OP_LSR_0),		// OP_LSR_0
/* 00 0010 0001 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 0010 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 0011 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 0100 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 0101 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 0110 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 0111 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 1000 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 1001 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 1010 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 1011 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 1100 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 1101 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 1110 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0010 1111 */ TABDECL(OP_LSR),		// OP_LSR

/* 00 0011 0000 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 0001 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 0010 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 0011 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 0100 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 0101 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 0110 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 0111 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 1000 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 1001 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 1010 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 1011 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 1100 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 1101 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 1110 */ TABDECL(OP_LSR),		// OP_LSR
/* 00 0011 1111 */ TABDECL(OP_LSR),		// OP_LSR

/* 00 0100 0000 */ TABDECL(OP_ASR_0),		// OP_ASR_0
/* 00 0100 0001 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 0010 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 0011 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 0100 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 0101 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 0110 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 0111 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 1000 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 1001 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 1010 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 1011 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 1100 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 1101 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 1110 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0100 1111 */ TABDECL(OP_ASR),		// OP_ASR

/* 00 0101 0000 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 0001 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 0010 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 0011 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 0100 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 0101 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 0110 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 0111 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 1000 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 1001 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 1010 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 1011 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 1100 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 1101 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 1110 */ TABDECL(OP_ASR),		// OP_ASR
/* 00 0101 1111 */ TABDECL(OP_ASR),		// OP_ASR

/* 00 0110 0000 */ TABDECL(OP_ADD_REG),		// OP_ADD_REG
/* 00 0110 0001 */ TABDECL(OP_ADD_REG),		// OP_ADD_REG
/* 00 0110 0010 */ TABDECL(OP_ADD_REG),		// OP_ADD_REG
/* 00 0110 0011 */ TABDECL(OP_ADD_REG),		// OP_ADD_REG
/* 00 0110 0100 */ TABDECL(OP_ADD_REG),		// OP_ADD_REG
/* 00 0110 0101 */ TABDECL(OP_ADD_REG),		// OP_ADD_REG
/* 00 0110 0110 */ TABDECL(OP_ADD_REG),		// OP_ADD_REG
/* 00 0110 0111 */ TABDECL(OP_ADD_REG),		// OP_ADD_REG
/* 00 0110 1000 */ TABDECL(OP_SUB_REG),		// OP_SUB_REG
/* 00 0110 1001 */ TABDECL(OP_SUB_REG),		// OP_SUB_REG
/* 00 0110 1010 */ TABDECL(OP_SUB_REG),		// OP_SUB_REG
/* 00 0110 1011 */ TABDECL(OP_SUB_REG),		// OP_SUB_REG
/* 00 0110 1100 */ TABDECL(OP_SUB_REG),		// OP_SUB_REG
/* 00 0110 1101 */ TABDECL(OP_SUB_REG),		// OP_SUB_REG
/* 00 0110 1110 */ TABDECL(OP_SUB_REG),		// OP_SUB_REG
/* 00 0110 1111 */ TABDECL(OP_SUB_REG),		// OP_SUB_REG

/* 00 0111 0000 */ TABDECL(OP_ADD_IMM3),		// OP_ADD_IMM3
/* 00 0111 0001 */ TABDECL(OP_ADD_IMM3),		// OP_ADD_IMM3
/* 00 0111 0010 */ TABDECL(OP_ADD_IMM3),		// OP_ADD_IMM3
/* 00 0111 0011 */ TABDECL(OP_ADD_IMM3),		// OP_ADD_IMM3
/* 00 0111 0100 */ TABDECL(OP_ADD_IMM3),		// OP_ADD_IMM3
/* 00 0111 0101 */ TABDECL(OP_ADD_IMM3),		// OP_ADD_IMM3
/* 00 0111 0110 */ TABDECL(OP_ADD_IMM3),		// OP_ADD_IMM3
/* 00 0111 0111 */ TABDECL(OP_ADD_IMM3),		// OP_ADD_IMM3
/* 00 0111 1000 */ TABDECL(OP_SUB_IMM3),		// OP_SUB_IMM3
/* 00 0111 1001 */ TABDECL(OP_SUB_IMM3),		// OP_SUB_IMM3
/* 00 0111 1010 */ TABDECL(OP_SUB_IMM3),		// OP_SUB_IMM3
/* 00 0111 1011 */ TABDECL(OP_SUB_IMM3),		// OP_SUB_IMM3
/* 00 0111 1100 */ TABDECL(OP_SUB_IMM3),		// OP_SUB_IMM3
/* 00 0111 1101 */ TABDECL(OP_SUB_IMM3),		// OP_SUB_IMM3
/* 00 0111 1110 */ TABDECL(OP_SUB_IMM3),		// OP_SUB_IMM3
/* 00 0111 1111 */ TABDECL(OP_SUB_IMM3),		// OP_SUB_IMM3

/* 00 1000 0000 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 0001 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 0010 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 0011 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 0100 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 0101 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 0110 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 0111 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 1000 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 1001 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 1010 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 1011 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 1100 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 1101 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 1110 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1000 1111 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8

/* 00 1001 0000 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 0001 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 0010 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 0011 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 0100 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 0101 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 0110 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 0111 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 1000 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 1001 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 1010 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 1011 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 1100 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 1101 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 1110 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8
/* 00 1001 1111 */ TABDECL(OP_MOV_IMM8),		// OP_MOV_IMM8

/* 00 1010 0000 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 0001 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 0010 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 0011 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 0100 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 0101 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 0110 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 0111 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 1000 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 1001 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 1010 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 1011 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 1100 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 1101 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 1110 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1010 1111 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8

/* 00 1011 0000 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 0001 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 0010 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 0011 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 0100 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 0101 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 0110 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 0111 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 1000 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 1001 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 1010 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 1011 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 1100 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 1101 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 1110 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8
/* 00 1011 1111 */ TABDECL(OP_CMP_IMM8),		// OP_CMP_IMM8

/* 00 1100 0000 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 0001 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 0010 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 0011 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 0100 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 0101 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 0110 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 0111 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 1000 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 1001 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 1010 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 1011 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 1100 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 1101 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 1110 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1100 1111 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8

/* 00 1101 0000 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 0001 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 0010 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 0011 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 0100 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 0101 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 0110 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 0111 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 1000 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 1001 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 1010 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 1011 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 1100 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 1101 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 1110 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8
/* 00 1101 1111 */ TABDECL(OP_ADD_IMM8),		// OP_ADD_IMM8

/* 00 1110 0000 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 0001 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 0010 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 0011 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 0100 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 0101 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 0110 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 0111 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 1000 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 1001 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 1010 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 1011 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 1100 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 1101 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 1110 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1110 1111 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8

/* 00 1111 0000 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 0001 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 0010 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 0011 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 0100 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 0101 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 0110 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 0111 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 1000 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 1001 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 1010 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 1011 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 1100 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 1101 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 1110 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8
/* 00 1111 1111 */ TABDECL(OP_SUB_IMM8),		// OP_SUB_IMM8

/* 01 0000 0000 */ TABDECL(OP_AND),		// OP_AND
/* 01 0000 0001 */ TABDECL(OP_EOR),		// OP_EOR
/* 01 0000 0010 */ TABDECL(OP_LSL_REG),		// OP_LSL_REG
/* 01 0000 0011 */ TABDECL(OP_LSR_REG),		// OP_LSR_REG
/* 01 0000 0100 */ TABDECL(OP_ASR_REG),		// OP_ASR_REG
/* 01 0000 0101 */ TABDECL(OP_ADC_REG),		// OP_ADC_REG
/* 01 0000 0110 */ TABDECL(OP_SBC_REG),		// OP_SBC_REG
/* 01 0000 0111 */ TABDECL(OP_ROR_REG),		// OP_ROR_REG
/* 01 0000 1000 */ TABDECL(OP_TST),		// OP_TST
/* 01 0000 1001 */ TABDECL(OP_NEG),		// OP_NEG
/* 01 0000 1010 */ TABDECL(OP_CMP),		// OP_CMP
/* 01 0000 1011 */ TABDECL(OP_CMN),		// OP_CMN
/* 01 0000 1100 */ TABDECL(OP_ORR),		// OP_ORR
/* 01 0000 1101 */ TABDECL(OP_MUL_REG),		// OP_MUL_REG
/* 01 0000 1110 */ TABDECL(OP_BIC),		// OP_BIC
/* 01 0000 1111 */ TABDECL(OP_MVN),		// OP_MVN

/* 01 0001 0000 */ TABDECL(OP_ADD_SPE),		// OP_ADD_SPE
/* 01 0001 0001 */ TABDECL(OP_ADD_SPE),		// OP_ADD_SPE
/* 01 0001 0010 */ TABDECL(OP_ADD_SPE),		// OP_ADD_SPE
/* 01 0001 0011 */ TABDECL(OP_ADD_SPE),		// OP_ADD_SPE
/* 01 0001 0100 */ TABDECL(OP_CMP_SPE),		// OP_CMP_SPE
/* 01 0001 0101 */ TABDECL(OP_CMP_SPE),		// OP_CMP_SPE
/* 01 0001 0110 */ TABDECL(OP_CMP_SPE),		// OP_CMP_SPE
/* 01 0001 0111 */ TABDECL(OP_CMP_SPE),		// OP_CMP_SPE
/* 01 0001 1000 */ TABDECL(OP_MOV_SPE),		// OP_MOV_SPE
/* 01 0001 1001 */ TABDECL(OP_MOV_SPE),		// OP_MOV_SPE
/* 01 0001 1010 */ TABDECL(OP_MOV_SPE),		// OP_MOV_SPE
/* 01 0001 1011 */ TABDECL(OP_MOV_SPE),		// OP_MOV_SPE
/* 01 0001 1100 */ TABDECL(OP_BX_THUMB),		// OP_BX_THUMB
/* 01 0001 1101 */ TABDECL(OP_BX_THUMB),		// OP_BX_THUMB
/* 01 0001 1110 */ TABDECL(OP_BLX_THUMB),		// OP_BLX_THUMB
/* 01 0001 1111 */ TABDECL(OP_BLX_THUMB),		// OP_BLX_THUMB

/* 01 0010 0000 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 0001 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 0010 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 0011 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 0100 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 0101 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 0110 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 0111 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 1000 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 1001 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 1010 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 1011 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 1100 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 1101 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 1110 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0010 1111 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL

/* 01 0011 0000 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 0001 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 0010 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 0011 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 0100 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 0101 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 0110 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 0111 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 1000 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 1001 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 1010 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 1011 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 1100 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 1101 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 1110 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL
/* 01 0011 1111 */ TABDECL(OP_LDR_PCREL),		// OP_LDR_PCREL

/* 01 0100 0000 */ TABDECL(OP_STR_REG_OFF),		// OP_STR_REG_OFF
/* 01 0100 0001 */ TABDECL(OP_STR_REG_OFF),		// OP_STR_REG_OFF
/* 01 0100 0010 */ TABDECL(OP_STR_REG_OFF),		// OP_STR_REG_OFF
/* 01 0100 0011 */ TABDECL(OP_STR_REG_OFF),		// OP_STR_REG_OFF
/* 01 0100 0100 */ TABDECL(OP_STR_REG_OFF),		// OP_STR_REG_OFF
/* 01 0100 0101 */ TABDECL(OP_STR_REG_OFF),		// OP_STR_REG_OFF
/* 01 0100 0110 */ TABDECL(OP_STR_REG_OFF),		// OP_STR_REG_OFF
/* 01 0100 0111 */ TABDECL(OP_STR_REG_OFF),		// OP_STR_REG_OFF
/* 01 0100 1000 */ TABDECL(OP_STRH_REG_OFF),		// OP_STRH_REG_OFF
/* 01 0100 1001 */ TABDECL(OP_STRH_REG_OFF),		// OP_STRH_REG_OFF
/* 01 0100 1010 */ TABDECL(OP_STRH_REG_OFF),		// OP_STRH_REG_OFF
/* 01 0100 1011 */ TABDECL(OP_STRH_REG_OFF),		// OP_STRH_REG_OFF
/* 01 0100 1100 */ TABDECL(OP_STRH_REG_OFF),		// OP_STRH_REG_OFF
/* 01 0100 1101 */ TABDECL(OP_STRH_REG_OFF),		// OP_STRH_REG_OFF
/* 01 0100 1110 */ TABDECL(OP_STRH_REG_OFF),		// OP_STRH_REG_OFF
/* 01 0100 1111 */ TABDECL(OP_STRH_REG_OFF),		// OP_STRH_REG_OFF

/* 01 0101 0000 */ TABDECL(OP_STRB_REG_OFF),		// OP_STRB_REG_OFF
/* 01 0101 0001 */ TABDECL(OP_STRB_REG_OFF),		// OP_STRB_REG_OFF
/* 01 0101 0010 */ TABDECL(OP_STRB_REG_OFF),		// OP_STRB_REG_OFF
/* 01 0101 0011 */ TABDECL(OP_STRB_REG_OFF),		// OP_STRB_REG_OFF
/* 01 0101 0100 */ TABDECL(OP_STRB_REG_OFF),		// OP_STRB_REG_OFF
/* 01 0101 0101 */ TABDECL(OP_STRB_REG_OFF),		// OP_STRB_REG_OFF
/* 01 0101 0110 */ TABDECL(OP_STRB_REG_OFF),		// OP_STRB_REG_OFF
/* 01 0101 0111 */ TABDECL(OP_STRB_REG_OFF),		// OP_STRB_REG_OFF
/* 01 0101 1000 */ TABDECL(OP_LDRSB_REG_OFF),		// OP_LDRSB_REG_OFF
/* 01 0101 1001 */ TABDECL(OP_LDRSB_REG_OFF),		// OP_LDRSB_REG_OFF
/* 01 0101 1010 */ TABDECL(OP_LDRSB_REG_OFF),		// OP_LDRSB_REG_OFF
/* 01 0101 1011 */ TABDECL(OP_LDRSB_REG_OFF),		// OP_LDRSB_REG_OFF
/* 01 0101 1100 */ TABDECL(OP_LDRSB_REG_OFF),		// OP_LDRSB_REG_OFF
/* 01 0101 1101 */ TABDECL(OP_LDRSB_REG_OFF),		// OP_LDRSB_REG_OFF
/* 01 0101 1110 */ TABDECL(OP_LDRSB_REG_OFF),		// OP_LDRSB_REG_OFF
/* 01 0101 1111 */ TABDECL(OP_LDRSB_REG_OFF),		// OP_LDRSB_REG_OFF

/* 01 0110 0000 */ TABDECL(OP_LDR_REG_OFF),		// OP_LDR_REG_OFF
/* 01 0110 0001 */ TABDECL(OP_LDR_REG_OFF),		// OP_LDR_REG_OFF
/* 01 0110 0010 */ TABDECL(OP_LDR_REG_OFF),		// OP_LDR_REG_OFF
/* 01 0110 0011 */ TABDECL(OP_LDR_REG_OFF),		// OP_LDR_REG_OFF
/* 01 0110 0100 */ TABDECL(OP_LDR_REG_OFF),		// OP_LDR_REG_OFF
/* 01 0110 0101 */ TABDECL(OP_LDR_REG_OFF),		// OP_LDR_REG_OFF
/* 01 0110 0110 */ TABDECL(OP_LDR_REG_OFF),		// OP_LDR_REG_OFF
/* 01 0110 0111 */ TABDECL(OP_LDR_REG_OFF),		// OP_LDR_REG_OFF
/* 01 0110 1000 */ TABDECL(OP_LDRH_REG_OFF),		// OP_LDRH_REG_OFF
/* 01 0110 1001 */ TABDECL(OP_LDRH_REG_OFF),		// OP_LDRH_REG_OFF
/* 01 0110 1010 */ TABDECL(OP_LDRH_REG_OFF),		// OP_LDRH_REG_OFF
/* 01 0110 1011 */ TABDECL(OP_LDRH_REG_OFF),		// OP_LDRH_REG_OFF
/* 01 0110 1100 */ TABDECL(OP_LDRH_REG_OFF),		// OP_LDRH_REG_OFF
/* 01 0110 1101 */ TABDECL(OP_LDRH_REG_OFF),		// OP_LDRH_REG_OFF
/* 01 0110 1110 */ TABDECL(OP_LDRH_REG_OFF),		// OP_LDRH_REG_OFF
/* 01 0110 1111 */ TABDECL(OP_LDRH_REG_OFF),		// OP_LDRH_REG_OFF

/* 01 0111 0000 */ TABDECL(OP_LDRB_REG_OFF),		// OP_LDRB_REG_OFF
/* 01 0111 0001 */ TABDECL(OP_LDRB_REG_OFF),		// OP_LDRB_REG_OFF
/* 01 0111 0010 */ TABDECL(OP_LDRB_REG_OFF),		// OP_LDRB_REG_OFF
/* 01 0111 0011 */ TABDECL(OP_LDRB_REG_OFF),		// OP_LDRB_REG_OFF
/* 01 0111 0100 */ TABDECL(OP_LDRB_REG_OFF),		// OP_LDRB_REG_OFF
/* 01 0111 0101 */ TABDECL(OP_LDRB_REG_OFF),		// OP_LDRB_REG_OFF
/* 01 0111 0110 */ TABDECL(OP_LDRB_REG_OFF),		// OP_LDRB_REG_OFF
/* 01 0111 0111 */ TABDECL(OP_LDRB_REG_OFF),		// OP_LDRB_REG_OFF
/* 01 0111 1000 */ TABDECL(OP_LDRSH_REG_OFF),		// OP_LDRSH_REG_OFF
/* 01 0111 1001 */ TABDECL(OP_LDRSH_REG_OFF),		// OP_LDRSH_REG_OFF
/* 01 0111 1010 */ TABDECL(OP_LDRSH_REG_OFF),		// OP_LDRSH_REG_OFF
/* 01 0111 1011 */ TABDECL(OP_LDRSH_REG_OFF),		// OP_LDRSH_REG_OFF
/* 01 0111 1100 */ TABDECL(OP_LDRSH_REG_OFF),		// OP_LDRSH_REG_OFF
/* 01 0111 1101 */ TABDECL(OP_LDRSH_REG_OFF),		// OP_LDRSH_REG_OFF
/* 01 0111 1110 */ TABDECL(OP_LDRSH_REG_OFF),		// OP_LDRSH_REG_OFF
/* 01 0111 1111 */ TABDECL(OP_LDRSH_REG_OFF),		// OP_LDRSH_REG_OFF

/* 01 1000 0000 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 0001 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 0010 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 0011 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 0100 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 0101 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 0110 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 0111 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 1000 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 1001 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 1010 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 1011 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 1100 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 1101 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 1110 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1000 1111 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF

/* 01 1001 0000 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 0001 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 0010 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 0011 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 0100 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 0101 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 0110 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 0111 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 1000 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 1001 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 1010 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 1011 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 1100 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 1101 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 1110 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF
/* 01 1001 1111 */ TABDECL(OP_STR_IMM_OFF),		// OP_STR_IMM_OFF

/* 01 1010 0000 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 0001 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 0010 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 0011 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 0100 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 0101 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 0110 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 0111 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 1000 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 1001 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 1010 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 1011 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 1100 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 1101 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 1110 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1010 1111 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF

/* 01 1011 0000 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 0001 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 0010 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 0011 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 0100 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 0101 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 0110 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 0111 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 1000 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 1001 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 1010 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 1011 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 1100 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 1101 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 1110 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF
/* 01 1011 1111 */ TABDECL(OP_LDR_IMM_OFF),		// OP_LDR_IMM_OFF

/* 01 1100 0000 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 0001 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 0010 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 0011 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 0100 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 0101 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 0110 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 0111 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 1000 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 1001 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 1010 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 1011 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 1100 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 1101 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 1110 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1100 1111 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF

/* 01 1101 0000 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 0001 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 0010 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 0011 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 0100 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 0101 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 0110 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 0111 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 1000 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 1001 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 1010 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 1011 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 1100 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 1101 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 1110 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF
/* 01 1101 1111 */ TABDECL(OP_STRB_IMM_OFF),		// OP_STRB_IMM_OFF

/* 01 1110 0000 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 0001 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 0010 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 0011 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 0100 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 0101 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 0110 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 0111 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 1000 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 1001 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 1010 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 1011 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 1100 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 1101 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 1110 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1110 1111 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF

/* 01 1111 0000 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 0001 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 0010 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 0011 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 0100 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 0101 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 0110 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 0111 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 1000 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 1001 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 1010 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 1011 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 1100 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 1101 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 1110 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF
/* 01 1111 1111 */ TABDECL(OP_LDRB_IMM_OFF),		// OP_LDRB_IMM_OFF

/* 10 0000 0000 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 0001 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 0010 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 0011 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 0100 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 0101 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 0110 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 0111 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 1000 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 1001 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 1010 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 1011 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 1100 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 1101 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 1110 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0000 1111 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF

/* 10 0001 0000 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 0001 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 0010 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 0011 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 0100 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 0101 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 0110 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 0111 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 1000 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 1001 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 1010 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 1011 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 1100 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 1101 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 1110 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF
/* 10 0001 1111 */ TABDECL(OP_STRH_IMM_OFF),		// OP_STRH_IMM_OFF

/* 10 0010 0000 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 0001 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 0010 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 0011 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 0100 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 0101 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 0110 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 0111 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 1000 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 1001 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 1010 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 1011 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 1100 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 1101 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 1110 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0010 1111 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF

/* 10 0011 0000 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 0001 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 0010 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 0011 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 0100 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 0101 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 0110 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 0111 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 1000 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 1001 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 1010 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 1011 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 1100 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 1101 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 1110 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF
/* 10 0011 1111 */ TABDECL(OP_LDRH_IMM_OFF),		// OP_LDRH_IMM_OFF

/* 10 0100 0000 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 0001 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 0010 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 0011 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 0100 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 0101 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 0110 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 0111 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 1000 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 1001 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 1010 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 1011 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 1100 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 1101 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 1110 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0100 1111 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL

/* 10 0101 0000 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 0001 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 0010 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 0011 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 0100 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 0101 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 0110 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 0111 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 1000 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 1001 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 1010 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 1011 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 1100 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 1101 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 1110 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL
/* 10 0101 1111 */ TABDECL(OP_STR_SPREL),		// OP_STR_SPREL

/* 10 0110 0000 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 0001 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 0010 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 0011 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 0100 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 0101 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 0110 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 0111 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 1000 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 1001 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 1010 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 1011 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 1100 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 1101 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 1110 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0110 1111 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL

/* 10 0111 0000 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 0001 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 0010 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 0011 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 0100 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 0101 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 0110 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 0111 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 1000 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 1001 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 1010 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 1011 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 1100 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 1101 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 1110 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL
/* 10 0111 1111 */ TABDECL(OP_LDR_SPREL),		// OP_LDR_SPREL

/* 10 1000 0000 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 0001 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 0010 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 0011 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 0100 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 0101 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 0110 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 0111 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 1000 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 1001 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 1010 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 1011 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 1100 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 1101 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 1110 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1000 1111 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC

/* 10 1001 0000 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 0001 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 0010 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 0011 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 0100 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 0101 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 0110 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 0111 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 1000 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 1001 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 1010 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 1011 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 1100 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 1101 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 1110 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC
/* 10 1001 1111 */ TABDECL(OP_ADD_2PC),		// OP_ADD_2PC

/* 10 1010 0000 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 0001 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 0010 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 0011 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 0100 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 0101 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 0110 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 0111 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 1000 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 1001 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 1010 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 1011 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 1100 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 1101 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 1110 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1010 1111 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP

/* 10 1011 0000 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 0001 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 0010 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 0011 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 0100 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 0101 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 0110 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 0111 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 1000 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 1001 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 1010 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 1011 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 1100 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 1101 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 1110 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP
/* 10 1011 1111 */ TABDECL(OP_ADD_2SP),		// OP_ADD_2SP

/* 10 1100 0000 */ TABDECL(OP_ADJUST_P_SP),		// OP_ADJUST_P_SP
/* 10 1100 0001 */ TABDECL(OP_ADJUST_P_SP),		// OP_ADJUST_P_SP
/* 10 1100 0010 */ TABDECL(OP_ADJUST_M_SP),		// OP_ADJUST_M_SP
/* 10 1100 0011 */ TABDECL(OP_ADJUST_M_SP),		// OP_ADJUST_M_SP
/* 10 1100 0100 */ TABDECL(OP_UND_THUMB),
/* 10 1100 0101 */ TABDECL(OP_UND_THUMB),
/* 10 1100 0110 */ TABDECL(OP_UND_THUMB),
/* 10 1100 0111 */ TABDECL(OP_UND_THUMB),
/* 10 1100 1000 */ TABDECL(OP_UND_THUMB),
/* 10 1100 1001 */ TABDECL(OP_UND_THUMB),
/* 10 1100 1010 */ TABDECL(OP_UND_THUMB),
/* 10 1100 1011 */ TABDECL(OP_UND_THUMB),
/* 10 1100 1100 */ TABDECL(OP_UND_THUMB),
/* 10 1100 1101 */ TABDECL(OP_UND_THUMB),
/* 10 1100 1110 */ TABDECL(OP_UND_THUMB),
/* 10 1100 1111 */ TABDECL(OP_UND_THUMB),

/* 10 1101 0000 */ TABDECL(OP_PUSH),		// OP_PUSH
/* 10 1101 0001 */ TABDECL(OP_PUSH),		// OP_PUSH
/* 10 1101 0010 */ TABDECL(OP_PUSH),		// OP_PUSH
/* 10 1101 0011 */ TABDECL(OP_PUSH),		// OP_PUSH
/* 10 1101 0100 */ TABDECL(OP_PUSH_LR),		// OP_PUSH_LR
/* 10 1101 0101 */ TABDECL(OP_PUSH_LR),		// OP_PUSH_LR
/* 10 1101 0110 */ TABDECL(OP_PUSH_LR),		// OP_PUSH_LR
/* 10 1101 0111 */ TABDECL(OP_PUSH_LR),		// OP_PUSH_LR
/* 10 1101 1000 */ TABDECL(OP_UND_THUMB),
/* 10 1101 1001 */ TABDECL(OP_UND_THUMB),
/* 10 1101 1010 */ TABDECL(OP_UND_THUMB),
/* 10 1101 1011 */ TABDECL(OP_UND_THUMB),
/* 10 1101 1100 */ TABDECL(OP_UND_THUMB),
/* 10 1101 1101 */ TABDECL(OP_UND_THUMB),
/* 10 1101 1110 */ TABDECL(OP_UND_THUMB),
/* 10 1101 1111 */ TABDECL(OP_UND_THUMB),

/* 10 1110 0000 */ TABDECL(OP_UND_THUMB),
/* 10 1110 0001 */ TABDECL(OP_UND_THUMB),
/* 10 1110 0010 */ TABDECL(OP_UND_THUMB),
/* 10 1110 0011 */ TABDECL(OP_UND_THUMB),
/* 10 1110 0100 */ TABDECL(OP_UND_THUMB),
/* 10 1110 0101 */ TABDECL(OP_UND_THUMB),
/* 10 1110 0110 */ TABDECL(OP_UND_THUMB),
/* 10 1110 0111 */ TABDECL(OP_UND_THUMB),
/* 10 1110 1000 */ TABDECL(OP_UND_THUMB),
/* 10 1110 1001 */ TABDECL(OP_UND_THUMB),
/* 10 1110 1010 */ TABDECL(OP_UND_THUMB),
/* 10 1110 1011 */ TABDECL(OP_UND_THUMB),
/* 10 1110 1100 */ TABDECL(OP_UND_THUMB),
/* 10 1110 1101 */ TABDECL(OP_UND_THUMB),
/* 10 1110 1110 */ TABDECL(OP_UND_THUMB),
/* 10 1110 1111 */ TABDECL(OP_UND_THUMB),

/* 10 1111 0000 */ TABDECL(OP_POP),		// OP_POP
/* 10 1111 0001 */ TABDECL(OP_POP),		// OP_POP
/* 10 1111 0010 */ TABDECL(OP_POP),		// OP_POP
/* 10 1111 0011 */ TABDECL(OP_POP),		// OP_POP
/* 10 1111 0100 */ TABDECL(OP_POP_PC),		// OP_POP_PC
/* 10 1111 0101 */ TABDECL(OP_POP_PC),		// OP_POP_PC
/* 10 1111 0110 */ TABDECL(OP_POP_PC),		// OP_POP_PC
/* 10 1111 0111 */ TABDECL(OP_POP_PC),		// OP_POP_PC
/* 10 1111 1000 */ TABDECL(OP_BKPT_THUMB),		// OP_BKPT_THUMB
/* 10 1111 1001 */ TABDECL(OP_BKPT_THUMB),		// OP_BKPT_THUMB
/* 10 1111 1010 */ TABDECL(OP_BKPT_THUMB),		// OP_BKPT_THUMB
/* 10 1111 1011 */ TABDECL(OP_BKPT_THUMB),		// OP_BKPT_THUMB
/* 10 1111 1100 */ TABDECL(OP_UND_THUMB),
/* 10 1111 1101 */ TABDECL(OP_UND_THUMB),
/* 10 1111 1110 */ TABDECL(OP_UND_THUMB),
/* 10 1111 1111 */ TABDECL(OP_UND_THUMB),

/* 11 0000 0000 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 0001 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 0010 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 0011 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 0100 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 0101 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 0110 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 0111 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 1000 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 1001 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 1010 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 1011 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 1100 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 1101 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 1110 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0000 1111 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB

/* 11 0001 0000 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 0001 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 0010 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 0011 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 0100 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 0101 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 0110 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 0111 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 1000 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 1001 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 1010 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 1011 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 1100 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 1101 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 1110 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB
/* 11 0001 1111 */ TABDECL(OP_STMIA_THUMB),		// OP_STMIA_THUMB

/* 11 0010 0000 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 0001 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 0010 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 0011 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 0100 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 0101 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 0110 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 0111 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 1000 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 1001 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 1010 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 1011 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 1100 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 1101 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 1110 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0010 1111 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB

/* 11 0011 0000 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 0001 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 0010 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 0011 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 0100 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 0101 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 0110 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 0111 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 1000 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 1001 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 1010 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 1011 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 1100 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 1101 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 1110 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB
/* 11 0011 1111 */ TABDECL(OP_LDMIA_THUMB),		// OP_LDMIA_THUMB

/* 11 0100 0000 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 0001 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 0010 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 0011 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 0100 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 0101 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 0110 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 0111 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 1000 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 1001 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 1010 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 1011 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 1100 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 1101 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 1110 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0100 1111 */ TABDECL(OP_B_COND),		// OP_B_COND

/* 11 0101 0000 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 0001 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 0010 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 0011 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 0100 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 0101 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 0110 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 0111 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 1000 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 1001 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 1010 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 1011 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 1100 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 1101 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 1110 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0101 1111 */ TABDECL(OP_B_COND),		// OP_B_COND

/* 11 0110 0000 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 0001 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 0010 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 0011 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 0100 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 0101 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 0110 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 0111 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 1000 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 1001 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 1010 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 1011 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 1100 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 1101 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 1110 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0110 1111 */ TABDECL(OP_B_COND),		// OP_B_COND

/* 11 0111 0000 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 0001 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 0010 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 0011 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 0100 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 0101 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 0110 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 0111 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 1000 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 1001 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 1010 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 1011 */ TABDECL(OP_B_COND),		// OP_B_COND
/* 11 0111 1100 */ TABDECL(OP_SWI_THUMB),		// OP_SWI_THUMB
/* 11 0111 1101 */ TABDECL(OP_SWI_THUMB),		// OP_SWI_THUMB
/* 11 0111 1110 */ TABDECL(OP_SWI_THUMB),		// OP_SWI_THUMB
/* 11 0111 1111 */ TABDECL(OP_SWI_THUMB),		// OP_SWI_THUMB

/* 11 1000 0000 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 0001 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 0010 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 0011 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 0100 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 0101 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 0110 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 0111 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 1000 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 1001 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 1010 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 1011 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 1100 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 1101 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 1110 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1000 1111 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND

/* 11 1001 0000 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 0001 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 0010 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 0011 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 0100 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 0101 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 0110 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 0111 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 1000 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 1001 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 1010 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 1011 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 1100 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 1101 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 1110 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND
/* 11 1001 1111 */ TABDECL(OP_B_UNCOND),		// OP_B_UNCOND

/* 11 1010 0000 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 0001 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 0010 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 0011 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 0100 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 0101 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 0110 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 0111 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 1000 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 1001 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 1010 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 1011 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 1100 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 1101 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 1110 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1010 1111 */ TABDECL(OP_BLX),		// OP_BLX

/* 11 1011 0000 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 0001 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 0010 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 0011 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 0100 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 0101 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 0110 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 0111 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 1000 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 1001 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 1010 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 1011 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 1100 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 1101 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 1110 */ TABDECL(OP_BLX),		// OP_BLX
/* 11 1011 1111 */ TABDECL(OP_BLX),		// OP_BLX

/* 11 1100 0000 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 0001 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 0010 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 0011 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 0100 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 0101 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 0110 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 0111 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 1000 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 1001 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 1010 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 1011 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 1100 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 1101 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 1110 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1100 1111 */ TABDECL(OP_BL_10),		// OP_BL_10

/* 11 1101 0000 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 0001 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 0010 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 0011 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 0100 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 0101 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 0110 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 0111 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 1000 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 1001 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 1010 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 1011 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 1100 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 1101 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 1110 */ TABDECL(OP_BL_10),		// OP_BL_10
/* 11 1101 1111 */ TABDECL(OP_BL_10),		// OP_BL_10

/* 11 1110 0000 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 0001 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 0010 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 0011 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 0100 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 0101 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 0110 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 0111 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 1000 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 1001 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 1010 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 1011 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 1100 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 1101 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 1110 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1110 1111 */ TABDECL(OP_BL_11),		// OP_BL_11

/* 11 1111 0000 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 0001 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 0010 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 0011 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 0100 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 0101 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 0110 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 0111 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 1000 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 1001 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 1010 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 1011 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 1100 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 1101 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 1110 */ TABDECL(OP_BL_11),		// OP_BL_11
/* 11 1111 1111 */ TABDECL(OP_BL_11),		// OP_BL_11

