/*
	Copyright (C) 2006 yopyop
	Copyright (C) 2007-2012 DeSmuME team
	This file is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 2 of the License, or
	(at your option) any later version.

	This file is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the this software.  If not, see <http://www.gnu.org/licenses/>.
*/

/* 0000 0000 0000 */ TABDECL(OP_AND_LSL_IMM),		// OP_AND_LSL_IMM
/* 0000 0000 0001 */ TABDECL(OP_AND_LSL_REG),		// OP_AND_LSL_REG
/* 0000 0000 0010 */ TABDECL(OP_AND_LSR_IMM),		// OP_AND_LSR_IMM
/* 0000 0000 0011 */ TABDECL(OP_AND_LSR_REG),		// OP_AND_LSR_REG
/* 0000 0000 0100 */ TABDECL(OP_AND_ASR_IMM),		// OP_AND_ASR_IMM
/* 0000 0000 0101 */ TABDECL(OP_AND_ASR_REG),		// OP_AND_ASR_REG
/* 0000 0000 0110 */ TABDECL(OP_AND_ROR_IMM),		// OP_AND_ROR_IMM
/* 0000 0000 0111 */ TABDECL(OP_AND_ROR_REG),		// OP_AND_ROR_REG
/* 0000 0000 1000 */ TABDECL(OP_AND_LSL_IMM),		// OP_AND_LSL_IMM
/* 0000 0000 1001 */ TABDECL(OP_MUL),		// OP_MUL
/* 0000 0000 1010 */ TABDECL(OP_AND_LSR_IMM),		// OP_AND_LSR_IMM
/* 0000 0000 1011 */ TABDECL(OP_STRH_POS_INDE_M_REG_OFF),		// OP_STRH_POS_INDE_M_REG_OFF
/* 0000 0000 1100 */ TABDECL(OP_AND_ASR_IMM),		// OP_AND_ASR_IMM
/* 0000 0000 1101 */ TABDECL(OP_LDRD_STRD_POST_INDEX),		// OP_LDRD_STRD_POST_INDEX
/* 0000 0000 1110 */ TABDECL(OP_AND_ROR_IMM),		// OP_AND_ROR_IMM
/* 0000 0000 1111 */ TABDECL(OP_LDRD_STRD_POST_INDEX),		// OP_LDRD_STRD_POST_INDEX

/* 0000 0001 0000 */ TABDECL(OP_AND_S_LSL_IMM),		// OP_AND_S_LSL_IMM
/* 0000 0001 0001 */ TABDECL(OP_AND_S_LSL_REG),		// OP_AND_S_LSL_REG
/* 0000 0001 0010 */ TABDECL(OP_AND_S_LSR_IMM),		// OP_AND_S_LSR_IMM
/* 0000 0001 0011 */ TABDECL(OP_AND_S_LSR_REG),		// OP_AND_S_LSR_REG
/* 0000 0001 0100 */ TABDECL(OP_AND_S_ASR_IMM),		// OP_AND_S_ASR_IMM
/* 0000 0001 0101 */ TABDECL(OP_AND_S_ASR_REG),		// OP_AND_S_ASR_REG
/* 0000 0001 0110 */ TABDECL(OP_AND_S_ROR_IMM),		// OP_AND_S_ROR_IMM
/* 0000 0001 0111 */ TABDECL(OP_AND_S_ROR_REG),		// OP_AND_S_ROR_REG
/* 0000 0001 1000 */ TABDECL(OP_AND_S_LSL_IMM),		// OP_AND_S_LSL_IMM
/* 0000 0001 1001 */ TABDECL(OP_MUL_S),		// OP_MUL_S
/* 0000 0001 1010 */ TABDECL(OP_AND_S_LSR_IMM),		// OP_AND_S_LSR_IMM
/* 0000 0001 1011 */ TABDECL(OP_LDRH_POS_INDE_M_REG_OFF),		// OP_LDRH_POS_INDE_M_REG_OFF
/* 0000 0001 1100 */ TABDECL(OP_AND_S_ASR_IMM),		// OP_AND_S_ASR_IMM
/* 0000 0001 1101 */ TABDECL(OP_LDRSB_POS_INDE_M_REG_OFF),		// OP_LDRSB_POS_INDE_M_REG_OFF
/* 0000 0001 1110 */ TABDECL(OP_AND_S_ROR_IMM),		// OP_AND_S_ROR_IMM
/* 0000 0001 1111 */ TABDECL(OP_LDRSH_POS_INDE_M_REG_OFF),		// OP_LDRSH_POS_INDE_M_REG_OFF

/* 0000 0010 0000 */ TABDECL(OP_EOR_LSL_IMM),		// OP_EOR_LSL_IMM
/* 0000 0010 0001 */ TABDECL(OP_EOR_LSL_REG),		// OP_EOR_LSL_REG
/* 0000 0010 0010 */ TABDECL(OP_EOR_LSR_IMM),		// OP_EOR_LSR_IMM
/* 0000 0010 0011 */ TABDECL(OP_EOR_LSR_REG),		// OP_EOR_LSR_REG
/* 0000 0010 0100 */ TABDECL(OP_EOR_ASR_IMM),		// OP_EOR_ASR_IMM
/* 0000 0010 0101 */ TABDECL(OP_EOR_ASR_REG),		// OP_EOR_ASR_REG
/* 0000 0010 0110 */ TABDECL(OP_EOR_ROR_IMM),		// OP_EOR_ROR_IMM
/* 0000 0010 0111 */ TABDECL(OP_EOR_ROR_REG),		// OP_EOR_ROR_REG
/* 0000 0010 1000 */ TABDECL(OP_EOR_LSL_IMM),		// OP_EOR_LSL_IMM
/* 0000 0010 1001 */ TABDECL(OP_MLA),		// OP_MLA
/* 0000 0010 1010 */ TABDECL(OP_EOR_LSR_IMM),		// OP_EOR_LSR_IMM
/* 0000 0010 1011 */ TABDECL(OP_UND),
/* 0000 0010 1100 */ TABDECL(OP_EOR_ASR_IMM),		// OP_EOR_ASR_IMM
/* 0000 0010 1101 */ TABDECL(OP_UND),
/* 0000 0010 1110 */ TABDECL(OP_EOR_ROR_IMM),		// OP_EOR_ROR_IMM
/* 0000 0010 1111 */ TABDECL(OP_UND),

/* 0000 0011 0000 */ TABDECL(OP_EOR_S_LSL_IMM),		// OP_EOR_S_LSL_IMM
/* 0000 0011 0001 */ TABDECL(OP_EOR_S_LSL_REG),		// OP_EOR_S_LSL_REG
/* 0000 0011 0010 */ TABDECL(OP_EOR_S_LSR_IMM),		// OP_EOR_S_LSR_IMM
/* 0000 0011 0011 */ TABDECL(OP_EOR_S_LSR_REG),		// OP_EOR_S_LSR_REG
/* 0000 0011 0100 */ TABDECL(OP_EOR_S_ASR_IMM),		// OP_EOR_S_ASR_IMM
/* 0000 0011 0101 */ TABDECL(OP_EOR_S_ASR_REG),		// OP_EOR_S_ASR_REG
/* 0000 0011 0110 */ TABDECL(OP_EOR_S_ROR_IMM),		// OP_EOR_S_ROR_IMM
/* 0000 0011 0111 */ TABDECL(OP_EOR_S_ROR_REG),		// OP_EOR_S_ROR_REG
/* 0000 0011 1000 */ TABDECL(OP_EOR_S_LSL_IMM),		// OP_EOR_S_LSL_IMM
/* 0000 0011 1001 */ TABDECL(OP_MLA_S),		// OP_MLA_S
/* 0000 0011 1010 */ TABDECL(OP_EOR_S_LSR_IMM),		// OP_EOR_S_LSR_IMM
/* 0000 0011 1011 */ TABDECL(OP_UND),
/* 0000 0011 1100 */ TABDECL(OP_EOR_S_ASR_IMM),		// OP_EOR_S_ASR_IMM
/* 0000 0011 1101 */ TABDECL(OP_UND),
/* 0000 0011 1110 */ TABDECL(OP_EOR_S_ROR_IMM),		// OP_EOR_S_ROR_IMM
/* 0000 0011 1111 */ TABDECL(OP_UND),

/* 0000 0100 0000 */ TABDECL(OP_SUB_LSL_IMM),		// OP_SUB_LSL_IMM
/* 0000 0100 0001 */ TABDECL(OP_SUB_LSL_REG),		// OP_SUB_LSL_REG
/* 0000 0100 0010 */ TABDECL(OP_SUB_LSR_IMM),		// OP_SUB_LSR_IMM
/* 0000 0100 0011 */ TABDECL(OP_SUB_LSR_REG),		// OP_SUB_LSR_REG
/* 0000 0100 0100 */ TABDECL(OP_SUB_ASR_IMM),		// OP_SUB_ASR_IMM
/* 0000 0100 0101 */ TABDECL(OP_SUB_ASR_REG),		// OP_SUB_ASR_REG
/* 0000 0100 0110 */ TABDECL(OP_SUB_ROR_IMM),		// OP_SUB_ROR_IMM
/* 0000 0100 0111 */ TABDECL(OP_SUB_ROR_REG),		// OP_SUB_ROR_REG
/* 0000 0100 1000 */ TABDECL(OP_SUB_LSL_IMM),		// OP_SUB_LSL_IMM
/* 0000 0100 1001 */ TABDECL(OP_UND),
/* 0000 0100 1010 */ TABDECL(OP_SUB_LSR_IMM),		// OP_SUB_LSR_IMM
/* 0000 0100 1011 */ TABDECL(OP_STRH_POS_INDE_M_IMM_OFF),		// OP_STRH_POS_INDE_M_IMM_OFF
/* 0000 0100 1100 */ TABDECL(OP_SUB_ASR_IMM),		// OP_SUB_ASR_IMM
/* 0000 0100 1101 */ TABDECL(OP_LDRD_STRD_POST_INDEX),		// OP_LDRD_STRD_POST_INDEX
/* 0000 0100 1110 */ TABDECL(OP_SUB_ROR_IMM),		// OP_SUB_ROR_IMM
/* 0000 0100 1111 */ TABDECL(OP_LDRD_STRD_POST_INDEX),		// OP_LDRD_STRD_POST_INDEX

/* 0000 0101 0000 */ TABDECL(OP_SUB_S_LSL_IMM),		// OP_SUB_S_LSL_IMM
/* 0000 0101 0001 */ TABDECL(OP_SUB_S_LSL_REG),		// OP_SUB_S_LSL_REG
/* 0000 0101 0010 */ TABDECL(OP_SUB_S_LSR_IMM),		// OP_SUB_S_LSR_IMM
/* 0000 0101 0011 */ TABDECL(OP_SUB_S_LSR_REG),		// OP_SUB_S_LSR_REG
/* 0000 0101 0100 */ TABDECL(OP_SUB_S_ASR_IMM),		// OP_SUB_S_ASR_IMM
/* 0000 0101 0101 */ TABDECL(OP_SUB_S_ASR_REG),		// OP_SUB_S_ASR_REG
/* 0000 0101 0110 */ TABDECL(OP_SUB_S_ROR_IMM),		// OP_SUB_S_ROR_IMM
/* 0000 0101 0111 */ TABDECL(OP_SUB_S_ROR_REG),		// OP_SUB_S_ROR_REG
/* 0000 0101 1000 */ TABDECL(OP_SUB_S_LSL_IMM),		// OP_SUB_S_LSL_IMM
/* 0000 0101 1001 */ TABDECL(OP_UND),
/* 0000 0101 1010 */ TABDECL(OP_SUB_S_LSR_IMM),		// OP_SUB_S_LSR_IMM
/* 0000 0101 1011 */ TABDECL(OP_LDRH_POS_INDE_M_IMM_OFF),		// OP_LDRH_POS_INDE_M_IMM_OFF
/* 0000 0101 1100 */ TABDECL(OP_SUB_S_ASR_IMM),		// OP_SUB_S_ASR_IMM
/* 0000 0101 1101 */ TABDECL(OP_LDRSB_POS_INDE_M_IMM_OFF),		// OP_LDRSB_POS_INDE_M_IMM_OFF
/* 0000 0101 1110 */ TABDECL(OP_SUB_S_ROR_IMM),		// OP_SUB_S_ROR_IMM
/* 0000 0101 1111 */ TABDECL(OP_LDRSH_POS_INDE_M_IMM_OFF),		// OP_LDRSH_POS_INDE_M_IMM_OFF

/* 0000 0110 0000 */ TABDECL(OP_RSB_LSL_IMM),		// OP_RSB_LSL_IMM
/* 0000 0110 0001 */ TABDECL(OP_RSB_LSL_REG),		// OP_RSB_LSL_REG
/* 0000 0110 0010 */ TABDECL(OP_RSB_LSR_IMM),		// OP_RSB_LSR_IMM
/* 0000 0110 0011 */ TABDECL(OP_RSB_LSR_REG),		// OP_RSB_LSR_REG
/* 0000 0110 0100 */ TABDECL(OP_RSB_ASR_IMM),		// OP_RSB_ASR_IMM
/* 0000 0110 0101 */ TABDECL(OP_RSB_ASR_REG),		// OP_RSB_ASR_REG
/* 0000 0110 0110 */ TABDECL(OP_RSB_ROR_IMM),		// OP_RSB_ROR_IMM
/* 0000 0110 0111 */ TABDECL(OP_RSB_ROR_REG),		// OP_RSB_ROR_REG
/* 0000 0110 1000 */ TABDECL(OP_RSB_LSL_IMM),		// OP_RSB_LSL_IMM
/* 0000 0110 1001 */ TABDECL(OP_UND),
/* 0000 0110 1010 */ TABDECL(OP_RSB_LSR_IMM),		// OP_RSB_LSR_IMM
/* 0000 0110 1011 */ TABDECL(OP_UND),
/* 0000 0110 1100 */ TABDECL(OP_RSB_ASR_IMM),		// OP_RSB_ASR_IMM
/* 0000 0110 1101 */ TABDECL(OP_UND),
/* 0000 0110 1110 */ TABDECL(OP_RSB_ROR_IMM),		// OP_RSB_ROR_IMM
/* 0000 0110 1111 */ TABDECL(OP_UND),

/* 0000 0111 0000 */ TABDECL(OP_RSB_S_LSL_IMM),		// OP_RSB_S_LSL_IMM
/* 0000 0111 0001 */ TABDECL(OP_RSB_S_LSL_REG),		// OP_RSB_S_LSL_REG
/* 0000 0111 0010 */ TABDECL(OP_RSB_S_LSR_IMM),		// OP_RSB_S_LSR_IMM
/* 0000 0111 0011 */ TABDECL(OP_RSB_S_LSR_REG),		// OP_RSB_S_LSR_REG
/* 0000 0111 0100 */ TABDECL(OP_RSB_S_ASR_IMM),		// OP_RSB_S_ASR_IMM
/* 0000 0111 0101 */ TABDECL(OP_RSB_S_ASR_REG),		// OP_RSB_S_ASR_REG
/* 0000 0111 0110 */ TABDECL(OP_RSB_S_ROR_IMM),		// OP_RSB_S_ROR_IMM
/* 0000 0111 0111 */ TABDECL(OP_RSB_S_ROR_REG),		// OP_RSB_S_ROR_REG
/* 0000 0111 1000 */ TABDECL(OP_RSB_S_LSL_IMM),		// OP_RSB_S_LSL_IMM
/* 0000 0111 1001 */ TABDECL(OP_UND),
/* 0000 0111 1010 */ TABDECL(OP_RSB_S_LSR_IMM),		// OP_RSB_S_LSR_IMM
/* 0000 0111 1011 */ TABDECL(OP_UND),
/* 0000 0111 1100 */ TABDECL(OP_RSB_S_ASR_IMM),		// OP_RSB_S_ASR_IMM
/* 0000 0111 1101 */ TABDECL(OP_UND),
/* 0000 0111 1110 */ TABDECL(OP_RSB_S_ROR_IMM),		// OP_RSB_S_ROR_IMM
/* 0000 0111 1111 */ TABDECL(OP_UND),

/* 0000 1000 0000 */ TABDECL(OP_ADD_LSL_IMM),		// OP_ADD_LSL_IMM
/* 0000 1000 0001 */ TABDECL(OP_ADD_LSL_REG),		// OP_ADD_LSL_REG
/* 0000 1000 0010 */ TABDECL(OP_ADD_LSR_IMM),		// OP_ADD_LSR_IMM
/* 0000 1000 0011 */ TABDECL(OP_ADD_LSR_REG),		// OP_ADD_LSR_REG
/* 0000 1000 0100 */ TABDECL(OP_ADD_ASR_IMM),		// OP_ADD_ASR_IMM
/* 0000 1000 0101 */ TABDECL(OP_ADD_ASR_REG),		// OP_ADD_ASR_REG
/* 0000 1000 0110 */ TABDECL(OP_ADD_ROR_IMM),		// OP_ADD_ROR_IMM
/* 0000 1000 0111 */ TABDECL(OP_ADD_ROR_REG),		// OP_ADD_ROR_REG
/* 0000 1000 1000 */ TABDECL(OP_ADD_LSL_IMM),		// OP_ADD_LSL_IMM
/* 0000 1000 1001 */ TABDECL(OP_UMULL),		// OP_UMULL
/* 0000 1000 1010 */ TABDECL(OP_ADD_LSR_IMM),		// OP_ADD_LSR_IMM
/* 0000 1000 1011 */ TABDECL(OP_STRH_POS_INDE_P_REG_OFF),		// OP_STRH_POS_INDE_P_REG_OFF
/* 0000 1000 1100 */ TABDECL(OP_ADD_ASR_IMM),		// OP_ADD_ASR_IMM
/* 0000 1000 1101 */ TABDECL(OP_LDRD_STRD_POST_INDEX),		// OP_LDRD_STRD_POST_INDEX
/* 0000 1000 1110 */ TABDECL(OP_ADD_ROR_IMM),		// OP_ADD_ROR_IMM
/* 0000 1000 1111 */ TABDECL(OP_LDRD_STRD_POST_INDEX),		// OP_LDRD_STRD_POST_INDEX

/* 0000 1001 0000 */ TABDECL(OP_ADD_S_LSL_IMM),		// OP_ADD_S_LSL_IMM
/* 0000 1001 0001 */ TABDECL(OP_ADD_S_LSL_REG),		// OP_ADD_S_LSL_REG
/* 0000 1001 0010 */ TABDECL(OP_ADD_S_LSR_IMM),		// OP_ADD_S_LSR_IMM
/* 0000 1001 0011 */ TABDECL(OP_ADD_S_LSR_REG),		// OP_ADD_S_LSR_REG
/* 0000 1001 0100 */ TABDECL(OP_ADD_S_ASR_IMM),		// OP_ADD_S_ASR_IMM
/* 0000 1001 0101 */ TABDECL(OP_ADD_S_ASR_REG),		// OP_ADD_S_ASR_REG
/* 0000 1001 0110 */ TABDECL(OP_ADD_S_ROR_IMM),		// OP_ADD_S_ROR_IMM
/* 0000 1001 0111 */ TABDECL(OP_ADD_S_ROR_REG),		// OP_ADD_S_ROR_REG
/* 0000 1001 1000 */ TABDECL(OP_ADD_S_LSL_IMM),		// OP_ADD_S_LSL_IMM
/* 0000 1001 1001 */ TABDECL(OP_UMULL_S),		// OP_UMULL_S
/* 0000 1001 1010 */ TABDECL(OP_ADD_S_LSR_IMM),		// OP_ADD_S_LSR_IMM
/* 0000 1001 1011 */ TABDECL(OP_LDRH_POS_INDE_P_REG_OFF),		// OP_LDRH_POS_INDE_P_REG_OFF
/* 0000 1001 1100 */ TABDECL(OP_ADD_S_ASR_IMM),		// OP_ADD_S_ASR_IMM
/* 0000 1001 1101 */ TABDECL(OP_LDRSB_POS_INDE_P_REG_OFF),		// OP_LDRSB_POS_INDE_P_REG_OFF
/* 0000 1001 1110 */ TABDECL(OP_ADD_S_ROR_IMM),		// OP_ADD_S_ROR_IMM
/* 0000 1001 1111 */ TABDECL(OP_LDRSH_POS_INDE_P_REG_OFF),		// OP_LDRSH_POS_INDE_P_REG_OFF

/* 0000 1010 0000 */ TABDECL(OP_ADC_LSL_IMM),		// OP_ADC_LSL_IMM
/* 0000 1010 0001 */ TABDECL(OP_ADC_LSL_REG),		// OP_ADC_LSL_REG
/* 0000 1010 0010 */ TABDECL(OP_ADC_LSR_IMM),		// OP_ADC_LSR_IMM
/* 0000 1010 0011 */ TABDECL(OP_ADC_LSR_REG),		// OP_ADC_LSR_REG
/* 0000 1010 0100 */ TABDECL(OP_ADC_ASR_IMM),		// OP_ADC_ASR_IMM
/* 0000 1010 0101 */ TABDECL(OP_ADC_ASR_REG),		// OP_ADC_ASR_REG
/* 0000 1010 0110 */ TABDECL(OP_ADC_ROR_IMM),		// OP_ADC_ROR_IMM
/* 0000 1010 0111 */ TABDECL(OP_ADC_ROR_REG),		// OP_ADC_ROR_REG
/* 0000 1010 1000 */ TABDECL(OP_ADC_LSL_IMM),		// OP_ADC_LSL_IMM
/* 0000 1010 1001 */ TABDECL(OP_UMLAL),		// OP_UMLAL
/* 0000 1010 1010 */ TABDECL(OP_ADC_LSR_IMM),		// OP_ADC_LSR_IMM
/* 0000 1010 1011 */ TABDECL(OP_UND),
/* 0000 1010 1100 */ TABDECL(OP_ADC_ASR_IMM),		// OP_ADC_ASR_IMM
/* 0000 1010 1101 */ TABDECL(OP_UND),
/* 0000 1010 1110 */ TABDECL(OP_ADC_ROR_IMM),		// OP_ADC_ROR_IMM
/* 0000 1010 1111 */ TABDECL(OP_UND),

/* 0000 1011 0000 */ TABDECL(OP_ADC_S_LSL_IMM),		// OP_ADC_S_LSL_IMM
/* 0000 1011 0001 */ TABDECL(OP_ADC_S_LSL_REG),		// OP_ADC_S_LSL_REG
/* 0000 1011 0010 */ TABDECL(OP_ADC_S_LSR_IMM),		// OP_ADC_S_LSR_IMM
/* 0000 1011 0011 */ TABDECL(OP_ADC_S_LSR_REG),		// OP_ADC_S_LSR_REG
/* 0000 1011 0100 */ TABDECL(OP_ADC_S_ASR_IMM),		// OP_ADC_S_ASR_IMM
/* 0000 1011 0101 */ TABDECL(OP_ADC_S_ASR_REG),		// OP_ADC_S_ASR_REG
/* 0000 1011 0110 */ TABDECL(OP_ADC_S_ROR_IMM),		// OP_ADC_S_ROR_IMM
/* 0000 1011 0111 */ TABDECL(OP_ADC_S_ROR_REG),		// OP_ADC_S_ROR_REG
/* 0000 1011 1000 */ TABDECL(OP_ADC_S_LSL_IMM),		// OP_ADC_S_LSL_IMM
/* 0000 1011 1001 */ TABDECL(OP_UMLAL_S),		// OP_UMLAL_S
/* 0000 1011 1010 */ TABDECL(OP_ADC_S_LSR_IMM),		// OP_ADC_S_LSR_IMM
/* 0000 1011 1011 */ TABDECL(OP_UND),
/* 0000 1011 1100 */ TABDECL(OP_ADC_S_ASR_IMM),		// OP_ADC_S_ASR_IMM
/* 0000 1011 1101 */ TABDECL(OP_UND),
/* 0000 1011 1110 */ TABDECL(OP_ADC_S_ROR_IMM),		// OP_ADC_S_ROR_IMM
/* 0000 1011 1111 */ TABDECL(OP_UND),

/* 0000 1100 0000 */ TABDECL(OP_SBC_LSL_IMM),		// OP_SBC_LSL_IMM
/* 0000 1100 0001 */ TABDECL(OP_SBC_LSL_REG),		// OP_SBC_LSL_REG
/* 0000 1100 0010 */ TABDECL(OP_SBC_LSR_IMM),		// OP_SBC_LSR_IMM
/* 0000 1100 0011 */ TABDECL(OP_SBC_LSR_REG),		// OP_SBC_LSR_REG
/* 0000 1100 0100 */ TABDECL(OP_SBC_ASR_IMM),		// OP_SBC_ASR_IMM
/* 0000 1100 0101 */ TABDECL(OP_SBC_ASR_REG),		// OP_SBC_ASR_REG
/* 0000 1100 0110 */ TABDECL(OP_SBC_ROR_IMM),		// OP_SBC_ROR_IMM
/* 0000 1100 0111 */ TABDECL(OP_SBC_ROR_REG),		// OP_SBC_ROR_REG
/* 0000 1100 1000 */ TABDECL(OP_SBC_LSL_IMM),		// OP_SBC_LSL_IMM
/* 0000 1100 1001 */ TABDECL(OP_SMULL),		// OP_SMULL
/* 0000 1100 1010 */ TABDECL(OP_SBC_LSR_IMM),		// OP_SBC_LSR_IMM
/* 0000 1100 1011 */ TABDECL(OP_STRH_POS_INDE_P_IMM_OFF),		// OP_STRH_POS_INDE_P_IMM_OFF
/* 0000 1100 1100 */ TABDECL(OP_SBC_ASR_IMM),		// OP_SBC_ASR_IMM
/* 0000 1100 1101 */ TABDECL(OP_LDRD_STRD_POST_INDEX),		// OP_LDRD_STRD_POST_INDEX
/* 0000 1100 1110 */ TABDECL(OP_SBC_ROR_IMM),		// OP_SBC_ROR_IMM
/* 0000 1100 1111 */ TABDECL(OP_LDRD_STRD_POST_INDEX),		// OP_LDRD_STRD_POST_INDEX

/* 0000 1101 0000 */ TABDECL(OP_SBC_S_LSL_IMM),		// OP_SBC_S_LSL_IMM
/* 0000 1101 0001 */ TABDECL(OP_SBC_S_LSL_REG),		// OP_SBC_S_LSL_REG
/* 0000 1101 0010 */ TABDECL(OP_SBC_S_LSR_IMM),		// OP_SBC_S_LSR_IMM
/* 0000 1101 0011 */ TABDECL(OP_SBC_S_LSR_REG),		// OP_SBC_S_LSR_REG
/* 0000 1101 0100 */ TABDECL(OP_SBC_S_ASR_IMM),		// OP_SBC_S_ASR_IMM
/* 0000 1101 0101 */ TABDECL(OP_SBC_S_ASR_REG),		// OP_SBC_S_ASR_REG
/* 0000 1101 0110 */ TABDECL(OP_SBC_S_ROR_IMM),		// OP_SBC_S_ROR_IMM
/* 0000 1101 0111 */ TABDECL(OP_SBC_S_ROR_REG),		// OP_SBC_S_ROR_REG
/* 0000 1101 1000 */ TABDECL(OP_SBC_S_LSL_IMM),		// OP_SBC_S_LSL_IMM
/* 0000 1101 1001 */ TABDECL(OP_SMULL_S),		// OP_SMULL_S
/* 0000 1101 1010 */ TABDECL(OP_SBC_S_LSR_IMM),		// OP_SBC_S_LSR_IMM
/* 0000 1101 1011 */ TABDECL(OP_LDRH_POS_INDE_P_IMM_OFF),		// OP_LDRH_POS_INDE_P_IMM_OFF
/* 0000 1101 1100 */ TABDECL(OP_SBC_S_ASR_IMM),		// OP_SBC_S_ASR_IMM
/* 0000 1101 1101 */ TABDECL(OP_LDRSB_POS_INDE_P_IMM_OFF),		// OP_LDRSB_POS_INDE_P_IMM_OFF
/* 0000 1101 1110 */ TABDECL(OP_SBC_S_ROR_IMM),		// OP_SBC_S_ROR_IMM
/* 0000 1101 1111 */ TABDECL(OP_LDRSH_POS_INDE_P_IMM_OFF),		// OP_LDRSH_POS_INDE_P_IMM_OFF

/* 0000 1110 0000 */ TABDECL(OP_RSC_LSL_IMM),		// OP_RSC_LSL_IMM
/* 0000 1110 0001 */ TABDECL(OP_RSC_LSL_REG),		// OP_RSC_LSL_REG
/* 0000 1110 0010 */ TABDECL(OP_RSC_LSR_IMM),		// OP_RSC_LSR_IMM
/* 0000 1110 0011 */ TABDECL(OP_RSC_LSR_REG),		// OP_RSC_LSR_REG
/* 0000 1110 0100 */ TABDECL(OP_RSC_ASR_IMM),		// OP_RSC_ASR_IMM
/* 0000 1110 0101 */ TABDECL(OP_RSC_ASR_REG),		// OP_RSC_ASR_REG
/* 0000 1110 0110 */ TABDECL(OP_RSC_ROR_IMM),		// OP_RSC_ROR_IMM
/* 0000 1110 0111 */ TABDECL(OP_RSC_ROR_REG),		// OP_RSC_ROR_REG
/* 0000 1110 1000 */ TABDECL(OP_RSC_LSL_IMM),		// OP_RSC_LSL_IMM
/* 0000 1110 1001 */ TABDECL(OP_SMLAL),		// OP_SMLAL
/* 0000 1110 1010 */ TABDECL(OP_RSC_LSR_IMM),		// OP_RSC_LSR_IMM
/* 0000 1110 1011 */ TABDECL(OP_UND),
/* 0000 1110 1100 */ TABDECL(OP_RSC_ASR_IMM),		// OP_RSC_ASR_IMM
/* 0000 1110 1101 */ TABDECL(OP_UND),
/* 0000 1110 1110 */ TABDECL(OP_RSC_ROR_IMM),		// OP_RSC_ROR_IMM
/* 0000 1110 1111 */ TABDECL(OP_UND),

/* 0000 1111 0000 */ TABDECL(OP_RSC_S_LSL_IMM),		// OP_RSC_S_LSL_IMM
/* 0000 1111 0001 */ TABDECL(OP_RSC_S_LSL_REG),		// OP_RSC_S_LSL_REG
/* 0000 1111 0010 */ TABDECL(OP_RSC_S_LSR_IMM),		// OP_RSC_S_LSR_IMM
/* 0000 1111 0011 */ TABDECL(OP_RSC_S_LSR_REG),		// OP_RSC_S_LSR_REG
/* 0000 1111 0100 */ TABDECL(OP_RSC_S_ASR_IMM),		// OP_RSC_S_ASR_IMM
/* 0000 1111 0101 */ TABDECL(OP_RSC_S_ASR_REG),		// OP_RSC_S_ASR_REG
/* 0000 1111 0110 */ TABDECL(OP_RSC_S_ROR_IMM),		// OP_RSC_S_ROR_IMM
/* 0000 1111 0111 */ TABDECL(OP_RSC_S_ROR_REG),		// OP_RSC_S_ROR_REG
/* 0000 1111 1000 */ TABDECL(OP_RSC_S_LSL_IMM),		// OP_RSC_S_LSL_IMM
/* 0000 1111 1001 */ TABDECL(OP_SMLAL_S),		// OP_SMLAL_S
/* 0000 1111 1010 */ TABDECL(OP_RSC_S_LSR_IMM),		// OP_RSC_S_LSR_IMM
/* 0000 1111 1011 */ TABDECL(OP_UND),
/* 0000 1111 1100 */ TABDECL(OP_RSC_S_ASR_IMM),		// OP_RSC_S_ASR_IMM
/* 0000 1111 1101 */ TABDECL(OP_UND),
/* 0000 1111 1110 */ TABDECL(OP_RSC_S_ROR_IMM),		// OP_RSC_S_ROR_IMM
/* 0000 1111 1111 */ TABDECL(OP_UND),

/* 0001 0000 0000 */ TABDECL(OP_MRS_CPSR),		// OP_MRS_CPSR
/* 0001 0000 0001 */ TABDECL(OP_UND),
/* 0001 0000 0010 */ TABDECL(OP_UND),
/* 0001 0000 0011 */ TABDECL(OP_UND),
/* 0001 0000 0100 */ TABDECL(OP_UND),
/* 0001 0000 0101 */ TABDECL(OP_QADD),		// OP_QADD
/* 0001 0000 0110 */ TABDECL(OP_UND),
/* 0001 0000 0111 */ TABDECL(OP_UND),
/* 0001 0000 1000 */ TABDECL(OP_SMLA_B_B),		// OP_SMLA_B_B
/* 0001 0000 1001 */ TABDECL(OP_SWP),		// OP_SWP
/* 0001 0000 1010 */ TABDECL(OP_SMLA_T_B),		// OP_SMLA_T_B
/* 0001 0000 1011 */ TABDECL(OP_STRH_M_REG_OFF),		// OP_STRH_M_REG_OFF
/* 0001 0000 1100 */ TABDECL(OP_SMLA_B_T),		// OP_SMLA_B_T
/* 0001 0000 1101 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX
/* 0001 0000 1110 */ TABDECL(OP_SMLA_T_T),		// OP_SMLA_T_T
/* 0001 0000 1111 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX

/* 0001 0001 0000 */ TABDECL(OP_TST_LSL_IMM),		// OP_TST_LSL_IMM
/* 0001 0001 0001 */ TABDECL(OP_TST_LSL_REG),		// OP_TST_LSL_REG
/* 0001 0001 0010 */ TABDECL(OP_TST_LSR_IMM),		// OP_TST_LSR_IMM
/* 0001 0001 0011 */ TABDECL(OP_TST_LSR_REG),		// OP_TST_LSR_REG
/* 0001 0001 0100 */ TABDECL(OP_TST_ASR_IMM),		// OP_TST_ASR_IMM
/* 0001 0001 0101 */ TABDECL(OP_TST_ASR_REG),		// OP_TST_ASR_REG
/* 0001 0001 0110 */ TABDECL(OP_TST_ROR_IMM),		// OP_TST_ROR_IMM
/* 0001 0001 0111 */ TABDECL(OP_TST_ROR_REG),		// OP_TST_ROR_REG
/* 0001 0001 1000 */ TABDECL(OP_TST_LSL_IMM),		// OP_TST_LSL_IMM
/* 0001 0001 1001 */ TABDECL(OP_UND),
/* 0001 0001 1010 */ TABDECL(OP_TST_LSR_IMM),		// OP_TST_LSR_IMM
/* 0001 0001 1011 */ TABDECL(OP_LDRH_M_REG_OFF),		// OP_LDRH_M_REG_OFF
/* 0001 0001 1100 */ TABDECL(OP_TST_ASR_IMM),		// OP_TST_ASR_IMM
/* 0001 0001 1101 */ TABDECL(OP_LDRSB_M_REG_OFF),		// OP_LDRSB_M_REG_OFF
/* 0001 0001 1110 */ TABDECL(OP_TST_ROR_IMM),		// OP_TST_ROR_IMM
/* 0001 0001 1111 */ TABDECL(OP_LDRSH_M_REG_OFF),		// OP_LDRSH_M_REG_OFF

/* 0001 0010 0000 */ TABDECL(OP_MSR_CPSR),		// OP_MSR_CPSR
/* 0001 0010 0001 */ TABDECL(OP_BX),		// OP_BX
/* 0001 0010 0010 */ TABDECL(OP_UND),
/* 0001 0010 0011 */ TABDECL(OP_BLX_REG),		// OP_BLX_REG
/* 0001 0010 0100 */ TABDECL(OP_UND),
/* 0001 0010 0101 */ TABDECL(OP_QSUB),		// OP_QSUB
/* 0001 0010 0110 */ TABDECL(OP_UND),
/* 0001 0010 0111 */ TABDECL(OP_BKPT),		// OP_BKPT
/* 0001 0010 1000 */ TABDECL(OP_SMLAW_B),		// OP_SMLAW_B
/* 0001 0010 1001 */ TABDECL(OP_UND),
/* 0001 0010 1010 */ TABDECL(OP_SMULW_B),		// OP_SMULW_B
/* 0001 0010 1011 */ TABDECL(OP_STRH_PRE_INDE_M_REG_OFF),		// OP_STRH_PRE_INDE_M_REG_OFF
/* 0001 0010 1100 */ TABDECL(OP_SMLAW_T),		// OP_SMLAW_T
/* 0001 0010 1101 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX
/* 0001 0010 1110 */ TABDECL(OP_SMULW_T),		// OP_SMULW_T
/* 0001 0010 1111 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX

/* 0001 0011 0000 */ TABDECL(OP_TEQ_LSL_IMM),		// OP_TEQ_LSL_IMM
/* 0001 0011 0001 */ TABDECL(OP_TEQ_LSL_REG),		// OP_TEQ_LSL_REG
/* 0001 0011 0010 */ TABDECL(OP_TEQ_LSR_IMM),		// OP_TEQ_LSR_IMM
/* 0001 0011 0011 */ TABDECL(OP_TEQ_LSR_REG),		// OP_TEQ_LSR_REG
/* 0001 0011 0100 */ TABDECL(OP_TEQ_ASR_IMM),		// OP_TEQ_ASR_IMM
/* 0001 0011 0101 */ TABDECL(OP_TEQ_ASR_REG),		// OP_TEQ_ASR_REG
/* 0001 0011 0110 */ TABDECL(OP_TEQ_ROR_IMM),		// OP_TEQ_ROR_IMM
/* 0001 0011 0111 */ TABDECL(OP_TEQ_ROR_REG),		// OP_TEQ_ROR_REG
/* 0001 0011 1000 */ TABDECL(OP_TEQ_LSL_IMM),		// OP_TEQ_LSL_IMM
/* 0001 0011 1001 */ TABDECL(OP_UND),
/* 0001 0011 1010 */ TABDECL(OP_TEQ_LSR_IMM),		// OP_TEQ_LSR_IMM
/* 0001 0011 1011 */ TABDECL(OP_LDRH_PRE_INDE_M_REG_OFF),		// OP_LDRH_PRE_INDE_M_REG_OFF
/* 0001 0011 1100 */ TABDECL(OP_TEQ_ASR_IMM),		// OP_TEQ_ASR_IMM
/* 0001 0011 1101 */ TABDECL(OP_LDRSB_PRE_INDE_M_REG_OFF),		// OP_LDRSB_PRE_INDE_M_REG_OFF
/* 0001 0011 1110 */ TABDECL(OP_TEQ_ROR_IMM),		// OP_TEQ_ROR_IMM
/* 0001 0011 1111 */ TABDECL(OP_LDRSH_PRE_INDE_M_REG_OFF),		// OP_LDRSH_PRE_INDE_M_REG_OFF

/* 0001 0100 0000 */ TABDECL(OP_MRS_SPSR),		// OP_MRS_SPSR
/* 0001 0100 0001 */ TABDECL(OP_UND),
/* 0001 0100 0010 */ TABDECL(OP_UND),
/* 0001 0100 0011 */ TABDECL(OP_UND),
/* 0001 0100 0100 */ TABDECL(OP_UND),
/* 0001 0100 0101 */ TABDECL(OP_QDADD),		// OP_QDADD
/* 0001 0100 0110 */ TABDECL(OP_UND),
/* 0001 0100 0111 */ TABDECL(OP_UND),
/* 0001 0100 1000 */ TABDECL(OP_SMLAL_B_B),		// OP_SMLAL_B_B
/* 0001 0100 1001 */ TABDECL(OP_SWPB),		// OP_SWPB
/* 0001 0100 1010 */ TABDECL(OP_SMLAL_T_B),		// OP_SMLAL_T_B
/* 0001 0100 1011 */ TABDECL(OP_STRH_M_IMM_OFF),		// OP_STRH_M_IMM_OFF
/* 0001 0100 1100 */ TABDECL(OP_SMLAL_B_T),		// OP_SMLAL_B_T
/* 0001 0100 1101 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX
/* 0001 0100 1110 */ TABDECL(OP_SMLAL_T_T),		// OP_SMLAL_T_T
/* 0001 0100 1111 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX

/* 0001 0101 0000 */ TABDECL(OP_CMP_LSL_IMM),		// OP_CMP_LSL_IMM
/* 0001 0101 0001 */ TABDECL(OP_CMP_LSL_REG),		// OP_CMP_LSL_REG
/* 0001 0101 0010 */ TABDECL(OP_CMP_LSR_IMM),		// OP_CMP_LSR_IMM
/* 0001 0101 0011 */ TABDECL(OP_CMP_LSR_REG),		// OP_CMP_LSR_REG
/* 0001 0101 0100 */ TABDECL(OP_CMP_ASR_IMM),		// OP_CMP_ASR_IMM
/* 0001 0101 0101 */ TABDECL(OP_CMP_ASR_REG),		// OP_CMP_ASR_REG
/* 0001 0101 0110 */ TABDECL(OP_CMP_ROR_IMM),		// OP_CMP_ROR_IMM
/* 0001 0101 0111 */ TABDECL(OP_CMP_ROR_REG),		// OP_CMP_ROR_REG
/* 0001 0101 1000 */ TABDECL(OP_CMP_LSL_IMM),		// OP_CMP_LSL_IMM
/* 0001 0101 1001 */ TABDECL(OP_UND),
/* 0001 0101 1010 */ TABDECL(OP_CMP_LSR_IMM),		// OP_CMP_LSR_IMM
/* 0001 0101 1011 */ TABDECL(OP_LDRH_M_IMM_OFF),		// OP_LDRH_M_IMM_OFF
/* 0001 0101 1100 */ TABDECL(OP_CMP_ASR_IMM),		// OP_CMP_ASR_IMM
/* 0001 0101 1101 */ TABDECL(OP_LDRSB_M_IMM_OFF),		// OP_LDRSB_M_IMM_OFF
/* 0001 0101 1110 */ TABDECL(OP_CMP_ROR_IMM),		// OP_CMP_ROR_IMM
/* 0001 0101 1111 */ TABDECL(OP_LDRSH_M_IMM_OFF),		// OP_LDRSH_M_IMM_OFF

/* 0001 0110 0000 */ TABDECL(OP_MSR_SPSR),		// OP_MSR_SPSR
/* 0001 0110 0001 */ TABDECL(OP_CLZ),		// OP_CLZ
/* 0001 0110 0010 */ TABDECL(OP_UND),
/* 0001 0110 0011 */ TABDECL(OP_UND),
/* 0001 0110 0100 */ TABDECL(OP_UND),
/* 0001 0110 0101 */ TABDECL(OP_QDSUB),		// OP_QDSUB
/* 0001 0110 0110 */ TABDECL(OP_UND),
/* 0001 0110 0111 */ TABDECL(OP_UND),
/* 0001 0110 1000 */ TABDECL(OP_SMUL_B_B),		// OP_SMUL_B_B
/* 0001 0110 1001 */ TABDECL(OP_UND),
/* 0001 0110 1010 */ TABDECL(OP_SMUL_T_B),		// OP_SMUL_T_B
/* 0001 0110 1011 */ TABDECL(OP_STRH_PRE_INDE_M_IMM_OFF),		// OP_STRH_PRE_INDE_M_IMM_OFF
/* 0001 0110 1100 */ TABDECL(OP_SMUL_B_T),		// OP_SMUL_B_T
/* 0001 0110 1101 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX
/* 0001 0110 1110 */ TABDECL(OP_SMUL_T_T),		// OP_SMUL_T_T
/* 0001 0110 1111 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX

/* 0001 0111 0000 */ TABDECL(OP_CMN_LSL_IMM),		// OP_CMN_LSL_IMM
/* 0001 0111 0001 */ TABDECL(OP_CMN_LSL_REG),		// OP_CMN_LSL_REG
/* 0001 0111 0010 */ TABDECL(OP_CMN_LSR_IMM),		// OP_CMN_LSR_IMM
/* 0001 0111 0011 */ TABDECL(OP_CMN_LSR_REG),		// OP_CMN_LSR_REG
/* 0001 0111 0100 */ TABDECL(OP_CMN_ASR_IMM),		// OP_CMN_ASR_IMM
/* 0001 0111 0101 */ TABDECL(OP_CMN_ASR_REG),		// OP_CMN_ASR_REG
/* 0001 0111 0110 */ TABDECL(OP_CMN_ROR_IMM),		// OP_CMN_ROR_IMM
/* 0001 0111 0111 */ TABDECL(OP_CMN_ROR_REG),		// OP_CMN_ROR_REG
/* 0001 0111 1000 */ TABDECL(OP_CMN_LSL_IMM),		// OP_CMN_LSL_IMM
/* 0001 0111 1001 */ TABDECL(OP_UND),
/* 0001 0111 1010 */ TABDECL(OP_CMN_LSR_IMM),		// OP_CMN_LSR_IMM
/* 0001 0111 1011 */ TABDECL(OP_LDRH_PRE_INDE_M_IMM_OFF),		// OP_LDRH_PRE_INDE_M_IMM_OFF
/* 0001 0111 1100 */ TABDECL(OP_CMN_ASR_IMM),		// OP_CMN_ASR_IMM
/* 0001 0111 1101 */ TABDECL(OP_LDRSB_PRE_INDE_M_IMM_OFF),		// OP_LDRSB_PRE_INDE_M_IMM_OFF
/* 0001 0111 1110 */ TABDECL(OP_CMN_ROR_IMM),		// OP_CMN_ROR_IMM
/* 0001 0111 1111 */ TABDECL(OP_LDRSH_PRE_INDE_M_IMM_OFF),		// OP_LDRSH_PRE_INDE_M_IMM_OFF

/* 0001 1000 0000 */ TABDECL(OP_ORR_LSL_IMM),		// OP_ORR_LSL_IMM
/* 0001 1000 0001 */ TABDECL(OP_ORR_LSL_REG),		// OP_ORR_LSL_REG
/* 0001 1000 0010 */ TABDECL(OP_ORR_LSR_IMM),		// OP_ORR_LSR_IMM
/* 0001 1000 0011 */ TABDECL(OP_ORR_LSR_REG),		// OP_ORR_LSR_REG
/* 0001 1000 0100 */ TABDECL(OP_ORR_ASR_IMM),		// OP_ORR_ASR_IMM
/* 0001 1000 0101 */ TABDECL(OP_ORR_ASR_REG),		// OP_ORR_ASR_REG
/* 0001 1000 0110 */ TABDECL(OP_ORR_ROR_IMM),		// OP_ORR_ROR_IMM
/* 0001 1000 0111 */ TABDECL(OP_ORR_ROR_REG),		// OP_ORR_ROR_REG
/* 0001 1000 1000 */ TABDECL(OP_ORR_LSL_IMM),		// OP_ORR_LSL_IMM
/* 0001 1000 1001 */ TABDECL(OP_STREX),				// OP_STREX
/* 0001 1000 1010 */ TABDECL(OP_ORR_LSR_IMM),		// OP_ORR_LSR_IMM
/* 0001 1000 1011 */ TABDECL(OP_STRH_P_REG_OFF),		// OP_STRH_P_REG_OFF
/* 0001 1000 1100 */ TABDECL(OP_ORR_ASR_IMM),		// OP_ORR_ASR_IMM
/* 0001 1000 1101 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX
/* 0001 1000 1110 */ TABDECL(OP_ORR_ROR_IMM),		// OP_ORR_ROR_IMM
/* 0001 1000 1111 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX

/* 0001 1001 0000 */ TABDECL(OP_ORR_S_LSL_IMM),		// OP_ORR_S_LSL_IMM
/* 0001 1001 0001 */ TABDECL(OP_ORR_S_LSL_REG),		// OP_ORR_S_LSL_REG
/* 0001 1001 0010 */ TABDECL(OP_ORR_S_LSR_IMM),		// OP_ORR_S_LSR_IMM
/* 0001 1001 0011 */ TABDECL(OP_ORR_S_LSR_REG),		// OP_ORR_S_LSR_REG
/* 0001 1001 0100 */ TABDECL(OP_ORR_S_ASR_IMM),		// OP_ORR_S_ASR_IMM
/* 0001 1001 0101 */ TABDECL(OP_ORR_S_ASR_REG),		// OP_ORR_S_ASR_REG
/* 0001 1001 0110 */ TABDECL(OP_ORR_S_ROR_IMM),		// OP_ORR_S_ROR_IMM
/* 0001 1001 0111 */ TABDECL(OP_ORR_S_ROR_REG),		// OP_ORR_S_ROR_REG
/* 0001 1001 1000 */ TABDECL(OP_ORR_S_LSL_IMM),		// OP_ORR_S_LSL_IMM
/* 0001 1001 1001 */ TABDECL(OP_LDREX),				// OP_LDREX
/* 0001 1001 1010 */ TABDECL(OP_ORR_S_LSR_IMM),		// OP_ORR_S_LSR_IMM
/* 0001 1001 1011 */ TABDECL(OP_LDRH_P_REG_OFF),		// OP_LDRH_P_REG_OFF
/* 0001 1001 1100 */ TABDECL(OP_ORR_S_ASR_IMM),		// OP_ORR_S_ASR_IMM
/* 0001 1001 1101 */ TABDECL(OP_LDRSB_P_REG_OFF),		// OP_LDRSB_P_REG_OFF
/* 0001 1001 1110 */ TABDECL(OP_ORR_S_ROR_IMM),		// OP_ORR_S_ROR_IMM
/* 0001 1001 1111 */ TABDECL(OP_LDRSH_P_REG_OFF),		// OP_LDRSH_P_REG_OFF

/* 0001 1010 0000 */ TABDECL(OP_MOV_LSL_IMM),		// OP_MOV_LSL_IMM
/* 0001 1010 0001 */ TABDECL(OP_MOV_LSL_REG),		// OP_MOV_LSL_REG
/* 0001 1010 0010 */ TABDECL(OP_MOV_LSR_IMM),		// OP_MOV_LSR_IMM
/* 0001 1010 0011 */ TABDECL(OP_MOV_LSR_REG),		// OP_MOV_LSR_REG
/* 0001 1010 0100 */ TABDECL(OP_MOV_ASR_IMM),		// OP_MOV_ASR_IMM
/* 0001 1010 0101 */ TABDECL(OP_MOV_ASR_REG),		// OP_MOV_ASR_REG
/* 0001 1010 0110 */ TABDECL(OP_MOV_ROR_IMM),		// OP_MOV_ROR_IMM
/* 0001 1010 0111 */ TABDECL(OP_MOV_ROR_REG),		// OP_MOV_ROR_REG
/* 0001 1010 1000 */ TABDECL(OP_MOV_LSL_IMM),		// OP_MOV_LSL_IMM
/* 0001 1010 1001 */ TABDECL(OP_UND),
/* 0001 1010 1010 */ TABDECL(OP_MOV_LSR_IMM),		// OP_MOV_LSR_IMM
/* 0001 1010 1011 */ TABDECL(OP_STRH_PRE_INDE_P_REG_OFF),		// OP_STRH_PRE_INDE_P_REG_OFF
/* 0001 1010 1100 */ TABDECL(OP_MOV_ASR_IMM),		// OP_MOV_ASR_IMM
/* 0001 1010 1101 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX
/* 0001 1010 1110 */ TABDECL(OP_MOV_ROR_IMM),		// OP_MOV_ROR_IMM
/* 0001 1010 1111 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX

/* 0001 1011 0000 */ TABDECL(OP_MOV_S_LSL_IMM),		// OP_MOV_S_LSL_IMM
/* 0001 1011 0001 */ TABDECL(OP_MOV_S_LSL_REG),		// OP_MOV_S_LSL_REG
/* 0001 1011 0010 */ TABDECL(OP_MOV_S_LSR_IMM),		// OP_MOV_S_LSR_IMM
/* 0001 1011 0011 */ TABDECL(OP_MOV_S_LSR_REG),		// OP_MOV_S_LSR_REG
/* 0001 1011 0100 */ TABDECL(OP_MOV_S_ASR_IMM),		// OP_MOV_S_ASR_IMM
/* 0001 1011 0101 */ TABDECL(OP_MOV_S_ASR_REG),		// OP_MOV_S_ASR_REG
/* 0001 1011 0110 */ TABDECL(OP_MOV_S_ROR_IMM),		// OP_MOV_S_ROR_IMM
/* 0001 1011 0111 */ TABDECL(OP_MOV_S_ROR_REG),		// OP_MOV_S_ROR_REG
/* 0001 1011 1000 */ TABDECL(OP_MOV_S_LSL_IMM),		// OP_MOV_S_LSL_IMM
/* 0001 1011 1001 */ TABDECL(OP_UND),
/* 0001 1011 1010 */ TABDECL(OP_MOV_S_LSR_IMM),		// OP_MOV_S_LSR_IMM
/* 0001 1011 1011 */ TABDECL(OP_LDRH_PRE_INDE_P_REG_OFF),		// OP_LDRH_PRE_INDE_P_REG_OFF
/* 0001 1011 1100 */ TABDECL(OP_MOV_S_ASR_IMM),		// OP_MOV_S_ASR_IMM
/* 0001 1011 1101 */ TABDECL(OP_LDRSB_PRE_INDE_P_REG_OFF),		// OP_LDRSB_PRE_INDE_P_REG_OFF
/* 0001 1011 1110 */ TABDECL(OP_MOV_S_ROR_IMM),		// OP_MOV_S_ROR_IMM
/* 0001 1011 1111 */ TABDECL(OP_LDRSH_PRE_INDE_P_REG_OFF),		// OP_LDRSH_PRE_INDE_P_REG_OFF

/* 0001 1100 0000 */ TABDECL(OP_BIC_LSL_IMM),		// OP_BIC_LSL_IMM
/* 0001 1100 0001 */ TABDECL(OP_BIC_LSL_REG),		// OP_BIC_LSL_REG
/* 0001 1100 0010 */ TABDECL(OP_BIC_LSR_IMM),		// OP_BIC_LSR_IMM
/* 0001 1100 0011 */ TABDECL(OP_BIC_LSR_REG),		// OP_BIC_LSR_REG
/* 0001 1100 0100 */ TABDECL(OP_BIC_ASR_IMM),		// OP_BIC_ASR_IMM
/* 0001 1100 0101 */ TABDECL(OP_BIC_ASR_REG),		// OP_BIC_ASR_REG
/* 0001 1100 0110 */ TABDECL(OP_BIC_ROR_IMM),		// OP_BIC_ROR_IMM
/* 0001 1100 0111 */ TABDECL(OP_BIC_ROR_REG),		// OP_BIC_ROR_REG
/* 0001 1100 1000 */ TABDECL(OP_BIC_LSL_IMM),		// OP_BIC_LSL_IMM
/* 0001 1100 1001 */ TABDECL(OP_UND),
/* 0001 1100 1010 */ TABDECL(OP_BIC_LSR_IMM),		// OP_BIC_LSR_IMM
/* 0001 1100 1011 */ TABDECL(OP_STRH_P_IMM_OFF),		// OP_STRH_P_IMM_OFF
/* 0001 1100 1100 */ TABDECL(OP_BIC_ASR_IMM),		// OP_BIC_ASR_IMM
/* 0001 1100 1101 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX
/* 0001 1100 1110 */ TABDECL(OP_BIC_ROR_IMM),		// OP_BIC_ROR_IMM
/* 0001 1100 1111 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX

/* 0001 1101 0000 */ TABDECL(OP_BIC_S_LSL_IMM),		// OP_BIC_S_LSL_IMM
/* 0001 1101 0001 */ TABDECL(OP_BIC_S_LSL_REG),		// OP_BIC_S_LSL_REG
/* 0001 1101 0010 */ TABDECL(OP_BIC_S_LSR_IMM),		// OP_BIC_S_LSR_IMM
/* 0001 1101 0011 */ TABDECL(OP_BIC_S_LSR_REG),		// OP_BIC_S_LSR_REG
/* 0001 1101 0100 */ TABDECL(OP_BIC_S_ASR_IMM),		// OP_BIC_S_ASR_IMM
/* 0001 1101 0101 */ TABDECL(OP_BIC_S_ASR_REG),		// OP_BIC_S_ASR_REG
/* 0001 1101 0110 */ TABDECL(OP_BIC_S_ROR_IMM),		// OP_BIC_S_ROR_IMM
/* 0001 1101 0111 */ TABDECL(OP_BIC_S_ROR_REG),		// OP_BIC_S_ROR_REG
/* 0001 1101 1000 */ TABDECL(OP_BIC_S_LSL_IMM),		// OP_BIC_S_LSL_IMM
/* 0001 1101 1001 */ TABDECL(OP_UND),
/* 0001 1101 1010 */ TABDECL(OP_BIC_S_LSR_IMM),		// OP_BIC_S_LSR_IMM
/* 0001 1101 1011 */ TABDECL(OP_LDRH_P_IMM_OFF),		// OP_LDRH_P_IMM_OFF
/* 0001 1101 1100 */ TABDECL(OP_BIC_S_ASR_IMM),		// OP_BIC_S_ASR_IMM
/* 0001 1101 1101 */ TABDECL(OP_LDRSB_P_IMM_OFF),		// OP_LDRSB_P_IMM_OFF
/* 0001 1101 1110 */ TABDECL(OP_BIC_S_ROR_IMM),		// OP_BIC_S_ROR_IMM
/* 0001 1101 1111 */ TABDECL(OP_LDRSH_P_IMM_OFF),		// OP_LDRSH_P_IMM_OFF

/* 0001 1110 0000 */ TABDECL(OP_MVN_LSL_IMM),		// OP_MVN_LSL_IMM
/* 0001 1110 0001 */ TABDECL(OP_MVN_LSL_REG),		// OP_MVN_LSL_REG
/* 0001 1110 0010 */ TABDECL(OP_MVN_LSR_IMM),		// OP_MVN_LSR_IMM
/* 0001 1110 0011 */ TABDECL(OP_MVN_LSR_REG),		// OP_MVN_LSR_REG
/* 0001 1110 0100 */ TABDECL(OP_MVN_ASR_IMM),		// OP_MVN_ASR_IMM
/* 0001 1110 0101 */ TABDECL(OP_MVN_ASR_REG),		// OP_MVN_ASR_REG
/* 0001 1110 0110 */ TABDECL(OP_MVN_ROR_IMM),		// OP_MVN_ROR_IMM
/* 0001 1110 0111 */ TABDECL(OP_MVN_ROR_REG),		// OP_MVN_ROR_REG
/* 0001 1110 1000 */ TABDECL(OP_MVN_LSL_IMM),		// OP_MVN_LSL_IMM
/* 0001 1110 1001 */ TABDECL(OP_UND),
/* 0001 1110 1010 */ TABDECL(OP_MVN_LSR_IMM),		// OP_MVN_LSR_IMM
/* 0001 1110 1011 */ TABDECL(OP_STRH_PRE_INDE_P_IMM_OFF),		// OP_STRH_PRE_INDE_P_IMM_OFF
/* 0001 1110 1100 */ TABDECL(OP_MVN_ASR_IMM),		// OP_MVN_ASR_IMM
/* 0001 1110 1101 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX
/* 0001 1110 1110 */ TABDECL(OP_MVN_ROR_IMM),		// OP_MVN_ROR_IMM
/* 0001 1110 1111 */ TABDECL(OP_LDRD_STRD_OFFSET_PRE_INDEX),		// OP_LDRD_STRD_OFFSET_PRE_INDEX

/* 0001 1111 0000 */ TABDECL(OP_MVN_S_LSL_IMM),		// OP_MVN_S_LSL_IMM
/* 0001 1111 0001 */ TABDECL(OP_MVN_S_LSL_REG),		// OP_MVN_S_LSL_REG
/* 0001 1111 0010 */ TABDECL(OP_MVN_S_LSR_IMM),		// OP_MVN_S_LSR_IMM
/* 0001 1111 0011 */ TABDECL(OP_MVN_S_LSR_REG),		// OP_MVN_S_LSR_REG
/* 0001 1111 0100 */ TABDECL(OP_MVN_S_ASR_IMM),		// OP_MVN_S_ASR_IMM
/* 0001 1111 0101 */ TABDECL(OP_MVN_S_ASR_REG),		// OP_MVN_S_ASR_REG
/* 0001 1111 0110 */ TABDECL(OP_MVN_S_ROR_IMM),		// OP_MVN_S_ROR_IMM
/* 0001 1111 0111 */ TABDECL(OP_MVN_S_ROR_REG),		// OP_MVN_S_ROR_REG
/* 0001 1111 1000 */ TABDECL(OP_MVN_S_LSL_IMM),		// OP_MVN_S_LSL_IMM
/* 0001 1111 1001 */ TABDECL(OP_UND),
/* 0001 1111 1010 */ TABDECL(OP_MVN_S_LSR_IMM),		// OP_MVN_S_LSR_IMM
/* 0001 1111 1011 */ TABDECL(OP_LDRH_PRE_INDE_P_IMM_OFF),		// OP_LDRH_PRE_INDE_P_IMM_OFF
/* 0001 1111 1100 */ TABDECL(OP_MVN_S_ASR_IMM),		// OP_MVN_S_ASR_IMM
/* 0001 1111 1101 */ TABDECL(OP_LDRSB_PRE_INDE_P_IMM_OFF),		// OP_LDRSB_PRE_INDE_P_IMM_OFF
/* 0001 1111 1110 */ TABDECL(OP_MVN_S_ROR_IMM),		// OP_MVN_S_ROR_IMM
/* 0001 1111 1111 */ TABDECL(OP_LDRSH_PRE_INDE_P_IMM_OFF),		// OP_LDRSH_PRE_INDE_P_IMM_OFF

/* 0010 0000 0000 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 0001 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 0010 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 0011 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 0100 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 0101 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 0110 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 0111 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 1000 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 1001 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 1010 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 1011 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 1100 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 1101 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 1110 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL
/* 0010 0000 1111 */ TABDECL(OP_AND_IMM_VAL),		// OP_AND_IMM_VAL

/* 0010 0001 0000 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 0001 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 0010 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 0011 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 0100 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 0101 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 0110 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 0111 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 1000 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 1001 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 1010 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 1011 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 1100 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 1101 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 1110 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL
/* 0010 0001 1111 */ TABDECL(OP_AND_S_IMM_VAL),		// OP_AND_S_IMM_VAL

/* 0010 0010 0000 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 0001 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 0010 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 0011 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 0100 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 0101 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 0110 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 0111 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 1000 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 1001 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 1010 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 1011 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 1100 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 1101 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 1110 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL
/* 0010 0010 1111 */ TABDECL(OP_EOR_IMM_VAL),		// OP_EOR_IMM_VAL

/* 0010 0011 0000 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 0001 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 0010 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 0011 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 0100 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 0101 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 0110 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 0111 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 1000 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 1001 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 1010 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 1011 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 1100 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 1101 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 1110 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL
/* 0010 0011 1111 */ TABDECL(OP_EOR_S_IMM_VAL),		// OP_EOR_S_IMM_VAL

/* 0010 0100 0000 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 0001 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 0010 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 0011 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 0100 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 0101 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 0110 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 0111 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 1000 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 1001 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 1010 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 1011 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 1100 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 1101 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 1110 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL
/* 0010 0100 1111 */ TABDECL(OP_SUB_IMM_VAL),		// OP_SUB_IMM_VAL

/* 0010 0101 0000 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 0001 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 0010 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 0011 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 0100 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 0101 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 0110 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 0111 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 1000 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 1001 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 1010 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 1011 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 1100 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 1101 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 1110 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL
/* 0010 0101 1111 */ TABDECL(OP_SUB_S_IMM_VAL),		// OP_SUB_S_IMM_VAL

/* 0010 0110 0000 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 0001 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 0010 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 0011 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 0100 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 0101 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 0110 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 0111 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 1000 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 1001 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 1010 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 1011 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 1100 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 1101 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 1110 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL
/* 0010 0110 1111 */ TABDECL(OP_RSB_IMM_VAL),		// OP_RSB_IMM_VAL

/* 0010 0111 0000 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 0001 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 0010 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 0011 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 0100 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 0101 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 0110 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 0111 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 1000 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 1001 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 1010 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 1011 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 1100 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 1101 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 1110 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL
/* 0010 0111 1111 */ TABDECL(OP_RSB_S_IMM_VAL),		// OP_RSB_S_IMM_VAL

/* 0010 1000 0000 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 0001 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 0010 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 0011 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 0100 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 0101 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 0110 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 0111 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 1000 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 1001 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 1010 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 1011 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 1100 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 1101 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 1110 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL
/* 0010 1000 1111 */ TABDECL(OP_ADD_IMM_VAL),		// OP_ADD_IMM_VAL

/* 0010 1001 0000 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 0001 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 0010 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 0011 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 0100 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 0101 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 0110 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 0111 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 1000 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 1001 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 1010 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 1011 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 1100 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 1101 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 1110 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL
/* 0010 1001 1111 */ TABDECL(OP_ADD_S_IMM_VAL),		// OP_ADD_S_IMM_VAL

/* 0010 1010 0000 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 0001 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 0010 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 0011 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 0100 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 0101 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 0110 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 0111 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 1000 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 1001 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 1010 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 1011 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 1100 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 1101 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 1110 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL
/* 0010 1010 1111 */ TABDECL(OP_ADC_IMM_VAL),		// OP_ADC_IMM_VAL

/* 0010 1011 0000 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 0001 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 0010 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 0011 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 0100 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 0101 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 0110 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 0111 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 1000 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 1001 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 1010 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 1011 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 1100 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 1101 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 1110 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL
/* 0010 1011 1111 */ TABDECL(OP_ADC_S_IMM_VAL),		// OP_ADC_S_IMM_VAL

/* 0010 1100 0000 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 0001 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 0010 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 0011 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 0100 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 0101 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 0110 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 0111 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 1000 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 1001 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 1010 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 1011 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 1100 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 1101 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 1110 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL
/* 0010 1100 1111 */ TABDECL(OP_SBC_IMM_VAL),		// OP_SBC_IMM_VAL

/* 0010 1101 0000 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 0001 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 0010 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 0011 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 0100 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 0101 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 0110 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 0111 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 1000 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 1001 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 1010 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 1011 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 1100 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 1101 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 1110 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL
/* 0010 1101 1111 */ TABDECL(OP_SBC_S_IMM_VAL),		// OP_SBC_S_IMM_VAL

/* 0010 1110 0000 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 0001 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 0010 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 0011 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 0100 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 0101 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 0110 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 0111 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 1000 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 1001 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 1010 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 1011 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 1100 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 1101 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 1110 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL
/* 0010 1110 1111 */ TABDECL(OP_RSC_IMM_VAL),		// OP_RSC_IMM_VAL

/* 0010 1111 0000 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 0001 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 0010 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 0011 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 0100 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 0101 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 0110 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 0111 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 1000 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 1001 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 1010 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 1011 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 1100 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 1101 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 1110 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL
/* 0010 1111 1111 */ TABDECL(OP_RSC_S_IMM_VAL),		// OP_RSC_S_IMM_VAL

/* 0011 0000 0000 */ TABDECL(OP_UND),
/* 0011 0000 0001 */ TABDECL(OP_UND),
/* 0011 0000 0010 */ TABDECL(OP_UND),
/* 0011 0000 0011 */ TABDECL(OP_UND),
/* 0011 0000 0100 */ TABDECL(OP_UND),
/* 0011 0000 0101 */ TABDECL(OP_UND),
/* 0011 0000 0110 */ TABDECL(OP_UND),
/* 0011 0000 0111 */ TABDECL(OP_UND),
/* 0011 0000 1000 */ TABDECL(OP_UND),
/* 0011 0000 1001 */ TABDECL(OP_UND),
/* 0011 0000 1010 */ TABDECL(OP_UND),
/* 0011 0000 1011 */ TABDECL(OP_UND),
/* 0011 0000 1100 */ TABDECL(OP_UND),
/* 0011 0000 1101 */ TABDECL(OP_UND),
/* 0011 0000 1110 */ TABDECL(OP_UND),
/* 0011 0000 1111 */ TABDECL(OP_UND),

/* 0011 0001 0000 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 0001 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 0010 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 0011 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 0100 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 0101 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 0110 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 0111 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 1000 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 1001 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 1010 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 1011 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 1100 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 1101 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 1110 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL
/* 0011 0001 1111 */ TABDECL(OP_TST_IMM_VAL),		// OP_TST_IMM_VAL

/* 0011 0010 0000 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 0001 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 0010 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 0011 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 0100 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 0101 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 0110 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 0111 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 1000 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 1001 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 1010 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 1011 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 1100 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 1101 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 1110 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL
/* 0011 0010 1111 */ TABDECL(OP_MSR_CPSR_IMM_VAL),		// OP_MSR_CPSR_IMM_VAL

/* 0011 0011 0000 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 0001 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 0010 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 0011 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 0100 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 0101 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 0110 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 0111 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 1000 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 1001 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 1010 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 1011 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 1100 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 1101 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 1110 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL
/* 0011 0011 1111 */ TABDECL(OP_TEQ_IMM_VAL),		// OP_TEQ_IMM_VAL

/* 0011 0100 0000 */ TABDECL(OP_UND),
/* 0011 0100 0001 */ TABDECL(OP_UND),
/* 0011 0100 0010 */ TABDECL(OP_UND),
/* 0011 0100 0011 */ TABDECL(OP_UND),
/* 0011 0100 0100 */ TABDECL(OP_UND),
/* 0011 0100 0101 */ TABDECL(OP_UND),
/* 0011 0100 0110 */ TABDECL(OP_UND),
/* 0011 0100 0111 */ TABDECL(OP_UND),
/* 0011 0100 1000 */ TABDECL(OP_UND),
/* 0011 0100 1001 */ TABDECL(OP_UND),
/* 0011 0100 1010 */ TABDECL(OP_UND),
/* 0011 0100 1011 */ TABDECL(OP_UND),
/* 0011 0100 1100 */ TABDECL(OP_UND),
/* 0011 0100 1101 */ TABDECL(OP_UND),
/* 0011 0100 1110 */ TABDECL(OP_UND),
/* 0011 0100 1111 */ TABDECL(OP_UND),

/* 0011 0101 0000 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 0001 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 0010 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 0011 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 0100 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 0101 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 0110 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 0111 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 1000 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 1001 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 1010 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 1011 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 1100 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 1101 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 1110 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL
/* 0011 0101 1111 */ TABDECL(OP_CMP_IMM_VAL),		// OP_CMP_IMM_VAL

/* 0011 0110 0000 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 0001 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 0010 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 0011 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 0100 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 0101 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 0110 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 0111 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 1000 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 1001 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 1010 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 1011 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 1100 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 1101 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 1110 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL
/* 0011 0110 1111 */ TABDECL(OP_MSR_SPSR_IMM_VAL),		// OP_MSR_SPSR_IMM_VAL

/* 0011 0111 0000 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 0001 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 0010 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 0011 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 0100 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 0101 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 0110 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 0111 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 1000 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 1001 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 1010 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 1011 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 1100 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 1101 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 1110 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL
/* 0011 0111 1111 */ TABDECL(OP_CMN_IMM_VAL),		// OP_CMN_IMM_VAL

/* 0011 1000 0000 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 0001 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 0010 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 0011 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 0100 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 0101 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 0110 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 0111 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 1000 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 1001 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 1010 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 1011 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 1100 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 1101 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 1110 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL
/* 0011 1000 1111 */ TABDECL(OP_ORR_IMM_VAL),		// OP_ORR_IMM_VAL

/* 0011 1001 0000 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 0001 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 0010 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 0011 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 0100 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 0101 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 0110 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 0111 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 1000 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 1001 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 1010 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 1011 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 1100 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 1101 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 1110 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL
/* 0011 1001 1111 */ TABDECL(OP_ORR_S_IMM_VAL),		// OP_ORR_S_IMM_VAL

/* 0011 1010 0000 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 0001 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 0010 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 0011 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 0100 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 0101 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 0110 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 0111 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 1000 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 1001 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 1010 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 1011 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 1100 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 1101 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 1110 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL
/* 0011 1010 1111 */ TABDECL(OP_MOV_IMM_VAL),		// OP_MOV_IMM_VAL

/* 0011 1011 0000 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 0001 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 0010 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 0011 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 0100 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 0101 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 0110 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 0111 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 1000 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 1001 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 1010 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 1011 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 1100 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 1101 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 1110 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL
/* 0011 1011 1111 */ TABDECL(OP_MOV_S_IMM_VAL),		// OP_MOV_S_IMM_VAL

/* 0011 1100 0000 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 0001 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 0010 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 0011 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 0100 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 0101 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 0110 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 0111 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 1000 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 1001 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 1010 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 1011 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 1100 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 1101 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 1110 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL
/* 0011 1100 1111 */ TABDECL(OP_BIC_IMM_VAL),		// OP_BIC_IMM_VAL

/* 0011 1101 0000 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 0001 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 0010 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 0011 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 0100 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 0101 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 0110 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 0111 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 1000 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 1001 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 1010 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 1011 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 1100 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 1101 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 1110 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL
/* 0011 1101 1111 */ TABDECL(OP_BIC_S_IMM_VAL),		// OP_BIC_S_IMM_VAL

/* 0011 1110 0000 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 0001 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 0010 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 0011 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 0100 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 0101 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 0110 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 0111 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 1000 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 1001 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 1010 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 1011 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 1100 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 1101 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 1110 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL
/* 0011 1110 1111 */ TABDECL(OP_MVN_IMM_VAL),		// OP_MVN_IMM_VAL

/* 0011 1111 0000 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 0001 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 0010 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 0011 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 0100 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 0101 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 0110 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 0111 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 1000 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 1001 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 1010 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 1011 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 1100 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 1101 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 1110 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL
/* 0011 1111 1111 */ TABDECL(OP_MVN_S_IMM_VAL),		// OP_MVN_S_IMM_VAL

/* 0100 0000 0000 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 0001 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 0010 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 0011 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 0100 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 0101 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 0110 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 0111 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 1000 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 1001 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 1010 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 1011 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 1100 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 1101 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 1110 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0000 1111 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND

/* 0100 0001 0000 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 0001 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 0010 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 0011 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 0100 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 0101 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 0110 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 0111 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 1000 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 1001 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 1010 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 1011 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 1100 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 1101 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 1110 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0001 1111 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND

/* 0100 0010 0000 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 0001 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 0010 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 0011 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 0100 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 0101 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 0110 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 0111 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 1000 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 1001 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 1010 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 1011 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 1100 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 1101 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 1110 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND
/* 0100 0010 1111 */ TABDECL(OP_STR_M_IMM_OFF_POSTIND),		// OP_STR_M_IMM_OFF_POSTIND

/* 0100 0011 0000 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 0001 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 0010 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 0011 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 0100 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 0101 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 0110 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 0111 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 1000 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 1001 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 1010 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 1011 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 1100 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 1101 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 1110 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND
/* 0100 0011 1111 */ TABDECL(OP_LDR_M_IMM_OFF_POSTIND),		// OP_LDR_M_IMM_OFF_POSTIND

/* 0100 0100 0000 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 0001 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 0010 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 0011 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 0100 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 0101 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 0110 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 0111 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 1000 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 1001 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 1010 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 1011 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 1100 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 1101 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 1110 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0100 1111 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND

/* 0100 0101 0000 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 0001 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 0010 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 0011 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 0100 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 0101 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 0110 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 0111 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 1000 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 1001 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 1010 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 1011 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 1100 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 1101 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 1110 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0101 1111 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND

/* 0100 0110 0000 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 0001 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 0010 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 0011 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 0100 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 0101 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 0110 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 0111 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 1000 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 1001 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 1010 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 1011 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 1100 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 1101 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 1110 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND
/* 0100 0110 1111 */ TABDECL(OP_STRB_M_IMM_OFF_POSTIND),		// OP_STRB_M_IMM_OFF_POSTIND

/* 0100 0111 0000 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 0001 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 0010 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 0011 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 0100 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 0101 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 0110 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 0111 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 1000 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 1001 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 1010 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 1011 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 1100 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 1101 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 1110 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND
/* 0100 0111 1111 */ TABDECL(OP_LDRB_M_IMM_OFF_POSTIND),		// OP_LDRB_M_IMM_OFF_POSTIND

/* 0100 1000 0000 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 0001 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 0010 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 0011 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 0100 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 0101 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 0110 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 0111 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 1000 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 1001 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 1010 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 1011 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 1100 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 1101 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 1110 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1000 1111 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND

/* 0100 1001 0000 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 0001 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 0010 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 0011 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 0100 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 0101 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 0110 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 0111 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 1000 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 1001 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 1010 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 1011 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 1100 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 1101 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 1110 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1001 1111 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND

/* 0100 1010 0000 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 0001 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 0010 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 0011 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 0100 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 0101 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 0110 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 0111 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 1000 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 1001 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 1010 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 1011 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 1100 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 1101 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 1110 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND
/* 0100 1010 1111 */ TABDECL(OP_STR_P_IMM_OFF_POSTIND),		// OP_STR_P_IMM_OFF_POSTIND

/* 0100 1011 0000 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 0001 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 0010 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 0011 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 0100 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 0101 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 0110 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 0111 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 1000 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 1001 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 1010 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 1011 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 1100 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 1101 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 1110 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND
/* 0100 1011 1111 */ TABDECL(OP_LDR_P_IMM_OFF_POSTIND),		// OP_LDR_P_IMM_OFF_POSTIND

/* 0100 1100 0000 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 0001 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 0010 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 0011 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 0100 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 0101 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 0110 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 0111 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 1000 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 1001 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 1010 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 1011 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 1100 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 1101 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 1110 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1100 1111 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND

/* 0100 1101 0000 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 0001 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 0010 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 0011 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 0100 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 0101 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 0110 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 0111 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 1000 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 1001 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 1010 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 1011 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 1100 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 1101 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 1110 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1101 1111 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND

/* 0100 1110 0000 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 0001 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 0010 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 0011 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 0100 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 0101 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 0110 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 0111 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 1000 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 1001 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 1010 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 1011 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 1100 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 1101 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 1110 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND
/* 0100 1110 1111 */ TABDECL(OP_STRB_P_IMM_OFF_POSTIND),		// OP_STRB_P_IMM_OFF_POSTIND

/* 0100 1111 0000 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 0001 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 0010 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 0011 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 0100 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 0101 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 0110 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 0111 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 1000 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 1001 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 1010 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 1011 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 1100 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 1101 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 1110 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND
/* 0100 1111 1111 */ TABDECL(OP_LDRB_P_IMM_OFF_POSTIND),		// OP_LDRB_P_IMM_OFF_POSTIND

/* 0101 0000 0000 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 0001 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 0010 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 0011 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 0100 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 0101 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 0110 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 0111 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 1000 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 1001 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 1010 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 1011 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 1100 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 1101 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 1110 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF
/* 0101 0000 1111 */ TABDECL(OP_STR_M_IMM_OFF),		// OP_STR_M_IMM_OFF

/* 0101 0001 0000 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 0001 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 0010 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 0011 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 0100 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 0101 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 0110 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 0111 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 1000 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 1001 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 1010 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 1011 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 1100 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 1101 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 1110 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF
/* 0101 0001 1111 */ TABDECL(OP_LDR_M_IMM_OFF),		// OP_LDR_M_IMM_OFF

/* 0101 0010 0000 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 0001 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 0010 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 0011 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 0100 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 0101 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 0110 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 0111 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 1000 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 1001 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 1010 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 1011 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 1100 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 1101 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 1110 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND
/* 0101 0010 1111 */ TABDECL(OP_STR_M_IMM_OFF_PREIND),		// OP_STR_M_IMM_OFF_PREIND

/* 0101 0011 0000 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 0001 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 0010 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 0011 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 0100 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 0101 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 0110 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 0111 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 1000 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 1001 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 1010 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 1011 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 1100 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 1101 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 1110 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND
/* 0101 0011 1111 */ TABDECL(OP_LDR_M_IMM_OFF_PREIND),		// OP_LDR_M_IMM_OFF_PREIND

/* 0101 0100 0000 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 0001 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 0010 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 0011 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 0100 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 0101 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 0110 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 0111 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 1000 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 1001 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 1010 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 1011 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 1100 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 1101 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 1110 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF
/* 0101 0100 1111 */ TABDECL(OP_STRB_M_IMM_OFF),		// OP_STRB_M_IMM_OFF

/* 0101 0101 0000 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 0001 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 0010 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 0011 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 0100 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 0101 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 0110 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 0111 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 1000 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 1001 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 1010 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 1011 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 1100 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 1101 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 1110 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF
/* 0101 0101 1111 */ TABDECL(OP_LDRB_M_IMM_OFF),		// OP_LDRB_M_IMM_OFF

/* 0101 0110 0000 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 0001 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 0010 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 0011 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 0100 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 0101 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 0110 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 0111 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 1000 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 1001 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 1010 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 1011 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 1100 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 1101 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 1110 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND
/* 0101 0110 1111 */ TABDECL(OP_STRB_M_IMM_OFF_PREIND),		// OP_STRB_M_IMM_OFF_PREIND

/* 0101 0111 0000 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 0001 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 0010 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 0011 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 0100 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 0101 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 0110 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 0111 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 1000 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 1001 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 1010 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 1011 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 1100 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 1101 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 1110 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND
/* 0101 0111 1111 */ TABDECL(OP_LDRB_M_IMM_OFF_PREIND),		// OP_LDRB_M_IMM_OFF_PREIND

/* 0101 1000 0000 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 0001 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 0010 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 0011 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 0100 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 0101 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 0110 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 0111 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 1000 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 1001 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 1010 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 1011 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 1100 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 1101 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 1110 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF
/* 0101 1000 1111 */ TABDECL(OP_STR_P_IMM_OFF),		// OP_STR_P_IMM_OFF

/* 0101 1001 0000 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 0001 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 0010 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 0011 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 0100 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 0101 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 0110 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 0111 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 1000 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 1001 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 1010 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 1011 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 1100 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 1101 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 1110 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF
/* 0101 1001 1111 */ TABDECL(OP_LDR_P_IMM_OFF),		// OP_LDR_P_IMM_OFF

/* 0101 1010 0000 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 0001 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 0010 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 0011 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 0100 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 0101 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 0110 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 0111 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 1000 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 1001 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 1010 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 1011 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 1100 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 1101 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 1110 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND
/* 0101 1010 1111 */ TABDECL(OP_STR_P_IMM_OFF_PREIND),		// OP_STR_P_IMM_OFF_PREIND

/* 0101 1011 0000 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 0001 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 0010 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 0011 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 0100 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 0101 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 0110 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 0111 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 1000 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 1001 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 1010 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 1011 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 1100 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 1101 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 1110 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND
/* 0101 1011 1111 */ TABDECL(OP_LDR_P_IMM_OFF_PREIND),		// OP_LDR_P_IMM_OFF_PREIND

/* 0101 1100 0000 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 0001 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 0010 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 0011 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 0100 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 0101 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 0110 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 0111 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 1000 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 1001 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 1010 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 1011 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 1100 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 1101 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 1110 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF
/* 0101 1100 1111 */ TABDECL(OP_STRB_P_IMM_OFF),		// OP_STRB_P_IMM_OFF

/* 0101 1101 0000 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 0001 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 0010 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 0011 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 0100 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 0101 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 0110 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 0111 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 1000 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 1001 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 1010 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 1011 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 1100 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 1101 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 1110 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF
/* 0101 1101 1111 */ TABDECL(OP_LDRB_P_IMM_OFF),		// OP_LDRB_P_IMM_OFF

/* 0101 1110 0000 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 0001 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 0010 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 0011 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 0100 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 0101 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 0110 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 0111 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 1000 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 1001 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 1010 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 1011 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 1100 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 1101 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 1110 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND
/* 0101 1110 1111 */ TABDECL(OP_STRB_P_IMM_OFF_PREIND),		// OP_STRB_P_IMM_OFF_PREIND

/* 0101 1111 0000 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 0001 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 0010 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 0011 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 0100 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 0101 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 0110 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 0111 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 1000 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 1001 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 1010 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 1011 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 1100 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 1101 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 1110 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND
/* 0101 1111 1111 */ TABDECL(OP_LDRB_P_IMM_OFF_PREIND),		// OP_LDRB_P_IMM_OFF_PREIND

/* 0110 0000 0000 */ TABDECL(OP_STR_M_LSL_IMM_OFF_POSTIND),		// OP_STR_M_LSL_IMM_OFF_POSTIND
/* 0110 0000 0001 */ TABDECL(OP_UND),
/* 0110 0000 0010 */ TABDECL(OP_STR_M_LSR_IMM_OFF_POSTIND),		// OP_STR_M_LSR_IMM_OFF_POSTIND
/* 0110 0000 0011 */ TABDECL(OP_UND),
/* 0110 0000 0100 */ TABDECL(OP_STR_M_ASR_IMM_OFF_POSTIND),		// OP_STR_M_ASR_IMM_OFF_POSTIND
/* 0110 0000 0101 */ TABDECL(OP_UND),
/* 0110 0000 0110 */ TABDECL(OP_STR_M_ROR_IMM_OFF_POSTIND),		// OP_STR_M_ROR_IMM_OFF_POSTIND
/* 0110 0000 0111 */ TABDECL(OP_UND),
/* 0110 0000 1000 */ TABDECL(OP_STR_M_LSL_IMM_OFF_POSTIND),		// OP_STR_M_LSL_IMM_OFF_POSTIND
/* 0110 0000 1001 */ TABDECL(OP_UND),
/* 0110 0000 1010 */ TABDECL(OP_STR_M_LSR_IMM_OFF_POSTIND),		// OP_STR_M_LSR_IMM_OFF_POSTIND
/* 0110 0000 1011 */ TABDECL(OP_UND),
/* 0110 0000 1100 */ TABDECL(OP_STR_M_ASR_IMM_OFF_POSTIND),		// OP_STR_M_ASR_IMM_OFF_POSTIND
/* 0110 0000 1101 */ TABDECL(OP_UND),
/* 0110 0000 1110 */ TABDECL(OP_STR_M_ROR_IMM_OFF_POSTIND),		// OP_STR_M_ROR_IMM_OFF_POSTIND
/* 0110 0000 1111 */ TABDECL(OP_UND),

/* 0110 0001 0000 */ TABDECL(OP_LDR_M_LSL_IMM_OFF_POSTIND),		// OP_LDR_M_LSL_IMM_OFF_POSTIND
/* 0110 0001 0001 */ TABDECL(OP_UND),
/* 0110 0001 0010 */ TABDECL(OP_LDR_M_LSR_IMM_OFF_POSTIND),		// OP_LDR_M_LSR_IMM_OFF_POSTIND
/* 0110 0001 0011 */ TABDECL(OP_UND),
/* 0110 0001 0100 */ TABDECL(OP_LDR_M_ASR_IMM_OFF_POSTIND),		// OP_LDR_M_ASR_IMM_OFF_POSTIND
/* 0110 0001 0101 */ TABDECL(OP_UND),
/* 0110 0001 0110 */ TABDECL(OP_LDR_M_ROR_IMM_OFF_POSTIND),		// OP_LDR_M_ROR_IMM_OFF_POSTIND
/* 0110 0001 0111 */ TABDECL(OP_UND),
/* 0110 0001 1000 */ TABDECL(OP_LDR_M_LSL_IMM_OFF_POSTIND),		// OP_LDR_M_LSL_IMM_OFF_POSTIND
/* 0110 0001 1001 */ TABDECL(OP_UND),
/* 0110 0001 1010 */ TABDECL(OP_LDR_M_LSR_IMM_OFF_POSTIND),		// OP_LDR_M_LSR_IMM_OFF_POSTIND
/* 0110 0001 1011 */ TABDECL(OP_UND),
/* 0110 0001 1100 */ TABDECL(OP_LDR_M_ASR_IMM_OFF_POSTIND),		// OP_LDR_M_ASR_IMM_OFF_POSTIND
/* 0110 0001 1101 */ TABDECL(OP_UND),
/* 0110 0001 1110 */ TABDECL(OP_LDR_M_ROR_IMM_OFF_POSTIND),		// OP_LDR_M_ROR_IMM_OFF_POSTIND
/* 0110 0001 1111 */ TABDECL(OP_UND),

/* 0110 0010 0000 */ TABDECL(OP_STR_M_LSL_IMM_OFF_POSTIND),		// OP_STR_M_LSL_IMM_OFF_POSTIND
/* 0110 0010 0001 */ TABDECL(OP_UND),
/* 0110 0010 0010 */ TABDECL(OP_STR_M_LSR_IMM_OFF_POSTIND),		// OP_STR_M_LSR_IMM_OFF_POSTIND
/* 0110 0010 0011 */ TABDECL(OP_UND),
/* 0110 0010 0100 */ TABDECL(OP_STR_M_ASR_IMM_OFF_POSTIND),		// OP_STR_M_ASR_IMM_OFF_POSTIND
/* 0110 0010 0101 */ TABDECL(OP_UND),
/* 0110 0010 0110 */ TABDECL(OP_STR_M_ROR_IMM_OFF_POSTIND),		// OP_STR_M_ROR_IMM_OFF_POSTIND
/* 0110 0010 0111 */ TABDECL(OP_UND),
/* 0110 0010 1000 */ TABDECL(OP_STR_M_LSL_IMM_OFF_POSTIND),		// OP_STR_M_LSL_IMM_OFF_POSTIND
/* 0110 0010 1001 */ TABDECL(OP_UND),
/* 0110 0010 1010 */ TABDECL(OP_STR_M_LSR_IMM_OFF_POSTIND),		// OP_STR_M_LSR_IMM_OFF_POSTIND
/* 0110 0010 1011 */ TABDECL(OP_UND),
/* 0110 0010 1100 */ TABDECL(OP_STR_M_ASR_IMM_OFF_POSTIND),		// OP_STR_M_ASR_IMM_OFF_POSTIND
/* 0110 0010 1101 */ TABDECL(OP_UND),
/* 0110 0010 1110 */ TABDECL(OP_STR_M_ROR_IMM_OFF_POSTIND),		// OP_STR_M_ROR_IMM_OFF_POSTIND
/* 0110 0010 1111 */ TABDECL(OP_UND),

/* 0110 0011 0000 */ TABDECL(OP_LDR_M_LSL_IMM_OFF_POSTIND),		// OP_LDR_M_LSL_IMM_OFF_POSTIND
/* 0110 0011 0001 */ TABDECL(OP_UND),
/* 0110 0011 0010 */ TABDECL(OP_LDR_M_LSR_IMM_OFF_POSTIND),		// OP_LDR_M_LSR_IMM_OFF_POSTIND
/* 0110 0011 0011 */ TABDECL(OP_UND),
/* 0110 0011 0100 */ TABDECL(OP_LDR_M_ASR_IMM_OFF_POSTIND),		// OP_LDR_M_ASR_IMM_OFF_POSTIND
/* 0110 0011 0101 */ TABDECL(OP_UND),
/* 0110 0011 0110 */ TABDECL(OP_LDR_M_ROR_IMM_OFF_POSTIND),		// OP_LDR_M_ROR_IMM_OFF_POSTIND
/* 0110 0011 0111 */ TABDECL(OP_UND),
/* 0110 0011 1000 */ TABDECL(OP_LDR_M_LSL_IMM_OFF_POSTIND),		// OP_LDR_M_LSL_IMM_OFF_POSTIND
/* 0110 0011 1001 */ TABDECL(OP_UND),
/* 0110 0011 1010 */ TABDECL(OP_LDR_M_LSR_IMM_OFF_POSTIND),		// OP_LDR_M_LSR_IMM_OFF_POSTIND
/* 0110 0011 1011 */ TABDECL(OP_UND),
/* 0110 0011 1100 */ TABDECL(OP_LDR_M_ASR_IMM_OFF_POSTIND),		// OP_LDR_M_ASR_IMM_OFF_POSTIND
/* 0110 0011 1101 */ TABDECL(OP_UND),
/* 0110 0011 1110 */ TABDECL(OP_LDR_M_ROR_IMM_OFF_POSTIND),		// OP_LDR_M_ROR_IMM_OFF_POSTIND
/* 0110 0011 1111 */ TABDECL(OP_UND),

/* 0110 0100 0000 */ TABDECL(OP_STRB_M_LSL_IMM_OFF_POSTIND),		// OP_STRB_M_LSL_IMM_OFF_POSTIND
/* 0110 0100 0001 */ TABDECL(OP_UND),
/* 0110 0100 0010 */ TABDECL(OP_STRB_M_LSR_IMM_OFF_POSTIND),		// OP_STRB_M_LSR_IMM_OFF_POSTIND
/* 0110 0100 0011 */ TABDECL(OP_UND),
/* 0110 0100 0100 */ TABDECL(OP_STRB_M_ASR_IMM_OFF_POSTIND),		// OP_STRB_M_ASR_IMM_OFF_POSTIND
/* 0110 0100 0101 */ TABDECL(OP_UND),
/* 0110 0100 0110 */ TABDECL(OP_STRB_M_ROR_IMM_OFF_POSTIND),		// OP_STRB_M_ROR_IMM_OFF_POSTIND
/* 0110 0100 0111 */ TABDECL(OP_UND),
/* 0110 0100 1000 */ TABDECL(OP_STRB_M_LSL_IMM_OFF_POSTIND),		// OP_STRB_M_LSL_IMM_OFF_POSTIND
/* 0110 0100 1001 */ TABDECL(OP_UND),
/* 0110 0100 1010 */ TABDECL(OP_STRB_M_LSR_IMM_OFF_POSTIND),		// OP_STRB_M_LSR_IMM_OFF_POSTIND
/* 0110 0100 1011 */ TABDECL(OP_UND),
/* 0110 0100 1100 */ TABDECL(OP_STRB_M_ASR_IMM_OFF_POSTIND),		// OP_STRB_M_ASR_IMM_OFF_POSTIND
/* 0110 0100 1101 */ TABDECL(OP_UND),
/* 0110 0100 1110 */ TABDECL(OP_STRB_M_ROR_IMM_OFF_POSTIND),		// OP_STRB_M_ROR_IMM_OFF_POSTIND
/* 0110 0100 1111 */ TABDECL(OP_UND),

/* 0110 0101 0000 */ TABDECL(OP_LDRB_M_LSL_IMM_OFF_POSTIND),		// OP_LDRB_M_LSL_IMM_OFF_POSTIND
/* 0110 0101 0001 */ TABDECL(OP_UND),
/* 0110 0101 0010 */ TABDECL(OP_LDRB_M_LSR_IMM_OFF_POSTIND),		// OP_LDRB_M_LSR_IMM_OFF_POSTIND
/* 0110 0101 0011 */ TABDECL(OP_UND),
/* 0110 0101 0100 */ TABDECL(OP_LDRB_M_ASR_IMM_OFF_POSTIND),		// OP_LDRB_M_ASR_IMM_OFF_POSTIND
/* 0110 0101 0101 */ TABDECL(OP_UND),
/* 0110 0101 0110 */ TABDECL(OP_LDRB_M_ROR_IMM_OFF_POSTIND),		// OP_LDRB_M_ROR_IMM_OFF_POSTIND
/* 0110 0101 0111 */ TABDECL(OP_UND),
/* 0110 0101 1000 */ TABDECL(OP_LDRB_M_LSL_IMM_OFF_POSTIND),		// OP_LDRB_M_LSL_IMM_OFF_POSTIND
/* 0110 0101 1001 */ TABDECL(OP_UND),
/* 0110 0101 1010 */ TABDECL(OP_LDRB_M_LSR_IMM_OFF_POSTIND),		// OP_LDRB_M_LSR_IMM_OFF_POSTIND
/* 0110 0101 1011 */ TABDECL(OP_UND),
/* 0110 0101 1100 */ TABDECL(OP_LDRB_M_ASR_IMM_OFF_POSTIND),		// OP_LDRB_M_ASR_IMM_OFF_POSTIND
/* 0110 0101 1101 */ TABDECL(OP_UND),
/* 0110 0101 1110 */ TABDECL(OP_LDRB_M_ROR_IMM_OFF_POSTIND),		// OP_LDRB_M_ROR_IMM_OFF_POSTIND
/* 0110 0101 1111 */ TABDECL(OP_UND),

/* 0110 0110 0000 */ TABDECL(OP_STRB_M_LSL_IMM_OFF_POSTIND),		// OP_STRB_M_LSL_IMM_OFF_POSTIND
/* 0110 0110 0001 */ TABDECL(OP_UND),
/* 0110 0110 0010 */ TABDECL(OP_STRB_M_LSR_IMM_OFF_POSTIND),		// OP_STRB_M_LSR_IMM_OFF_POSTIND
/* 0110 0110 0011 */ TABDECL(OP_UND),
/* 0110 0110 0100 */ TABDECL(OP_STRB_M_ASR_IMM_OFF_POSTIND),		// OP_STRB_M_ASR_IMM_OFF_POSTIND
/* 0110 0110 0101 */ TABDECL(OP_UND),
/* 0110 0110 0110 */ TABDECL(OP_STRB_M_ROR_IMM_OFF_POSTIND),		// OP_STRB_M_ROR_IMM_OFF_POSTIND
/* 0110 0110 0111 */ TABDECL(OP_UND),
/* 0110 0110 1000 */ TABDECL(OP_STRB_M_LSL_IMM_OFF_POSTIND),		// OP_STRB_M_LSL_IMM_OFF_POSTIND
/* 0110 0110 1001 */ TABDECL(OP_UND),
/* 0110 0110 1010 */ TABDECL(OP_STRB_M_LSR_IMM_OFF_POSTIND),		// OP_STRB_M_LSR_IMM_OFF_POSTIND
/* 0110 0110 1011 */ TABDECL(OP_UND),
/* 0110 0110 1100 */ TABDECL(OP_STRB_M_ASR_IMM_OFF_POSTIND),		// OP_STRB_M_ASR_IMM_OFF_POSTIND
/* 0110 0110 1101 */ TABDECL(OP_UND),
/* 0110 0110 1110 */ TABDECL(OP_STRB_M_ROR_IMM_OFF_POSTIND),		// OP_STRB_M_ROR_IMM_OFF_POSTIND
/* 0110 0110 1111 */ TABDECL(OP_UND),

/* 0110 0111 0000 */ TABDECL(OP_LDRB_M_LSL_IMM_OFF_POSTIND),		// OP_LDRB_M_LSL_IMM_OFF_POSTIND
/* 0110 0111 0001 */ TABDECL(OP_UND),
/* 0110 0111 0010 */ TABDECL(OP_LDRB_M_LSR_IMM_OFF_POSTIND),		// OP_LDRB_M_LSR_IMM_OFF_POSTIND
/* 0110 0111 0011 */ TABDECL(OP_UND),
/* 0110 0111 0100 */ TABDECL(OP_LDRB_M_ASR_IMM_OFF_POSTIND),		// OP_LDRB_M_ASR_IMM_OFF_POSTIND
/* 0110 0111 0101 */ TABDECL(OP_UND),
/* 0110 0111 0110 */ TABDECL(OP_LDRB_M_ROR_IMM_OFF_POSTIND),		// OP_LDRB_M_ROR_IMM_OFF_POSTIND
/* 0110 0111 0111 */ TABDECL(OP_UND),
/* 0110 0111 1000 */ TABDECL(OP_LDRB_M_LSL_IMM_OFF_POSTIND),		// OP_LDRB_M_LSL_IMM_OFF_POSTIND
/* 0110 0111 1001 */ TABDECL(OP_UND),
/* 0110 0111 1010 */ TABDECL(OP_LDRB_M_LSR_IMM_OFF_POSTIND),		// OP_LDRB_M_LSR_IMM_OFF_POSTIND
/* 0110 0111 1011 */ TABDECL(OP_UND),
/* 0110 0111 1100 */ TABDECL(OP_LDRB_M_ASR_IMM_OFF_POSTIND),		// OP_LDRB_M_ASR_IMM_OFF_POSTIND
/* 0110 0111 1101 */ TABDECL(OP_UND),
/* 0110 0111 1110 */ TABDECL(OP_LDRB_M_ROR_IMM_OFF_POSTIND),		// OP_LDRB_M_ROR_IMM_OFF_POSTIND
/* 0110 0111 1111 */ TABDECL(OP_UND),

/* 0110 1000 0000 */ TABDECL(OP_STR_P_LSL_IMM_OFF_POSTIND),		// OP_STR_P_LSL_IMM_OFF_POSTIND
/* 0110 1000 0001 */ TABDECL(OP_UND),
/* 0110 1000 0010 */ TABDECL(OP_STR_P_LSR_IMM_OFF_POSTIND),		// OP_STR_P_LSR_IMM_OFF_POSTIND
/* 0110 1000 0011 */ TABDECL(OP_UND),
/* 0110 1000 0100 */ TABDECL(OP_STR_P_ASR_IMM_OFF_POSTIND),		// OP_STR_P_ASR_IMM_OFF_POSTIND
/* 0110 1000 0101 */ TABDECL(OP_UND),
/* 0110 1000 0110 */ TABDECL(OP_STR_P_ROR_IMM_OFF_POSTIND),		// OP_STR_P_ROR_IMM_OFF_POSTIND
/* 0110 1000 0111 */ TABDECL(OP_UND),
/* 0110 1000 1000 */ TABDECL(OP_STR_P_LSL_IMM_OFF_POSTIND),		// OP_STR_P_LSL_IMM_OFF_POSTIND
/* 0110 1000 1001 */ TABDECL(OP_UND),
/* 0110 1000 1010 */ TABDECL(OP_STR_P_LSR_IMM_OFF_POSTIND),		// OP_STR_P_LSR_IMM_OFF_POSTIND
/* 0110 1000 1011 */ TABDECL(OP_UND),
/* 0110 1000 1100 */ TABDECL(OP_STR_P_ASR_IMM_OFF_POSTIND),		// OP_STR_P_ASR_IMM_OFF_POSTIND
/* 0110 1000 1101 */ TABDECL(OP_UND),
/* 0110 1000 1110 */ TABDECL(OP_STR_P_ROR_IMM_OFF_POSTIND),		// OP_STR_P_ROR_IMM_OFF_POSTIND
/* 0110 1000 1111 */ TABDECL(OP_UND),

/* 0110 1001 0000 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1001 0001 */ TABDECL(OP_UND),
/* 0110 1001 0010 */ TABDECL(OP_LDR_P_LSR_IMM_OFF_POSTIND),		// OP_LDR_P_LSR_IMM_OFF_POSTIND
/* 0110 1001 0011 */ TABDECL(OP_UND),
/* 0110 1001 0100 */ TABDECL(OP_LDR_P_ASR_IMM_OFF_POSTIND),		// OP_LDR_P_ASR_IMM_OFF_POSTIND
/* 0110 1001 0101 */ TABDECL(OP_UND),
/* 0110 1001 0110 */ TABDECL(OP_LDR_P_ROR_IMM_OFF_POSTIND),		// OP_LDR_P_ROR_IMM_OFF_POSTIND
/* 0110 1001 0111 */ TABDECL(OP_UND),
/* 0110 1001 1000 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1001 1001 */ TABDECL(OP_UND),
/* 0110 1001 1010 */ TABDECL(OP_LDR_P_LSR_IMM_OFF_POSTIND),		// OP_LDR_P_LSR_IMM_OFF_POSTIND
/* 0110 1001 1011 */ TABDECL(OP_UND),
/* 0110 1001 1100 */ TABDECL(OP_LDR_P_ASR_IMM_OFF_POSTIND),		// OP_LDR_P_ASR_IMM_OFF_POSTIND
/* 0110 1001 1101 */ TABDECL(OP_UND),
/* 0110 1001 1110 */ TABDECL(OP_LDR_P_ROR_IMM_OFF_POSTIND),		// OP_LDR_P_ROR_IMM_OFF_POSTIND
/* 0110 1001 1111 */ TABDECL(OP_UND),

/* 0110 1010 0000 */ TABDECL(OP_STR_P_LSL_IMM_OFF_POSTIND),		// OP_STR_P_LSL_IMM_OFF_POSTIND
/* 0110 1010 0001 */ TABDECL(OP_UND),
/* 0110 1010 0010 */ TABDECL(OP_STR_P_LSR_IMM_OFF_POSTIND),		// OP_STR_P_LSR_IMM_OFF_POSTIND
/* 0110 1010 0011 */ TABDECL(OP_UND),
/* 0110 1010 0100 */ TABDECL(OP_STR_P_ASR_IMM_OFF_POSTIND),		// OP_STR_P_ASR_IMM_OFF_POSTIND
/* 0110 1010 0101 */ TABDECL(OP_UND),
/* 0110 1010 0110 */ TABDECL(OP_STR_P_ROR_IMM_OFF_POSTIND),		// OP_STR_P_ROR_IMM_OFF_POSTIND
/* 0110 1010 0111 */ TABDECL(OP_UND),
/* 0110 1010 1000 */ TABDECL(OP_STR_P_LSL_IMM_OFF_POSTIND),		// OP_STR_P_LSL_IMM_OFF_POSTIND
/* 0110 1010 1001 */ TABDECL(OP_UND),
/* 0110 1010 1010 */ TABDECL(OP_STR_P_LSR_IMM_OFF_POSTIND),		// OP_STR_P_LSR_IMM_OFF_POSTIND
/* 0110 1010 1011 */ TABDECL(OP_UND),
/* 0110 1010 1100 */ TABDECL(OP_STR_P_ASR_IMM_OFF_POSTIND),		// OP_STR_P_ASR_IMM_OFF_POSTIND
/* 0110 1010 1101 */ TABDECL(OP_UND),
/* 0110 1010 1110 */ TABDECL(OP_STR_P_ROR_IMM_OFF_POSTIND),		// OP_STR_P_ROR_IMM_OFF_POSTIND
/* 0110 1010 1111 */ TABDECL(OP_UND),

/* 0110 1011 0000 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1011 0001 */ TABDECL(OP_UND),
/* 0110 1011 0010 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1011 0011 */ TABDECL(OP_UND),
/* 0110 1011 0100 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1011 0101 */ TABDECL(OP_UND),
/* 0110 1011 0110 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1011 0111 */ TABDECL(OP_UND),
/* 0110 1011 1000 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1011 1001 */ TABDECL(OP_UND),
/* 0110 1011 1010 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1011 1011 */ TABDECL(OP_UND),
/* 0110 1011 1100 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1011 1101 */ TABDECL(OP_UND),
/* 0110 1011 1110 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_POSTIND),		// OP_LDR_P_LSL_IMM_OFF_POSTIND
/* 0110 1011 1111 */ TABDECL(OP_UND),

/* 0110 1100 0000 */ TABDECL(OP_STRB_P_LSL_IMM_OFF_POSTIND),		// OP_STRB_P_LSL_IMM_OFF_POSTIND
/* 0110 1100 0001 */ TABDECL(OP_UND),
/* 0110 1100 0010 */ TABDECL(OP_STRB_P_LSR_IMM_OFF_POSTIND),		// OP_STRB_P_LSR_IMM_OFF_POSTIND
/* 0110 1100 0011 */ TABDECL(OP_UND),
/* 0110 1100 0100 */ TABDECL(OP_STRB_P_ASR_IMM_OFF_POSTIND),		// OP_STRB_P_ASR_IMM_OFF_POSTIND
/* 0110 1100 0101 */ TABDECL(OP_UND),
/* 0110 1100 0110 */ TABDECL(OP_STRB_P_ROR_IMM_OFF_POSTIND),		// OP_STRB_P_ROR_IMM_OFF_POSTIND
/* 0110 1100 0111 */ TABDECL(OP_UND),
/* 0110 1100 1000 */ TABDECL(OP_STRB_P_LSL_IMM_OFF_POSTIND),		// OP_STRB_P_LSL_IMM_OFF_POSTIND
/* 0110 1100 1001 */ TABDECL(OP_UND),
/* 0110 1100 1010 */ TABDECL(OP_STRB_P_LSR_IMM_OFF_POSTIND),		// OP_STRB_P_LSR_IMM_OFF_POSTIND
/* 0110 1100 1011 */ TABDECL(OP_UND),
/* 0110 1100 1100 */ TABDECL(OP_STRB_P_ASR_IMM_OFF_POSTIND),		// OP_STRB_P_ASR_IMM_OFF_POSTIND
/* 0110 1100 1101 */ TABDECL(OP_UND),
/* 0110 1100 1110 */ TABDECL(OP_STRB_P_ROR_IMM_OFF_POSTIND),		// OP_STRB_P_ROR_IMM_OFF_POSTIND
/* 0110 1100 1111 */ TABDECL(OP_UND),

/* 0110 1101 0000 */ TABDECL(OP_LDRB_P_LSL_IMM_OFF_POSTIND),		// OP_LDRB_P_LSL_IMM_OFF_POSTIND
/* 0110 1101 0001 */ TABDECL(OP_UND),
/* 0110 1101 0010 */ TABDECL(OP_LDRB_P_LSR_IMM_OFF_POSTIND),		// OP_LDRB_P_LSR_IMM_OFF_POSTIND
/* 0110 1101 0011 */ TABDECL(OP_UND),
/* 0110 1101 0100 */ TABDECL(OP_LDRB_P_ASR_IMM_OFF_POSTIND),		// OP_LDRB_P_ASR_IMM_OFF_POSTIND
/* 0110 1101 0101 */ TABDECL(OP_UND),
/* 0110 1101 0110 */ TABDECL(OP_LDRB_P_ROR_IMM_OFF_POSTIND),		// OP_LDRB_P_ROR_IMM_OFF_POSTIND
/* 0110 1101 0111 */ TABDECL(OP_UND),
/* 0110 1101 1000 */ TABDECL(OP_LDRB_P_LSL_IMM_OFF_POSTIND),		// OP_LDRB_P_LSL_IMM_OFF_POSTIND
/* 0110 1101 1001 */ TABDECL(OP_UND),
/* 0110 1101 1010 */ TABDECL(OP_LDRB_P_LSR_IMM_OFF_POSTIND),		// OP_LDRB_P_LSR_IMM_OFF_POSTIND
/* 0110 1101 1011 */ TABDECL(OP_UND),
/* 0110 1101 1100 */ TABDECL(OP_LDRB_P_ASR_IMM_OFF_POSTIND),		// OP_LDRB_P_ASR_IMM_OFF_POSTIND
/* 0110 1101 1101 */ TABDECL(OP_UND),
/* 0110 1101 1110 */ TABDECL(OP_LDRB_P_ROR_IMM_OFF_POSTIND),		// OP_LDRB_P_ROR_IMM_OFF_POSTIND
/* 0110 1101 1111 */ TABDECL(OP_UND),

/* 0110 1110 0000 */ TABDECL(OP_STRB_P_LSL_IMM_OFF_POSTIND),		// OP_STRB_P_LSL_IMM_OFF_POSTIND
/* 0110 1110 0001 */ TABDECL(OP_UND),
/* 0110 1110 0010 */ TABDECL(OP_STRB_P_LSR_IMM_OFF_POSTIND),		// OP_STRB_P_LSR_IMM_OFF_POSTIND
/* 0110 1110 0011 */ TABDECL(OP_UND),
/* 0110 1110 0100 */ TABDECL(OP_STRB_P_ASR_IMM_OFF_POSTIND),		// OP_STRB_P_ASR_IMM_OFF_POSTIND
/* 0110 1110 0101 */ TABDECL(OP_UND),
/* 0110 1110 0110 */ TABDECL(OP_STRB_P_ROR_IMM_OFF_POSTIND),		// OP_STRB_P_ROR_IMM_OFF_POSTIND
/* 0110 1110 0111 */ TABDECL(OP_UND),
/* 0110 1110 1000 */ TABDECL(OP_STRB_P_LSL_IMM_OFF_POSTIND),		// OP_STRB_P_LSL_IMM_OFF_POSTIND
/* 0110 1110 1001 */ TABDECL(OP_UND),
/* 0110 1110 1010 */ TABDECL(OP_STRB_P_LSR_IMM_OFF_POSTIND),		// OP_STRB_P_LSR_IMM_OFF_POSTIND
/* 0110 1110 1011 */ TABDECL(OP_UND),
/* 0110 1110 1100 */ TABDECL(OP_STRB_P_ASR_IMM_OFF_POSTIND),		// OP_STRB_P_ASR_IMM_OFF_POSTIND
/* 0110 1110 1101 */ TABDECL(OP_UND),
/* 0110 1110 1110 */ TABDECL(OP_STRB_P_ROR_IMM_OFF_POSTIND),		// OP_STRB_P_ROR_IMM_OFF_POSTIND
/* 0110 1110 1111 */ TABDECL(OP_UND),

/* 0110 1111 0000 */ TABDECL(OP_LDRB_P_LSL_IMM_OFF_POSTIND),		// OP_LDRB_P_LSL_IMM_OFF_POSTIND
/* 0110 1111 0001 */ TABDECL(OP_UND),
/* 0110 1111 0010 */ TABDECL(OP_LDRB_P_LSR_IMM_OFF_POSTIND),		// OP_LDRB_P_LSR_IMM_OFF_POSTIND
/* 0110 1111 0011 */ TABDECL(OP_UND),
/* 0110 1111 0100 */ TABDECL(OP_LDRB_P_ASR_IMM_OFF_POSTIND),		// OP_LDRB_P_ASR_IMM_OFF_POSTIND
/* 0110 1111 0101 */ TABDECL(OP_UND),
/* 0110 1111 0110 */ TABDECL(OP_LDRB_P_ROR_IMM_OFF_POSTIND),		// OP_LDRB_P_ROR_IMM_OFF_POSTIND
/* 0110 1111 0111 */ TABDECL(OP_UND),
/* 0110 1111 1000 */ TABDECL(OP_LDRB_P_LSL_IMM_OFF_POSTIND),		// OP_LDRB_P_LSL_IMM_OFF_POSTIND
/* 0110 1111 1001 */ TABDECL(OP_UND),
/* 0110 1111 1010 */ TABDECL(OP_LDRB_P_LSR_IMM_OFF_POSTIND),		// OP_LDRB_P_LSR_IMM_OFF_POSTIND
/* 0110 1111 1011 */ TABDECL(OP_UND),
/* 0110 1111 1100 */ TABDECL(OP_LDRB_P_ASR_IMM_OFF_POSTIND),		// OP_LDRB_P_ASR_IMM_OFF_POSTIND
/* 0110 1111 1101 */ TABDECL(OP_UND),
/* 0110 1111 1110 */ TABDECL(OP_LDRB_P_ROR_IMM_OFF_POSTIND),		// OP_LDRB_P_ROR_IMM_OFF_POSTIND
/* 0110 1111 1111 */ TABDECL(OP_UND),

/* 0111 0000 0000 */ TABDECL(OP_STR_M_LSL_IMM_OFF),		// OP_STR_M_LSL_IMM_OFF
/* 0111 0000 0001 */ TABDECL(OP_UND),
/* 0111 0000 0010 */ TABDECL(OP_STR_M_LSR_IMM_OFF),		// OP_STR_M_LSR_IMM_OFF
/* 0111 0000 0011 */ TABDECL(OP_UND),
/* 0111 0000 0100 */ TABDECL(OP_STR_M_ASR_IMM_OFF),		// OP_STR_M_ASR_IMM_OFF
/* 0111 0000 0101 */ TABDECL(OP_UND),
/* 0111 0000 0110 */ TABDECL(OP_STR_M_ROR_IMM_OFF),		// OP_STR_M_ROR_IMM_OFF
/* 0111 0000 0111 */ TABDECL(OP_UND),
/* 0111 0000 1000 */ TABDECL(OP_STR_M_LSL_IMM_OFF),		// OP_STR_M_LSL_IMM_OFF
/* 0111 0000 1001 */ TABDECL(OP_UND),
/* 0111 0000 1010 */ TABDECL(OP_STR_M_LSR_IMM_OFF),		// OP_STR_M_LSR_IMM_OFF
/* 0111 0000 1011 */ TABDECL(OP_UND),
/* 0111 0000 1100 */ TABDECL(OP_STR_M_ASR_IMM_OFF),		// OP_STR_M_ASR_IMM_OFF
/* 0111 0000 1101 */ TABDECL(OP_UND),
/* 0111 0000 1110 */ TABDECL(OP_STR_M_ROR_IMM_OFF),		// OP_STR_M_ROR_IMM_OFF
/* 0111 0000 1111 */ TABDECL(OP_UND),

/* 0111 0001 0000 */ TABDECL(OP_LDR_M_LSL_IMM_OFF),		// OP_LDR_M_LSL_IMM_OFF
/* 0111 0001 0001 */ TABDECL(OP_UND),
/* 0111 0001 0010 */ TABDECL(OP_LDR_M_LSR_IMM_OFF),		// OP_LDR_M_LSR_IMM_OFF
/* 0111 0001 0011 */ TABDECL(OP_UND),
/* 0111 0001 0100 */ TABDECL(OP_LDR_M_ASR_IMM_OFF),		// OP_LDR_M_ASR_IMM_OFF
/* 0111 0001 0101 */ TABDECL(OP_UND),
/* 0111 0001 0110 */ TABDECL(OP_LDR_M_ROR_IMM_OFF),		// OP_LDR_M_ROR_IMM_OFF
/* 0111 0001 0111 */ TABDECL(OP_UND),
/* 0111 0001 1000 */ TABDECL(OP_LDR_M_LSL_IMM_OFF),		// OP_LDR_M_LSL_IMM_OFF
/* 0111 0001 1001 */ TABDECL(OP_UND),
/* 0111 0001 1010 */ TABDECL(OP_LDR_M_LSR_IMM_OFF),		// OP_LDR_M_LSR_IMM_OFF
/* 0111 0001 1011 */ TABDECL(OP_UND),
/* 0111 0001 1100 */ TABDECL(OP_LDR_M_ASR_IMM_OFF),		// OP_LDR_M_ASR_IMM_OFF
/* 0111 0001 1101 */ TABDECL(OP_UND),
/* 0111 0001 1110 */ TABDECL(OP_LDR_M_ROR_IMM_OFF),		// OP_LDR_M_ROR_IMM_OFF
/* 0111 0001 1111 */ TABDECL(OP_UND),

/* 0111 0010 0000 */ TABDECL(OP_STR_M_LSL_IMM_OFF_PREIND),		// OP_STR_M_LSL_IMM_OFF_PREIND
/* 0111 0010 0001 */ TABDECL(OP_UND),
/* 0111 0010 0010 */ TABDECL(OP_STR_M_LSR_IMM_OFF_PREIND),		// OP_STR_M_LSR_IMM_OFF_PREIND
/* 0111 0010 0011 */ TABDECL(OP_UND),
/* 0111 0010 0100 */ TABDECL(OP_STR_M_ASR_IMM_OFF_PREIND),		// OP_STR_M_ASR_IMM_OFF_PREIND
/* 0111 0010 0101 */ TABDECL(OP_UND),
/* 0111 0010 0110 */ TABDECL(OP_STR_M_ROR_IMM_OFF_PREIND),		// OP_STR_M_ROR_IMM_OFF_PREIND
/* 0111 0010 0111 */ TABDECL(OP_UND),
/* 0111 0010 1000 */ TABDECL(OP_STR_M_LSL_IMM_OFF_PREIND),		// OP_STR_M_LSL_IMM_OFF_PREIND
/* 0111 0010 1001 */ TABDECL(OP_UND),
/* 0111 0010 1010 */ TABDECL(OP_STR_M_LSR_IMM_OFF_PREIND),		// OP_STR_M_LSR_IMM_OFF_PREIND
/* 0111 0010 1011 */ TABDECL(OP_UND),
/* 0111 0010 1100 */ TABDECL(OP_STR_M_ASR_IMM_OFF_PREIND),		// OP_STR_M_ASR_IMM_OFF_PREIND
/* 0111 0010 1101 */ TABDECL(OP_UND),
/* 0111 0010 1110 */ TABDECL(OP_STR_M_ROR_IMM_OFF_PREIND),		// OP_STR_M_ROR_IMM_OFF_PREIND
/* 0111 0010 1111 */ TABDECL(OP_UND),

/* 0111 0011 0000 */ TABDECL(OP_LDR_M_LSL_IMM_OFF_PREIND),		// OP_LDR_M_LSL_IMM_OFF_PREIND
/* 0111 0011 0001 */ TABDECL(OP_UND),
/* 0111 0011 0010 */ TABDECL(OP_LDR_M_LSR_IMM_OFF_PREIND),		// OP_LDR_M_LSR_IMM_OFF_PREIND
/* 0111 0011 0011 */ TABDECL(OP_UND),
/* 0111 0011 0100 */ TABDECL(OP_LDR_M_ASR_IMM_OFF_PREIND),		// OP_LDR_M_ASR_IMM_OFF_PREIND
/* 0111 0011 0101 */ TABDECL(OP_UND),
/* 0111 0011 0110 */ TABDECL(OP_LDR_M_ROR_IMM_OFF_PREIND),		// OP_LDR_M_ROR_IMM_OFF_PREIND
/* 0111 0011 0111 */ TABDECL(OP_UND),
/* 0111 0011 1000 */ TABDECL(OP_LDR_M_LSL_IMM_OFF_PREIND),		// OP_LDR_M_LSL_IMM_OFF_PREIND
/* 0111 0011 1001 */ TABDECL(OP_UND),
/* 0111 0011 1010 */ TABDECL(OP_LDR_M_LSR_IMM_OFF_PREIND),		// OP_LDR_M_LSR_IMM_OFF_PREIND
/* 0111 0011 1011 */ TABDECL(OP_UND),
/* 0111 0011 1100 */ TABDECL(OP_LDR_M_ASR_IMM_OFF_PREIND),		// OP_LDR_M_ASR_IMM_OFF_PREIND
/* 0111 0011 1101 */ TABDECL(OP_UND),
/* 0111 0011 1110 */ TABDECL(OP_LDR_M_ROR_IMM_OFF_PREIND),		// OP_LDR_M_ROR_IMM_OFF_PREIND
/* 0111 0011 1111 */ TABDECL(OP_UND),

/* 0111 0100 0000 */ TABDECL(OP_STRB_M_LSL_IMM_OFF),		// OP_STRB_M_LSL_IMM_OFF
/* 0111 0100 0001 */ TABDECL(OP_UND),
/* 0111 0100 0010 */ TABDECL(OP_STRB_M_LSR_IMM_OFF),		// OP_STRB_M_LSR_IMM_OFF
/* 0111 0100 0011 */ TABDECL(OP_UND),
/* 0111 0100 0100 */ TABDECL(OP_STRB_M_ASR_IMM_OFF),		// OP_STRB_M_ASR_IMM_OFF
/* 0111 0100 0101 */ TABDECL(OP_UND),
/* 0111 0100 0110 */ TABDECL(OP_STRB_M_ROR_IMM_OFF),		// OP_STRB_M_ROR_IMM_OFF
/* 0111 0100 0111 */ TABDECL(OP_UND),
/* 0111 0100 1000 */ TABDECL(OP_STRB_M_LSL_IMM_OFF),		// OP_STRB_M_LSL_IMM_OFF
/* 0111 0100 1001 */ TABDECL(OP_UND),
/* 0111 0100 1010 */ TABDECL(OP_STRB_M_LSR_IMM_OFF),		// OP_STRB_M_LSR_IMM_OFF
/* 0111 0100 1011 */ TABDECL(OP_UND),
/* 0111 0100 1100 */ TABDECL(OP_STRB_M_ASR_IMM_OFF),		// OP_STRB_M_ASR_IMM_OFF
/* 0111 0100 1101 */ TABDECL(OP_UND),
/* 0111 0100 1110 */ TABDECL(OP_STRB_M_ROR_IMM_OFF),		// OP_STRB_M_ROR_IMM_OFF
/* 0111 0100 1111 */ TABDECL(OP_UND),

/* 0111 0101 0000 */ TABDECL(OP_LDRB_M_LSL_IMM_OFF),		// OP_LDRB_M_LSL_IMM_OFF
/* 0111 0101 0001 */ TABDECL(OP_UND),
/* 0111 0101 0010 */ TABDECL(OP_LDRB_M_LSR_IMM_OFF),		// OP_LDRB_M_LSR_IMM_OFF
/* 0111 0101 0011 */ TABDECL(OP_UND),
/* 0111 0101 0100 */ TABDECL(OP_LDRB_M_ASR_IMM_OFF),		// OP_LDRB_M_ASR_IMM_OFF
/* 0111 0101 0101 */ TABDECL(OP_UND),
/* 0111 0101 0110 */ TABDECL(OP_LDRB_M_ROR_IMM_OFF),		// OP_LDRB_M_ROR_IMM_OFF
/* 0111 0101 0111 */ TABDECL(OP_UND),
/* 0111 0101 1000 */ TABDECL(OP_LDRB_M_LSL_IMM_OFF),		// OP_LDRB_M_LSL_IMM_OFF
/* 0111 0101 1001 */ TABDECL(OP_UND),
/* 0111 0101 1010 */ TABDECL(OP_LDRB_M_LSR_IMM_OFF),		// OP_LDRB_M_LSR_IMM_OFF
/* 0111 0101 1011 */ TABDECL(OP_UND),
/* 0111 0101 1100 */ TABDECL(OP_LDRB_M_ASR_IMM_OFF),		// OP_LDRB_M_ASR_IMM_OFF
/* 0111 0101 1101 */ TABDECL(OP_UND),
/* 0111 0101 1110 */ TABDECL(OP_LDRB_M_ROR_IMM_OFF),		// OP_LDRB_M_ROR_IMM_OFF
/* 0111 0101 1111 */ TABDECL(OP_UND),

/* 0111 0110 0000 */ TABDECL(OP_STRB_M_LSL_IMM_OFF_PREIND),		// OP_STRB_M_LSL_IMM_OFF_PREIND
/* 0111 0110 0001 */ TABDECL(OP_UND),
/* 0111 0110 0010 */ TABDECL(OP_STRB_M_LSR_IMM_OFF_PREIND),		// OP_STRB_M_LSR_IMM_OFF_PREIND
/* 0111 0110 0011 */ TABDECL(OP_UND),
/* 0111 0110 0100 */ TABDECL(OP_STRB_M_ASR_IMM_OFF_PREIND),		// OP_STRB_M_ASR_IMM_OFF_PREIND
/* 0111 0110 0101 */ TABDECL(OP_UND),
/* 0111 0110 0110 */ TABDECL(OP_STRB_M_ROR_IMM_OFF_PREIND),		// OP_STRB_M_ROR_IMM_OFF_PREIND
/* 0111 0110 0111 */ TABDECL(OP_UND),
/* 0111 0110 1000 */ TABDECL(OP_STRB_M_LSL_IMM_OFF_PREIND),		// OP_STRB_M_LSL_IMM_OFF_PREIND
/* 0111 0110 1001 */ TABDECL(OP_UND),
/* 0111 0110 1010 */ TABDECL(OP_STRB_M_LSR_IMM_OFF_PREIND),		// OP_STRB_M_LSR_IMM_OFF_PREIND
/* 0111 0110 1011 */ TABDECL(OP_UND),
/* 0111 0110 1100 */ TABDECL(OP_STRB_M_ASR_IMM_OFF_PREIND),		// OP_STRB_M_ASR_IMM_OFF_PREIND
/* 0111 0110 1101 */ TABDECL(OP_UND),
/* 0111 0110 1110 */ TABDECL(OP_STRB_M_ROR_IMM_OFF_PREIND),		// OP_STRB_M_ROR_IMM_OFF_PREIND
/* 0111 0110 1111 */ TABDECL(OP_UND),

/* 0111 0111 0000 */ TABDECL(OP_LDRB_M_LSL_IMM_OFF_PREIND),		// OP_LDRB_M_LSL_IMM_OFF_PREIND
/* 0111 0111 0001 */ TABDECL(OP_UND),
/* 0111 0111 0010 */ TABDECL(OP_LDRB_M_LSR_IMM_OFF_PREIND),		// OP_LDRB_M_LSR_IMM_OFF_PREIND
/* 0111 0111 0011 */ TABDECL(OP_UND),
/* 0111 0111 0100 */ TABDECL(OP_LDRB_M_ASR_IMM_OFF_PREIND),		// OP_LDRB_M_ASR_IMM_OFF_PREIND
/* 0111 0111 0101 */ TABDECL(OP_UND),
/* 0111 0111 0110 */ TABDECL(OP_LDRB_M_ROR_IMM_OFF_PREIND),		// OP_LDRB_M_ROR_IMM_OFF_PREIND
/* 0111 0111 0111 */ TABDECL(OP_UND),
/* 0111 0111 1000 */ TABDECL(OP_LDRB_M_LSL_IMM_OFF_PREIND),		// OP_LDRB_M_LSL_IMM_OFF_PREIND
/* 0111 0111 1001 */ TABDECL(OP_UND),
/* 0111 0111 1010 */ TABDECL(OP_LDRB_M_LSR_IMM_OFF_PREIND),		// OP_LDRB_M_LSR_IMM_OFF_PREIND
/* 0111 0111 1011 */ TABDECL(OP_UND),
/* 0111 0111 1100 */ TABDECL(OP_LDRB_M_ASR_IMM_OFF_PREIND),		// OP_LDRB_M_ASR_IMM_OFF_PREIND
/* 0111 0111 1101 */ TABDECL(OP_UND),
/* 0111 0111 1110 */ TABDECL(OP_LDRB_M_ROR_IMM_OFF_PREIND),		// OP_LDRB_M_ROR_IMM_OFF_PREIND
/* 0111 0111 1111 */ TABDECL(OP_UND),

/* 0111 1000 0000 */ TABDECL(OP_STR_P_LSL_IMM_OFF),		// OP_STR_P_LSL_IMM_OFF
/* 0111 1000 0001 */ TABDECL(OP_UND),
/* 0111 1000 0010 */ TABDECL(OP_STR_P_LSR_IMM_OFF),		// OP_STR_P_LSR_IMM_OFF
/* 0111 1000 0011 */ TABDECL(OP_UND),
/* 0111 1000 0100 */ TABDECL(OP_STR_P_ASR_IMM_OFF),		// OP_STR_P_ASR_IMM_OFF
/* 0111 1000 0101 */ TABDECL(OP_UND),
/* 0111 1000 0110 */ TABDECL(OP_STR_P_ROR_IMM_OFF),		// OP_STR_P_ROR_IMM_OFF
/* 0111 1000 0111 */ TABDECL(OP_UND),
/* 0111 1000 1000 */ TABDECL(OP_STR_P_LSL_IMM_OFF),		// OP_STR_P_LSL_IMM_OFF
/* 0111 1000 1001 */ TABDECL(OP_UND),
/* 0111 1000 1010 */ TABDECL(OP_STR_P_LSR_IMM_OFF),		// OP_STR_P_LSR_IMM_OFF
/* 0111 1000 1011 */ TABDECL(OP_UND),
/* 0111 1000 1100 */ TABDECL(OP_STR_P_ASR_IMM_OFF),		// OP_STR_P_ASR_IMM_OFF
/* 0111 1000 1101 */ TABDECL(OP_UND),
/* 0111 1000 1110 */ TABDECL(OP_STR_P_ROR_IMM_OFF),		// OP_STR_P_ROR_IMM_OFF
/* 0111 1000 1111 */ TABDECL(OP_UND),

/* 0111 1001 0000 */ TABDECL(OP_LDR_P_LSL_IMM_OFF),		// OP_LDR_P_LSL_IMM_OFF
/* 0111 1001 0001 */ TABDECL(OP_UND),
/* 0111 1001 0010 */ TABDECL(OP_LDR_P_LSR_IMM_OFF),		// OP_LDR_P_LSR_IMM_OFF
/* 0111 1001 0011 */ TABDECL(OP_UND),
/* 0111 1001 0100 */ TABDECL(OP_LDR_P_ASR_IMM_OFF),		// OP_LDR_P_ASR_IMM_OFF
/* 0111 1001 0101 */ TABDECL(OP_UND),
/* 0111 1001 0110 */ TABDECL(OP_LDR_P_ROR_IMM_OFF),		// OP_LDR_P_ROR_IMM_OFF
/* 0111 1001 0111 */ TABDECL(OP_UND),
/* 0111 1001 1000 */ TABDECL(OP_LDR_P_LSL_IMM_OFF),		// OP_LDR_P_LSL_IMM_OFF
/* 0111 1001 1001 */ TABDECL(OP_UND),
/* 0111 1001 1010 */ TABDECL(OP_LDR_P_LSR_IMM_OFF),		// OP_LDR_P_LSR_IMM_OFF
/* 0111 1001 1011 */ TABDECL(OP_UND),
/* 0111 1001 1100 */ TABDECL(OP_LDR_P_ASR_IMM_OFF),		// OP_LDR_P_ASR_IMM_OFF
/* 0111 1001 1101 */ TABDECL(OP_UND),
/* 0111 1001 1110 */ TABDECL(OP_LDR_P_ROR_IMM_OFF),		// OP_LDR_P_ROR_IMM_OFF
/* 0111 1001 1111 */ TABDECL(OP_UND),

/* 0111 1010 0000 */ TABDECL(OP_STR_P_LSL_IMM_OFF_PREIND),		// OP_STR_P_LSL_IMM_OFF_PREIND
/* 0111 1010 0001 */ TABDECL(OP_UND),
/* 0111 1010 0010 */ TABDECL(OP_STR_P_LSR_IMM_OFF_PREIND),		// OP_STR_P_LSR_IMM_OFF_PREIND
/* 0111 1010 0011 */ TABDECL(OP_UND),
/* 0111 1010 0100 */ TABDECL(OP_STR_P_ASR_IMM_OFF_PREIND),		// OP_STR_P_ASR_IMM_OFF_PREIND
/* 0111 1010 0101 */ TABDECL(OP_UND),
/* 0111 1010 0110 */ TABDECL(OP_STR_P_ROR_IMM_OFF_PREIND),		// OP_STR_P_ROR_IMM_OFF_PREIND
/* 0111 1010 0111 */ TABDECL(OP_UND),
/* 0111 1010 1000 */ TABDECL(OP_STR_P_LSL_IMM_OFF_PREIND),		// OP_STR_P_LSL_IMM_OFF_PREIND
/* 0111 1010 1001 */ TABDECL(OP_UND),
/* 0111 1010 1010 */ TABDECL(OP_STR_P_LSR_IMM_OFF_PREIND),		// OP_STR_P_LSR_IMM_OFF_PREIND
/* 0111 1010 1011 */ TABDECL(OP_UND),
/* 0111 1010 1100 */ TABDECL(OP_STR_P_ASR_IMM_OFF_PREIND),		// OP_STR_P_ASR_IMM_OFF_PREIND
/* 0111 1010 1101 */ TABDECL(OP_UND),
/* 0111 1010 1110 */ TABDECL(OP_STR_P_ROR_IMM_OFF_PREIND),		// OP_STR_P_ROR_IMM_OFF_PREIND
/* 0111 1010 1111 */ TABDECL(OP_UND),

/* 0111 1011 0000 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_PREIND),		// OP_LDR_P_LSL_IMM_OFF_PREIND
/* 0111 1011 0001 */ TABDECL(OP_UND),
/* 0111 1011 0010 */ TABDECL(OP_LDR_P_LSR_IMM_OFF_PREIND),		// OP_LDR_P_LSR_IMM_OFF_PREIND
/* 0111 1011 0011 */ TABDECL(OP_UND),
/* 0111 1011 0100 */ TABDECL(OP_LDR_P_ASR_IMM_OFF_PREIND),		// OP_LDR_P_ASR_IMM_OFF_PREIND
/* 0111 1011 0101 */ TABDECL(OP_UND),
/* 0111 1011 0110 */ TABDECL(OP_LDR_P_ROR_IMM_OFF_PREIND),		// OP_LDR_P_ROR_IMM_OFF_PREIND
/* 0111 1011 0111 */ TABDECL(OP_UND),
/* 0111 1011 1000 */ TABDECL(OP_LDR_P_LSL_IMM_OFF_PREIND),		// OP_LDR_P_LSL_IMM_OFF_PREIND
/* 0111 1011 1001 */ TABDECL(OP_UND),
/* 0111 1011 1010 */ TABDECL(OP_LDR_P_LSR_IMM_OFF_PREIND),		// OP_LDR_P_LSR_IMM_OFF_PREIND
/* 0111 1011 1011 */ TABDECL(OP_UND),
/* 0111 1011 1100 */ TABDECL(OP_LDR_P_ASR_IMM_OFF_PREIND),		// OP_LDR_P_ASR_IMM_OFF_PREIND
/* 0111 1011 1101 */ TABDECL(OP_UND),
/* 0111 1011 1110 */ TABDECL(OP_LDR_P_ROR_IMM_OFF_PREIND),		// OP_LDR_P_ROR_IMM_OFF_PREIND
/* 0111 1011 1111 */ TABDECL(OP_UND),

/* 0111 1100 0000 */ TABDECL(OP_STRB_P_LSL_IMM_OFF),		// OP_STRB_P_LSL_IMM_OFF
/* 0111 1100 0001 */ TABDECL(OP_UND),
/* 0111 1100 0010 */ TABDECL(OP_STRB_P_LSR_IMM_OFF),		// OP_STRB_P_LSR_IMM_OFF
/* 0111 1100 0011 */ TABDECL(OP_UND),
/* 0111 1100 0100 */ TABDECL(OP_STRB_P_ASR_IMM_OFF),		// OP_STRB_P_ASR_IMM_OFF
/* 0111 1100 0101 */ TABDECL(OP_UND),
/* 0111 1100 0110 */ TABDECL(OP_STRB_P_ROR_IMM_OFF),		// OP_STRB_P_ROR_IMM_OFF
/* 0111 1100 0111 */ TABDECL(OP_UND),
/* 0111 1100 1000 */ TABDECL(OP_STRB_P_LSL_IMM_OFF),		// OP_STRB_P_LSL_IMM_OFF
/* 0111 1100 1001 */ TABDECL(OP_UND),
/* 0111 1100 1010 */ TABDECL(OP_STRB_P_LSR_IMM_OFF),		// OP_STRB_P_LSR_IMM_OFF
/* 0111 1100 1011 */ TABDECL(OP_UND),
/* 0111 1100 1100 */ TABDECL(OP_STRB_P_ASR_IMM_OFF),		// OP_STRB_P_ASR_IMM_OFF
/* 0111 1100 1101 */ TABDECL(OP_UND),
/* 0111 1100 1110 */ TABDECL(OP_STRB_P_ROR_IMM_OFF),		// OP_STRB_P_ROR_IMM_OFF
/* 0111 1100 1111 */ TABDECL(OP_UND),

/* 0111 1101 0000 */ TABDECL(OP_LDRB_P_LSL_IMM_OFF),		// OP_LDRB_P_LSL_IMM_OFF
/* 0111 1101 0001 */ TABDECL(OP_UND),
/* 0111 1101 0010 */ TABDECL(OP_LDRB_P_LSR_IMM_OFF),		// OP_LDRB_P_LSR_IMM_OFF
/* 0111 1101 0011 */ TABDECL(OP_UND),
/* 0111 1101 0100 */ TABDECL(OP_LDRB_P_ASR_IMM_OFF),		// OP_LDRB_P_ASR_IMM_OFF
/* 0111 1101 0101 */ TABDECL(OP_UND),
/* 0111 1101 0110 */ TABDECL(OP_LDRB_P_ROR_IMM_OFF),		// OP_LDRB_P_ROR_IMM_OFF
/* 0111 1101 0111 */ TABDECL(OP_UND),
/* 0111 1101 1000 */ TABDECL(OP_LDRB_P_LSL_IMM_OFF),		// OP_LDRB_P_LSL_IMM_OFF
/* 0111 1101 1001 */ TABDECL(OP_UND),
/* 0111 1101 1010 */ TABDECL(OP_LDRB_P_LSR_IMM_OFF),		// OP_LDRB_P_LSR_IMM_OFF
/* 0111 1101 1011 */ TABDECL(OP_UND),
/* 0111 1101 1100 */ TABDECL(OP_LDRB_P_ASR_IMM_OFF),		// OP_LDRB_P_ASR_IMM_OFF
/* 0111 1101 1101 */ TABDECL(OP_UND),
/* 0111 1101 1110 */ TABDECL(OP_LDRB_P_ROR_IMM_OFF),		// OP_LDRB_P_ROR_IMM_OFF
/* 0111 1101 1111 */ TABDECL(OP_UND),

/* 0111 1110 0000 */ TABDECL(OP_STRB_P_LSL_IMM_OFF_PREIND),		// OP_STRB_P_LSL_IMM_OFF_PREIND
/* 0111 1110 0001 */ TABDECL(OP_UND),
/* 0111 1110 0010 */ TABDECL(OP_STRB_P_LSR_IMM_OFF_PREIND),		// OP_STRB_P_LSR_IMM_OFF_PREIND
/* 0111 1110 0011 */ TABDECL(OP_UND),
/* 0111 1110 0100 */ TABDECL(OP_STRB_P_ASR_IMM_OFF_PREIND),		// OP_STRB_P_ASR_IMM_OFF_PREIND
/* 0111 1110 0101 */ TABDECL(OP_UND),
/* 0111 1110 0110 */ TABDECL(OP_STRB_P_ROR_IMM_OFF_PREIND),		// OP_STRB_P_ROR_IMM_OFF_PREIND
/* 0111 1110 0111 */ TABDECL(OP_UND),
/* 0111 1110 1000 */ TABDECL(OP_STRB_P_LSL_IMM_OFF_PREIND),		// OP_STRB_P_LSL_IMM_OFF_PREIND
/* 0111 1110 1001 */ TABDECL(OP_UND),
/* 0111 1110 1010 */ TABDECL(OP_STRB_P_LSR_IMM_OFF_PREIND),		// OP_STRB_P_LSR_IMM_OFF_PREIND
/* 0111 1110 1011 */ TABDECL(OP_UND),
/* 0111 1110 1100 */ TABDECL(OP_STRB_P_ASR_IMM_OFF_PREIND),		// OP_STRB_P_ASR_IMM_OFF_PREIND
/* 0111 1110 1101 */ TABDECL(OP_UND),
/* 0111 1110 1110 */ TABDECL(OP_STRB_P_ROR_IMM_OFF_PREIND),		// OP_STRB_P_ROR_IMM_OFF_PREIND
/* 0111 1110 1111 */ TABDECL(OP_UND),

/* 0111 1111 0000 */ TABDECL(OP_LDRB_P_LSL_IMM_OFF_PREIND),		// OP_LDRB_P_LSL_IMM_OFF_PREIND
/* 0111 1111 0001 */ TABDECL(OP_UND),
/* 0111 1111 0010 */ TABDECL(OP_LDRB_P_LSR_IMM_OFF_PREIND),		// OP_LDRB_P_LSR_IMM_OFF_PREIND
/* 0111 1111 0011 */ TABDECL(OP_UND),
/* 0111 1111 0100 */ TABDECL(OP_LDRB_P_ASR_IMM_OFF_PREIND),		// OP_LDRB_P_ASR_IMM_OFF_PREIND
/* 0111 1111 0101 */ TABDECL(OP_UND),
/* 0111 1111 0110 */ TABDECL(OP_LDRB_P_ROR_IMM_OFF_PREIND),		// OP_LDRB_P_ROR_IMM_OFF_PREIND
/* 0111 1111 0111 */ TABDECL(OP_UND),
/* 0111 1111 1000 */ TABDECL(OP_LDRB_P_LSL_IMM_OFF_PREIND),		// OP_LDRB_P_LSL_IMM_OFF_PREIND
/* 0111 1111 1001 */ TABDECL(OP_UND),
/* 0111 1111 1010 */ TABDECL(OP_LDRB_P_LSR_IMM_OFF_PREIND),		// OP_LDRB_P_LSR_IMM_OFF_PREIND
/* 0111 1111 1011 */ TABDECL(OP_UND),
/* 0111 1111 1100 */ TABDECL(OP_LDRB_P_ASR_IMM_OFF_PREIND),		// OP_LDRB_P_ASR_IMM_OFF_PREIND
/* 0111 1111 1101 */ TABDECL(OP_UND),
/* 0111 1111 1110 */ TABDECL(OP_LDRB_P_ROR_IMM_OFF_PREIND),		// OP_LDRB_P_ROR_IMM_OFF_PREIND
/* 0111 1111 1111 */ TABDECL(OP_UND),

/* 1000 0000 0000 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 0001 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 0010 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 0011 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 0100 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 0101 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 0110 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 0111 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 1000 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 1001 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 1010 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 1011 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 1100 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 1101 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 1110 */ TABDECL(OP_STMDA),		// OP_STMDA
/* 1000 0000 1111 */ TABDECL(OP_STMDA),		// OP_STMDA

/* 1000 0001 0000 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 0001 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 0010 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 0011 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 0100 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 0101 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 0110 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 0111 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 1000 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 1001 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 1010 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 1011 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 1100 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 1101 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 1110 */ TABDECL(OP_LDMDA),		// OP_LDMDA
/* 1000 0001 1111 */ TABDECL(OP_LDMDA),		// OP_LDMDA

/* 1000 0010 0000 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 0001 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 0010 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 0011 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 0100 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 0101 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 0110 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 0111 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 1000 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 1001 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 1010 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 1011 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 1100 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 1101 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 1110 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W
/* 1000 0010 1111 */ TABDECL(OP_STMDA_W),		// OP_STMDA_W

/* 1000 0011 0000 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 0001 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 0010 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 0011 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 0100 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 0101 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 0110 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 0111 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 1000 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 1001 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 1010 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 1011 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 1100 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 1101 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 1110 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W
/* 1000 0011 1111 */ TABDECL(OP_LDMDA_W),		// OP_LDMDA_W

/* 1000 0100 0000 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 0001 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 0010 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 0011 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 0100 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 0101 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 0110 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 0111 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 1000 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 1001 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 1010 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 1011 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 1100 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 1101 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 1110 */ TABDECL(OP_STMDA2),		// OP_STMDA2
/* 1000 0100 1111 */ TABDECL(OP_STMDA2),		// OP_STMDA2

/* 1000 0101 0000 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 0001 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 0010 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 0011 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 0100 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 0101 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 0110 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 0111 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 1000 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 1001 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 1010 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 1011 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 1100 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 1101 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 1110 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2
/* 1000 0101 1111 */ TABDECL(OP_LDMDA2),		// OP_LDMDA2

/* 1000 0110 0000 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 0001 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 0010 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 0011 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 0100 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 0101 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 0110 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 0111 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 1000 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 1001 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 1010 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 1011 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 1100 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 1101 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 1110 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W
/* 1000 0110 1111 */ TABDECL(OP_STMDA2_W),		// OP_STMDA2_W

/* 1000 0111 0000 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 0001 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 0010 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 0011 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 0100 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 0101 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 0110 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 0111 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 1000 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 1001 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 1010 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 1011 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 1100 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 1101 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 1110 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W
/* 1000 0111 1111 */ TABDECL(OP_LDMDA2_W),		// OP_LDMDA2_W

/* 1000 1000 0000 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 0001 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 0010 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 0011 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 0100 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 0101 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 0110 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 0111 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 1000 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 1001 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 1010 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 1011 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 1100 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 1101 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 1110 */ TABDECL(OP_STMIA),		// OP_STMIA
/* 1000 1000 1111 */ TABDECL(OP_STMIA),		// OP_STMIA

/* 1000 1001 0000 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 0001 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 0010 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 0011 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 0100 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 0101 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 0110 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 0111 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 1000 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 1001 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 1010 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 1011 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 1100 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 1101 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 1110 */ TABDECL(OP_LDMIA),		// OP_LDMIA
/* 1000 1001 1111 */ TABDECL(OP_LDMIA),		// OP_LDMIA

/* 1000 1010 0000 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 0001 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 0010 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 0011 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 0100 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 0101 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 0110 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 0111 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 1000 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 1001 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 1010 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 1011 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 1100 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 1101 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 1110 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W
/* 1000 1010 1111 */ TABDECL(OP_STMIA_W),		// OP_STMIA_W

/* 1000 1011 0000 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 0001 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 0010 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 0011 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 0100 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 0101 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 0110 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 0111 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 1000 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 1001 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 1010 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 1011 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 1100 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 1101 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 1110 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W
/* 1000 1011 1111 */ TABDECL(OP_LDMIA_W),		// OP_LDMIA_W

/* 1000 1100 0000 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 0001 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 0010 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 0011 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 0100 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 0101 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 0110 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 0111 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 1000 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 1001 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 1010 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 1011 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 1100 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 1101 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 1110 */ TABDECL(OP_STMIA2),		// OP_STMIA2
/* 1000 1100 1111 */ TABDECL(OP_STMIA2),		// OP_STMIA2

/* 1000 1101 0000 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 0001 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 0010 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 0011 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 0100 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 0101 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 0110 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 0111 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 1000 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 1001 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 1010 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 1011 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 1100 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 1101 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 1110 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2
/* 1000 1101 1111 */ TABDECL(OP_LDMIA2),		// OP_LDMIA2

/* 1000 1110 0000 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 0001 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 0010 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 0011 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 0100 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 0101 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 0110 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 0111 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 1000 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 1001 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 1010 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 1011 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 1100 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 1101 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 1110 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W
/* 1000 1110 1111 */ TABDECL(OP_STMIA2_W),		// OP_STMIA2_W

/* 1000 1111 0000 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 0001 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 0010 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 0011 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 0100 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 0101 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 0110 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 0111 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 1000 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 1001 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 1010 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 1011 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 1100 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 1101 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 1110 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W
/* 1000 1111 1111 */ TABDECL(OP_LDMIA2_W),		// OP_LDMIA2_W

/* 1001 0000 0000 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 0001 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 0010 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 0011 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 0100 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 0101 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 0110 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 0111 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 1000 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 1001 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 1010 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 1011 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 1100 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 1101 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 1110 */ TABDECL(OP_STMDB),		// OP_STMDB
/* 1001 0000 1111 */ TABDECL(OP_STMDB),		// OP_STMDB

/* 1001 0001 0000 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 0001 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 0010 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 0011 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 0100 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 0101 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 0110 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 0111 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 1000 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 1001 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 1010 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 1011 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 1100 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 1101 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 1110 */ TABDECL(OP_LDMDB),		// OP_LDMDB
/* 1001 0001 1111 */ TABDECL(OP_LDMDB),		// OP_LDMDB

/* 1001 0010 0000 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 0001 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 0010 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 0011 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 0100 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 0101 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 0110 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 0111 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 1000 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 1001 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 1010 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 1011 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 1100 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 1101 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 1110 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W
/* 1001 0010 1111 */ TABDECL(OP_STMDB_W),		// OP_STMDB_W

/* 1001 0011 0000 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 0001 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 0010 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 0011 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 0100 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 0101 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 0110 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 0111 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 1000 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 1001 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 1010 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 1011 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 1100 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 1101 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 1110 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W
/* 1001 0011 1111 */ TABDECL(OP_LDMDB_W),		// OP_LDMDB_W

/* 1001 0100 0000 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 0001 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 0010 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 0011 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 0100 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 0101 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 0110 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 0111 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 1000 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 1001 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 1010 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 1011 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 1100 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 1101 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 1110 */ TABDECL(OP_STMDB2),		// OP_STMDB2
/* 1001 0100 1111 */ TABDECL(OP_STMDB2),		// OP_STMDB2

/* 1001 0101 0000 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 0001 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 0010 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 0011 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 0100 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 0101 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 0110 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 0111 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 1000 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 1001 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 1010 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 1011 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 1100 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 1101 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 1110 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2
/* 1001 0101 1111 */ TABDECL(OP_LDMDB2),		// OP_LDMDB2

/* 1001 0110 0000 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 0001 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 0010 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 0011 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 0100 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 0101 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 0110 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 0111 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 1000 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 1001 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 1010 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 1011 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 1100 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 1101 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 1110 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W
/* 1001 0110 1111 */ TABDECL(OP_STMDB2_W),		// OP_STMDB2_W

/* 1001 0111 0000 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 0001 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 0010 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 0011 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 0100 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 0101 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 0110 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 0111 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 1000 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 1001 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 1010 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 1011 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 1100 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 1101 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 1110 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W
/* 1001 0111 1111 */ TABDECL(OP_LDMDB2_W),		// OP_LDMDB2_W

/* 1001 1000 0000 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 0001 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 0010 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 0011 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 0100 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 0101 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 0110 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 0111 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 1000 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 1001 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 1010 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 1011 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 1100 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 1101 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 1110 */ TABDECL(OP_STMIB),		// OP_STMIB
/* 1001 1000 1111 */ TABDECL(OP_STMIB),		// OP_STMIB

/* 1001 1001 0000 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 0001 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 0010 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 0011 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 0100 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 0101 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 0110 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 0111 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 1000 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 1001 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 1010 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 1011 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 1100 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 1101 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 1110 */ TABDECL(OP_LDMIB),		// OP_LDMIB
/* 1001 1001 1111 */ TABDECL(OP_LDMIB),		// OP_LDMIB

/* 1001 1010 0000 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 0001 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 0010 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 0011 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 0100 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 0101 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 0110 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 0111 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 1000 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 1001 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 1010 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 1011 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 1100 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 1101 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 1110 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W
/* 1001 1010 1111 */ TABDECL(OP_STMIB_W),		// OP_STMIB_W

/* 1001 1011 0000 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 0001 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 0010 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 0011 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 0100 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 0101 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 0110 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 0111 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 1000 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 1001 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 1010 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 1011 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 1100 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 1101 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 1110 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W
/* 1001 1011 1111 */ TABDECL(OP_LDMIB_W),		// OP_LDMIB_W

/* 1001 1100 0000 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 0001 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 0010 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 0011 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 0100 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 0101 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 0110 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 0111 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 1000 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 1001 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 1010 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 1011 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 1100 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 1101 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 1110 */ TABDECL(OP_STMIB2),		// OP_STMIB2
/* 1001 1100 1111 */ TABDECL(OP_STMIB2),		// OP_STMIB2

/* 1001 1101 0000 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 0001 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 0010 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 0011 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 0100 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 0101 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 0110 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 0111 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 1000 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 1001 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 1010 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 1011 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 1100 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 1101 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 1110 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2
/* 1001 1101 1111 */ TABDECL(OP_LDMIB2),		// OP_LDMIB2

/* 1001 1110 0000 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 0001 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 0010 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 0011 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 0100 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 0101 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 0110 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 0111 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 1000 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 1001 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 1010 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 1011 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 1100 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 1101 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 1110 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W
/* 1001 1110 1111 */ TABDECL(OP_STMIB2_W),		// OP_STMIB2_W

/* 1001 1111 0000 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 0001 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 0010 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 0011 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 0100 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 0101 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 0110 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 0111 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 1000 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 1001 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 1010 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 1011 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 1100 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 1101 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 1110 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W
/* 1001 1111 1111 */ TABDECL(OP_LDMIB2_W),		// OP_LDMIB2_W

/* 1010 0000 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 0000 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 0001 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 0001 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 0010 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 0010 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 0011 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 0011 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 0100 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 0100 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 0101 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 0101 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 0110 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 0110 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 0111 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 0111 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 1000 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 1000 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 1001 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 1001 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 1010 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 1010 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 1011 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 1011 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 1100 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 1100 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 1101 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 1101 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 1110 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 1110 1111 */ TABDECL(OP_B),		// OP_B

/* 1010 1111 0000 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 0001 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 0010 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 0011 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 0100 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 0101 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 0110 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 0111 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 1000 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 1001 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 1010 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 1011 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 1100 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 1101 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 1110 */ TABDECL(OP_B),		// OP_B
/* 1010 1111 1111 */ TABDECL(OP_B),		// OP_B

/* 1011 0000 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0000 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 0001 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0001 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 0010 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0010 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 0011 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0011 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 0100 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0100 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 0101 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0101 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 0110 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0110 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 0111 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 0111 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 1000 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1000 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 1001 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1001 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 1010 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1010 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 1011 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1011 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 1100 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1100 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 1101 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1101 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 1110 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1110 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1011 1111 0000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 0001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 0010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 0011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 0100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 0101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 0110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 0111 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 1000 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 1001 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 1010 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 1011 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 1100 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 1101 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 1110 */ TABDECL(OP_BL),		// OP_BL
/* 1011 1111 1111 */ TABDECL(OP_BL),		// OP_BL

/* 1100 0000 0000 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 0001 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 0010 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 0011 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 0100 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 0101 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 0110 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 0111 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 1000 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 1001 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 1010 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 1011 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 1100 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 1101 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 1110 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0000 1111 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION

/* 1100 0001 0000 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 0001 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 0010 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 0011 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 0100 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 0101 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 0110 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 0111 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 1000 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 1001 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 1010 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 1011 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 1100 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 1101 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 1110 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0001 1111 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION

/* 1100 0010 0000 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 0001 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 0010 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 0011 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 0100 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 0101 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 0110 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 0111 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 1000 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 1001 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 1010 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 1011 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 1100 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 1101 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 1110 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0010 1111 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND

/* 1100 0011 0000 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 0001 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 0010 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 0011 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 0100 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 0101 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 0110 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 0111 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 1000 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 1001 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 1010 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 1011 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 1100 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 1101 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 1110 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0011 1111 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND

/* 1100 0100 0000 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 0001 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 0010 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 0011 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 0100 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 0101 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 0110 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 0111 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 1000 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 1001 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 1010 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 1011 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 1100 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 1101 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 1110 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 0100 1111 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION

/* 1100 0101 0000 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 0001 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 0010 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 0011 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 0100 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 0101 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 0110 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 0111 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 1000 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 1001 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 1010 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 1011 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 1100 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 1101 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 1110 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 0101 1111 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION

/* 1100 0110 0000 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 0001 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 0010 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 0011 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 0100 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 0101 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 0110 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 0111 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 1000 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 1001 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 1010 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 1011 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 1100 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 1101 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 1110 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND
/* 1100 0110 1111 */ TABDECL(OP_STC_M_POSTIND),		// OP_STC_M_POSTIND

/* 1100 0111 0000 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 0001 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 0010 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 0011 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 0100 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 0101 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 0110 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 0111 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 1000 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 1001 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 1010 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 1011 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 1100 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 1101 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 1110 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND
/* 1100 0111 1111 */ TABDECL(OP_LDC_M_POSTIND),		// OP_LDC_M_POSTIND

/* 1100 1000 0000 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 0001 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 0010 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 0011 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 0100 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 0101 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 0110 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 0111 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 1000 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 1001 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 1010 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 1011 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 1100 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 1101 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 1110 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1000 1111 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION

/* 1100 1001 0000 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 0001 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 0010 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 0011 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 0100 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 0101 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 0110 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 0111 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 1000 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 1001 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 1010 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 1011 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 1100 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 1101 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 1110 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1001 1111 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION

/* 1100 1010 0000 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 0001 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 0010 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 0011 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 0100 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 0101 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 0110 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 0111 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 1000 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 1001 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 1010 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 1011 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 1100 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 1101 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 1110 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1010 1111 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND

/* 1100 1011 0000 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 0001 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 0010 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 0011 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 0100 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 0101 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 0110 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 0111 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 1000 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 1001 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 1010 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 1011 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 1100 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 1101 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 1110 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1011 1111 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND

/* 1100 1100 0000 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 0001 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 0010 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 0011 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 0100 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 0101 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 0110 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 0111 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 1000 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 1001 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 1010 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 1011 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 1100 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 1101 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 1110 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION
/* 1100 1100 1111 */ TABDECL(OP_STC_OPTION),		// OP_STC_OPTION

/* 1100 1101 0000 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 0001 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 0010 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 0011 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 0100 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 0101 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 0110 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 0111 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 1000 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 1001 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 1010 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 1011 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 1100 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 1101 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 1110 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION
/* 1100 1101 1111 */ TABDECL(OP_LDC_OPTION),		// OP_LDC_OPTION

/* 1100 1110 0000 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 0001 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 0010 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 0011 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 0100 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 0101 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 0110 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 0111 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 1000 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 1001 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 1010 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 1011 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 1100 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 1101 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 1110 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND
/* 1100 1110 1111 */ TABDECL(OP_STC_P_POSTIND),		// OP_STC_P_POSTIND

/* 1100 1111 0000 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 0001 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 0010 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 0011 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 0100 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 0101 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 0110 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 0111 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 1000 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 1001 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 1010 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 1011 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 1100 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 1101 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 1110 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND
/* 1100 1111 1111 */ TABDECL(OP_LDC_P_POSTIND),		// OP_LDC_P_POSTIND

/* 1101 0000 0000 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 0001 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 0010 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 0011 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 0100 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 0101 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 0110 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 0111 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 1000 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 1001 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 1010 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 1011 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 1100 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 1101 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 1110 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0000 1111 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF

/* 1101 0001 0000 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 0001 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 0010 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 0011 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 0100 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 0101 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 0110 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 0111 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 1000 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 1001 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 1010 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 1011 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 1100 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 1101 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 1110 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0001 1111 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF

/* 1101 0010 0000 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 0001 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 0010 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 0011 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 0100 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 0101 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 0110 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 0111 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 1000 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 1001 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 1010 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 1011 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 1100 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 1101 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 1110 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0010 1111 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND

/* 1101 0011 0000 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 0001 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 0010 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 0011 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 0100 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 0101 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 0110 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 0111 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 1000 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 1001 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 1010 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 1011 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 1100 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 1101 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 1110 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0011 1111 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND

/* 1101 0100 0000 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 0001 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 0010 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 0011 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 0100 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 0101 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 0110 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 0111 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 1000 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 1001 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 1010 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 1011 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 1100 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 1101 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 1110 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF
/* 1101 0100 1111 */ TABDECL(OP_STC_M_IMM_OFF),		// OP_STC_M_IMM_OFF

/* 1101 0101 0000 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 0001 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 0010 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 0011 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 0100 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 0101 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 0110 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 0111 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 1000 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 1001 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 1010 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 1011 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 1100 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 1101 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 1110 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF
/* 1101 0101 1111 */ TABDECL(OP_LDC_M_IMM_OFF),		// OP_LDC_M_IMM_OFF

/* 1101 0110 0000 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 0001 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 0010 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 0011 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 0100 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 0101 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 0110 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 0111 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 1000 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 1001 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 1010 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 1011 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 1100 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 1101 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 1110 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND
/* 1101 0110 1111 */ TABDECL(OP_STC_M_PREIND),		// OP_STC_M_PREIND

/* 1101 0111 0000 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 0001 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 0010 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 0011 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 0100 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 0101 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 0110 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 0111 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 1000 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 1001 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 1010 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 1011 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 1100 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 1101 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 1110 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND
/* 1101 0111 1111 */ TABDECL(OP_LDC_M_PREIND),		// OP_LDC_M_PREIND

/* 1101 1000 0000 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 0001 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 0010 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 0011 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 0100 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 0101 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 0110 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 0111 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 1000 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 1001 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 1010 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 1011 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 1100 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 1101 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 1110 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1000 1111 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF

/* 1101 1001 0000 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 0001 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 0010 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 0011 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 0100 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 0101 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 0110 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 0111 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 1000 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 1001 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 1010 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 1011 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 1100 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 1101 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 1110 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1001 1111 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF

/* 1101 1010 0000 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 0001 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 0010 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 0011 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 0100 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 0101 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 0110 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 0111 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 1000 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 1001 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 1010 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 1011 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 1100 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 1101 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 1110 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1010 1111 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND

/* 1101 1011 0000 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 0001 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 0010 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 0011 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 0100 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 0101 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 0110 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 0111 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 1000 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 1001 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 1010 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 1011 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 1100 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 1101 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 1110 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1011 1111 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND

/* 1101 1100 0000 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 0001 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 0010 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 0011 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 0100 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 0101 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 0110 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 0111 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 1000 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 1001 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 1010 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 1011 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 1100 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 1101 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 1110 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF
/* 1101 1100 1111 */ TABDECL(OP_STC_P_IMM_OFF),		// OP_STC_P_IMM_OFF

/* 1101 1101 0000 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 0001 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 0010 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 0011 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 0100 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 0101 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 0110 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 0111 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 1000 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 1001 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 1010 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 1011 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 1100 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 1101 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 1110 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF
/* 1101 1101 1111 */ TABDECL(OP_LDC_P_IMM_OFF),		// OP_LDC_P_IMM_OFF

/* 1101 1110 0000 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 0001 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 0010 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 0011 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 0100 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 0101 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 0110 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 0111 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 1000 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 1001 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 1010 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 1011 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 1100 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 1101 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 1110 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND
/* 1101 1110 1111 */ TABDECL(OP_STC_P_PREIND),		// OP_STC_P_PREIND

/* 1101 1111 0000 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 0001 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 0010 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 0011 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 0100 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 0101 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 0110 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 0111 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 1000 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 1001 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 1010 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 1011 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 1100 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 1101 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 1110 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND
/* 1101 1111 1111 */ TABDECL(OP_LDC_P_PREIND),		// OP_LDC_P_PREIND

/* 1110 0000 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0000 0001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0000 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0000 0011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0000 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0000 0101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0000 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0000 0111 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0000 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0000 1001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0000 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0000 1011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0000 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0000 1101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0000 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0000 1111 */ TABDECL(OP_MCR),		// OP_MCR

/* 1110 0001 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0001 0001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0001 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0001 0011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0001 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0001 0101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0001 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0001 0111 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0001 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0001 1001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0001 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0001 1011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0001 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0001 1101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0001 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0001 1111 */ TABDECL(OP_MRC),		// OP_MRC

/* 1110 0010 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0010 0001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0010 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0010 0011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0010 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0010 0101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0010 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0010 0111 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0010 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0010 1001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0010 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0010 1011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0010 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0010 1101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0010 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0010 1111 */ TABDECL(OP_MCR),		// OP_MCR

/* 1110 0011 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0011 0001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0011 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0011 0011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0011 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0011 0101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0011 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0011 0111 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0011 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0011 1001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0011 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0011 1011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0011 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0011 1101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0011 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0011 1111 */ TABDECL(OP_MRC),		// OP_MRC

/* 1110 0100 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0100 0001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0100 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0100 0011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0100 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0100 0101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0100 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0100 0111 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0100 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0100 1001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0100 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0100 1011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0100 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0100 1101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0100 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0100 1111 */ TABDECL(OP_MCR),		// OP_MCR

/* 1110 0101 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0101 0001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0101 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0101 0011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0101 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0101 0101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0101 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0101 0111 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0101 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0101 1001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0101 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0101 1011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0101 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0101 1101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0101 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0101 1111 */ TABDECL(OP_MRC),		// OP_MRC

/* 1110 0110 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0110 0001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0110 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0110 0011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0110 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0110 0101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0110 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0110 0111 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0110 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0110 1001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0110 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0110 1011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0110 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0110 1101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 0110 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0110 1111 */ TABDECL(OP_MCR),		// OP_MCR

/* 1110 0111 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0111 0001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0111 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0111 0011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0111 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0111 0101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0111 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0111 0111 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0111 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0111 1001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0111 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0111 1011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0111 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0111 1101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 0111 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 0111 1111 */ TABDECL(OP_MRC),		// OP_MRC

/* 1110 1000 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1000 0001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1000 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1000 0011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1000 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1000 0101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1000 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1000 0111 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1000 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1000 1001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1000 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1000 1011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1000 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1000 1101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1000 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1000 1111 */ TABDECL(OP_MCR),		// OP_MCR

/* 1110 1001 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1001 0001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1001 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1001 0011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1001 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1001 0101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1001 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1001 0111 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1001 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1001 1001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1001 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1001 1011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1001 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1001 1101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1001 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1001 1111 */ TABDECL(OP_MRC),		// OP_MRC

/* 1110 1010 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1010 0001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1010 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1010 0011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1010 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1010 0101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1010 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1010 0111 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1010 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1010 1001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1010 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1010 1011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1010 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1010 1101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1010 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1010 1111 */ TABDECL(OP_MCR),		// OP_MCR

/* 1110 1011 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1011 0001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1011 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1011 0011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1011 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1011 0101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1011 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1011 0111 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1011 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1011 1001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1011 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1011 1011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1011 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1011 1101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1011 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1011 1111 */ TABDECL(OP_MRC),		// OP_MRC

/* 1110 1100 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1100 0001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1100 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1100 0011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1100 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1100 0101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1100 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1100 0111 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1100 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1100 1001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1100 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1100 1011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1100 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1100 1101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1100 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1100 1111 */ TABDECL(OP_MCR),		// OP_MCR

/* 1110 1101 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1101 0001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1101 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1101 0011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1101 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1101 0101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1101 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1101 0111 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1101 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1101 1001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1101 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1101 1011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1101 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1101 1101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1101 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1101 1111 */ TABDECL(OP_MRC),		// OP_MRC

/* 1110 1110 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1110 0001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1110 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1110 0011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1110 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1110 0101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1110 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1110 0111 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1110 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1110 1001 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1110 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1110 1011 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1110 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1110 1101 */ TABDECL(OP_MCR),		// OP_MCR
/* 1110 1110 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1110 1111 */ TABDECL(OP_MCR),		// OP_MCR

/* 1110 1111 0000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1111 0001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1111 0010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1111 0011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1111 0100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1111 0101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1111 0110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1111 0111 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1111 1000 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1111 1001 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1111 1010 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1111 1011 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1111 1100 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1111 1101 */ TABDECL(OP_MRC),		// OP_MRC
/* 1110 1111 1110 */ TABDECL(OP_CDP),		// OP_CDP
/* 1110 1111 1111 */ TABDECL(OP_MRC),		// OP_MRC

/* 1111 0000 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0000 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 0001 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0001 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 0010 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0010 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 0011 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0011 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 0100 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0100 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 0101 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0101 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 0110 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0110 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 0111 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 0111 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 1000 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1000 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 1001 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1001 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 1010 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1010 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 1011 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1011 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 1100 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1100 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 1101 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1101 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 1110 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1110 1111 */ TABDECL(OP_SWI),		// OP_SWI

/* 1111 1111 0000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 0001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 0010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 0011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 0100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 0101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 0110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 0111 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 1000 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 1001 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 1010 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 1011 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 1100 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 1101 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 1110 */ TABDECL(OP_SWI),		// OP_SWI
/* 1111 1111 1111 */ TABDECL(OP_SWI),		// OP_SWI

